import modin.pandas as pandas


import load_data
import data_support

result_path = "../../../results/"

#Headerdata for CSV Files
headers = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "return_path", "header_date",
                "client", "folder",
                "mail_relay_count", "mail_virus_scans",
                "message_id", "dkim",
                "cert_data", "smime_signature", "smime_encryption", "autocrypt_data",
                "pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment",
                "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline",
                "pgp_inline_forward",
                "smime_error", "pgp_error", "other_error"]


def enc_signed_overview(df_meta_all, test, debug):
    """
        Ouput the key numbers over all crypto mails.
        * Signature vs Encryption (in %)
        * PGP vs S/MIME (in %)
        * Table with all data
    """
    filename = "general_crypto_overview"
    if test:
        filename = "test_"+filename
    filename = result_path + filename
    
    result = encryption_signatures(df_meta_all, debug)
    # match command with result
    result['Mails'] = df_meta_all.shape[0]
    total = result['Total']
    commands = {}
    commands['TotalMails'] = "{:,}".format(result['Mails'])
    commands['totalSignedMails'] = "{:,}".format(result['Signatures'])
    commands['totalEncryptedMails'] = "{:,}".format(result['Encryption'])
    commands['totalPGPMails'] = "{:,}".format(result['PGP'])
    commands['totalSMIMEMails'] = "{:,}".format(result['SMIME'])
    
    commands['totalSignedMailsPercentToAllMails'] = round(100.00 * result['Signatures']/result['Mails'], 2)
    commands['totalEncryptedMailsPercentToAllMails'] = round(100.00 * result['Encryption']/result['Mails'], 2)
    commands['totalPGPMailsPercent'] = round(100.00 * result['PGP']/total, 2)
    commands['totalSMIMEMailsPercent'] = round(100.00 * result['SMIME']/total, 2)
    
    commands['pgpEncryptedMails'] = "{:,}".format(result['PGP encryption'])
    commands['smimeEncryptedMails'] = "{:,}".format(result['SMIME encryption'])
    commands['pgpEncryptedMailsPercent'] = round(100.00 * result['PGP encryption']/result['Encryption'], 2)
    commands['smimeEncryptedMailsPercent'] = round(100.00 * result['SMIME encryption']/result['Encryption'], 2)
    
    commands['pgpSignedMails'] = "{:,}".format(result['PGP signatures'])
    commands['smimeSignedMails'] = "{:,}".format(result['SMIME signatures'])
    commands['pgpSignedMailsPercent'] = round(100.00 * result['PGP signatures']/result['Signatures'], 2)
    commands['smimeSignedMailsPercent'] = round(100.00 * result['SMIME signatures']/result['Signatures'], 2)


    
    load_data.write_commands(commands, filename + ".tex")
    load_data.latex_table(result, filename + "_table.tex", "", "")

def enc_signed_years(all_mails_df, test, debug, send_only):
    """
        First crypto mail: type, year
        First PGP mail: type, year
        First SMIME mail: type year
        First time SMIME > PGP: year + %
        First time PGP MIME > PGP Inline: year + %
        Snowden 2013 -> 2014: enc mails, signed mails
    """
    year = 1994
    max = 2022
    res = []
    commands = {}
    
    first_signed = True
    first_enc = True
    first_pgp = True
    first_smime = True
    first_more_smime = True
    first_more_pgp_mime = True
    
    pre_snowden_enc = 0
    snowden_enc = 0
    post_snowden_enc = 0
    
    pre_snowden_signed = 0
    snowden_signed = 0
    post_snowden_signed = 0
    if send_only:
        all_mails_df = data_support.get_sent_mails(all_mails_df)
    mails_per_year = all_mails_df.groupby('header_year', group_keys=False)['header_year'].count()
    years = []

    while year <= max:
        years.append(year)
        if debug:
            print("###############", year, "###############")
        data = data_support.filter_for_year(all_mails_df, year)
        mails_count = data.shape[0]
        # avoid div zero error
        if mails_count == 0:
            mails_count = 1
            
        info = encryption_signatures(data, debug)
        info["Year"] = year
        info["Mails"] = "{:,}".format(mails_count)
        info["SignatureRatio"] = "{:,}".format(round(100.0*info["Signatures"] / mails_count,2))
        info["EncryptionRatio"] = "{:,}".format(round(100.0*info["Encryption"] / mails_count,2))
        info["PGPRatio"] = "{:,}".format(round(100.0*info["PGP"] / mails_count,2))
        info["SMIMERatio"] = "{:,}".format(round(100.0*info["SMIME"] / mails_count,2))
        res.append(info)
        
        if debug:
            print("Mails:", mails_count)
            print("SignatureRatio:", info["SignatureRatio"])
            print("EncryptionRatio:", info["EncryptionRatio"])
            print("PGPRatio:", info["PGPRatio"])
            print("SMIMERatio:", info["SMIMERatio"])
        
        
        if first_signed and info['Signatures'] > 0:
            first_signed = False
            commands['firstSignedYear'] = year
            type = ""
            if info['SMIME'] > 0:
                type = 'S/MIME'
                first_smime = False
            if info['PGP'] > 0:
                if type != "":
                    type = 'S/MIME and PGP'
                else:
                    type = 'PGP'
                first_pgp = False
            commands['firstSignedType'] = type
            
        if first_enc and info['Encryption'] > 0:
            first_enc = False
            commands['firstEncYear'] = year
            type = ""
            if info['SMIME'] > 0:
                type = 'S/MIME'
                first_smime = False
            if info['PGP'] > 0:
                if type != "":
                    type = 'S/MIME and PGP'
                else:
                    type = 'PGP'
                first_pgp = False
            commands['firstEncType'] = type
            
        if first_pgp and info['PGP'] > 0:
            commands['firstPgpYear'] = year
            first_pgp = False
        if first_smime and info['SMIME'] > 0:
            commands['firstSmimeYear'] = year
            first_smime = False
        if first_more_smime and info['SMIME'] > info['PGP']:
            commands['firstMoreSmimeYear'] = year
            commands['firstMoreSmimeMailsSmime'] = info['SMIME']
            commands['firstMoreSmimeMailsPgp'] = info['PGP']
            first_more_smime = False
        if first_more_pgp_mime and info['PGP MIME'] > info['PGP inline']:
            commands['firstMorePgpMimeYear'] = year
            commands['firstMorePgpMimeMailsMime'] = info['PGP MIME']
            commands['firstMorePgpMimeMailsInline'] = info['PGP inline']
            first_more_pgp_mime = False
        if year == 2012:
            pre_snowden_enc = info['Encryption']
            pre_snowden_signed = info['Signatures']
        if year == 2013:
            commands['snowdenCryptoMails'] = info['Total']
            snowden_enc = info['Encryption']
            snowden_signed = info['Signatures']
        if year == 2014:
            post_snowden_enc = info['Encryption']
            post_snowden_signed = info['Signatures']
            
            commands['snowdenPreIncreaseEnc'] = snowden_enc - pre_snowden_enc
            commands['snowdenPreIncreaseSign'] = snowden_signed - pre_snowden_signed
            commands['snowdenPostIncreaseEnc'] = post_snowden_enc - snowden_enc
            commands['snowdenPostIncreaseSign'] = post_snowden_signed - snowden_signed
        year += 1
        
    data = pandas.DataFrame(res)
    data = data.set_index('Year')
    title = 'Communication per year'
    name = 'communication_per_year'
    if send_only:
        title = 'Sent emails per year'
        name = 'sent_emails_per_year'
    
    #load_data.generate_plot_line(years, data[['Mails', 'SMIME encryption', 'SMIME signatures', 'PGP signatures', 'PGP encryption', "PGP MIME signatures", "PGP MIME encryption", "PGP inline signatures", "PGP inline encryption"]], title, name, log = True)
    load_data.generate_plot_line(years, data[['Mails', 'SMIME encryption', 'SMIME signatures', 'PGP signatures', 'PGP encryption']], title, name, log = True)
    
    filename = result_path + name
    csv_filename =  filename + ".csv"
    latex_table = data.to_latex(index=True)
    
    # Store data
    data.to_csv(csv_filename, sep = ',', quotechar = '\'', encoding='utf-8')
    load_data.write_string(latex_table, filename+"_table.tex")
    load_data.write_commands(commands, filename + ".tex")
    
    return data

def encryption_signatures(data, debug):
    """
        Output:
                * #signed mails, # encrypted mails,
                * PGP or s/mime related mails without mails only with key or certificate
                * Signed/encrypted with S/MIME
                * Signed/encrypted with PGP
                * Signed/encrypted with PGP Inline
                * Signed/encrypted with PGP MIME
    """
    signature_data_smime = data["smime_sig_fixed"].value_counts()
    encryption_data_smime = data["smime_encryption"].value_counts()
        
    signature_data_pgp = data["pgp_sig_attachment_fixed"].value_counts()
    signature_data_pgp_inline = data["pgp_signature_inline"].value_counts()
    encryption_data_pgp = data["pgp_enc_attachment_fixed"].value_counts()
    encryption_data_pgp_inline = data["pgp_encryption_inline"].value_counts()

    res = {}
    total = data.shape[0]
    # SMIME
    if "SMIMEEnc.enc" in encryption_data_smime:
       res["SMIME encryption"] = encryption_data_smime["SMIMEEnc.enc"]
    else:
        res["SMIME encryption"] = 0
    if True in signature_data_smime:
        res["SMIME signatures"] = signature_data_smime[True]
    else:
        res["SMIME signatures"] = 0
        
    
    # PGP
    if True in signature_data_pgp:
        res["PGP MIME signatures"] = signature_data_pgp[True]
    else:
        res["PGP MIME signatures"] = 0
    if True in encryption_data_pgp:
        res["PGP MIME encryption"] = encryption_data_pgp[True]
    else:
        res["PGP MIME encryption"] = 0
        
    if "PGPSig.un" in signature_data_pgp_inline:
        res["PGP inline signatures"] = total - signature_data_pgp_inline["PGPSig.un"]
    else:
        res["PGP inline signatures"] = total
    if "PGPEnc.un" in encryption_data_pgp_inline:
        res["PGP inline encryption"] = total - encryption_data_pgp_inline["PGPEnc.un"]
    else:
        res["PGP inline encryption"] = total
    
    # Add summary
    res["SMIME"] = res["SMIME signatures"] + res["SMIME encryption"]
    res["PGP"] = res["PGP MIME signatures"] + res["PGP MIME encryption"] + res["PGP inline signatures"] + res["PGP inline encryption"]
    res['PGP signatures'] = res["PGP MIME signatures"] + res["PGP inline signatures"]
    res['PGP encryption'] = res["PGP MIME encryption"] + res["PGP inline encryption"]
    res['PGP inline'] = res["PGP inline signatures"] + res["PGP inline encryption"]
    res['PGP MIME'] = res["PGP MIME signatures"] + res["PGP MIME encryption"]
    res['Signatures'] = res['PGP signatures'] + res["SMIME signatures"]
    res['Encryption'] = res['PGP encryption'] + res["SMIME encryption"]
    res['Total'] = res['Signatures'] + res['Encryption']

    if debug:
        for key, value in res.items():
            print(key + ":"+ str(value))
    
    return res
    
  
def main():
    debug = True
    
    overview = True
    perYear = True
    
    test = True

    overall = input("Run overall data? y/n\n")
    if overall == 'y':
        test = False
        print("> Load all data...")

    global result_path
    result_path = load_data.get_cache_path("", test)
    
    df_meta_all = load_data.load_all_mails(test)
    # maybe an overhead...
    df_meta_all = data_support.add_level_info_fix_crypto(df_meta_all)
    df_meta_all = load_data.drop_error_mails(df_meta_all)
    
    if overview:
        enc_signed_overview(df_meta_all, test, debug)
    if perYear:
        enc_signed_years(df_meta_all, test, debug, send_only = True)
        enc_signed_years(df_meta_all, test, debug, send_only = False)
    print("> Done")
    

if __name__ == "__main__":
    main()

