import pandas as pd
import modin.pandas as pandas
import numpy as np
from pathlib import Path
from collections import namedtuple
from statistics import mean
import datetime
import time
import numbers
import re
import os
from pprint import pprint
import helper, data_support, load_data, graph_support
import matplotlib.pyplot as plt
import scipy.stats as st
import seaborn as sns
import functools
import pickle
import json
from matplotlib.lines import Line2D

# Set to false to recreate numbers. Warning - long runtime. Does not recreate base csv
# Recreates graphs anyway
use_cache = True

#Use to skip loading after 1000 entries, will use different cache than non-testrun
testrun = True

#Set this to true to regenerate cache files required by analysis_pgp
recreate_analysis_pgp = False

base_path = load_data.base_path

tex_results = {}
tex_var_results = {}
table_results = ""

usergroup_mapping = {}
usergroup_mapping["sc.h"] = "Scientific"
usergroup_mapping["sa.h"] = "Staff"
usergroup_mapping["su.h"] = "Student"
usergroup_mapping["NX.h"] = "NX Internal"
usergroup_mapping["u.h"] = "Unknown Internal"
usergroup_mapping["unc.h"] = "Unclear"
usergroup_mapping["oos.h"] = "OutOfScope"
usergroup_mapping["err.h"] = "Errored"
usergroup_mapping["ext"] = "External"

#columns = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "header_date",
#                "client", "folder",
#                "mail_relay_count", "mail_virus_scans",
#                "message_id", "dkim",
#                "other_error"]

def cache_it(cache_name, is_df = True, reload = False, is_json = False):
    def real_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if is_cached(cache_name):
                if is_df:
                    result = load_from_cache(cache_name)
                else:
                    with open(load_data.get_cache_path(cache_name, testrun), 'r') as fp:
                        result = fp.read()
                        if is_json:
                            result = json.loads(result)
            else:
                result = function(*args, **kwargs)
                if is_df:
                    store_to_cache(cache_name, result)
                    if reload:
                        #Fix for some index problems on some DFs after grouping
                        result = load_from_cache(cache_name)
                else:
                    with open(load_data.get_cache_path(cache_name, testrun), 'w') as fp:
                        if is_json:
                            fp.write(json.dumps(result))
                        else:
                            fp.write(str(result))
            return result
        return wrapper
    return real_decorator

def store_to_cache(name, data):
    global testrun
    return load_data.store_to_cache(name, data, testrun)

def is_cached(name):
    global use_cache
    global testrun
    return load_data.is_cached(name, testrun, use_cache)

def load_from_cache(name):
    global testrun
    return load_data.load_from_cache(name, testrun)

def get_domain_and_mailcount(path='../data/'):
    domains = set()
    mailboxes = set()
    filenames = Path(path).rglob('*.csv')
    for f in filenames:
        domain = str(f).rsplit("/", 2)[1]
        filename = str(f).rsplit("/", 2)[2]
        if (len("ff5c96d866cad16908d4a7cc6484fa2c2e3e41e3.csv") == len(filename)):
            domains.add(domain)
            mailboxes.add(filename)

    filenames = Path(path).rglob('*.csv.gz')
    for f in filenames:
        domain = str(f).rsplit("/", 2)[1]
        filename = str(f).rsplit("/", 2)[2]
        domains.add(domain)
        mailboxes.add(filename)

    return len(mailboxes), len(domains)

def read_csv(path='../data/'):
    df = []

    files_read = 0

    filenames = Path(path).rglob('*.csv')
    for f in filenames:
        domain = str(f).rsplit("/", 2)[1]
        filename = str(f).rsplit("generate_plot/", 2)[2]
        if (len("ff5c96d866cad16908d4a7cc6484fa2c2e3e41e3.csv") == len(filename)):
            print(filename)
            read_df = pandas.read_csv(f, sep=',', quotechar='\'', low_memory=True)#, usecols=columns)
            read_df["domain"] = domain
            read_df["ID"] = filename
            read_df.fillna("empty", inplace=True)
            # print(read_df)
            df.append(read_df)
            files_read += 1
            print("CSV_read", files_read)

    filenames = Path(path).rglob('*.csv.gz')
    for f in filenames:
        domain = str(f).rsplit("/", 2)[1]
        filename = str(f).rsplit("/", 2)[2]
        try:
            read_df = pandas.read_csv(f, sep=',', quotechar='\'', compression='gzip', low_memory=True)#, usecols=columns)
            read_df["domain"] = domain
            read_df["ID"] = filename
            read_df.fillna("empty", inplace=True)
            # print(read_df)
            df.append(read_df)
            files_read += 1
            print(files_read)
        except:
            print("Skipped", f)

    data = pandas.concat(df, ignore_index=True)
    print("Read", files_read, "files for data analysis.")
    return data

def store_all_mails(data):
    filename = "all_mails_all_fields.csv"
    filename = "../../../data/" + filename

    data.to_csv(filename, sep = ',', quotechar = '\'', encoding='utf-8')

def tex_command(name, value):
    return "\\newcommand{\\"+str(name)+"}{"+str(value)+"}"

def tex_var(name, value):
    return "\\definevar{"+str(name)+"}{"+str(value)+"}"

def _print_df(df):
    pprint(dict(df))

def generate_plot(data_x, data_y, title, filename):
    plt.figure(figsize=(9, 6))
    plt.bar(x=data_x,
            height=data_y,
            color='midnightblue',
            log=True)
    plt.xticks(rotation=45)
    plt.title(title)
    plt.savefig('/home/mailanalysis/results/graphs/'+filename)

def generate_plot_no_log(data_x, data_y, title, filename):
    plt.figure(figsize=(9, 6))
    plt.bar(x=data_x,
            height=data_y,
            color='midnightblue',
            log=False)
    plt.xticks(rotation=45)
    plt.title(title)
    plt.savefig('/home/mailanalysis/results/graphs/'+filename)

def mails_per_year(data):
    cache_name = "meta_mails_per_year.csv"
    if is_cached(cache_name):
        mails_per_year = load_from_cache(cache_name)
    else:
        print("Mails Per Year:")
        mails_per_year = data.groupby("header_year", group_keys=False).size().reset_index(name='count')
        store_to_cache(cache_name, mails_per_year)
    generate_plot(mails_per_year["header_year"], mails_per_year["count"], "Mails per year", "MailPerYear.png")

def mails_per_year_received(data):
    cache_name = "meta_mails_per_year_received.csv"
    if is_cached(cache_name):
        mails_per_year = load_from_cache(cache_name)
    else:
        print("Mails Per Year:")
        mails_per_year = data.groupby("header_year", group_keys=False).size().reset_index(name='count')
        store_to_cache(cache_name, mails_per_year)
    generate_plot(mails_per_year["header_year"], mails_per_year["count"], "Mails received per year", "ReceivedPerYear.png")

def mails_per_year_sent(data_sent):
    cache_name = "meta_sent_mails_per_year.csv"
    if is_cached(cache_name):
        sent_per_year = load_from_cache(cache_name)
    else:
        print("Mails Per Year: (Sent)")
        sent_per_year = data_sent.groupby("header_year", group_keys=False).size().reset_index(name='count')
        store_to_cache(cache_name, sent_per_year)
    generate_plot(sent_per_year["header_year"], sent_per_year["count"], "Mails sent per year", "SentPerYear.png")

def filter_row_before(row):
    print(row)
    return row["min"] > row["header_date"]

@cache_it("grouped_useragent_general", reload = True)
def grouped_encryption_useragent_general(data):
    start = time.time()
    grouped_multiple_general = data.groupby(['user', 'client_merged']).agg(
        {'header_date': ['count', 'min', 'max']})
    grouped_multiple_general.columns = grouped_multiple_general.columns.get_level_values(1)
    end = time.time()
    print("Grouping useragent finished in", ((end - start) / 60), "min")
    return grouped_multiple_general

@cache_it("grouped_useragent_smime", reload = True)
def grouped_encryption_useragent_smime(data):
    start = time.time()
    grouped_multiple_smime = data.groupby(['user', 'client_merged', 'smime_sig_fixed', 'smime_encryption']).agg(
        {'header_date': ['count', 'min', 'max']})
    grouped_multiple_smime.columns = grouped_multiple_smime.columns.get_level_values(1)
    end = time.time()
    print("Grouping SMIME finished in", ((end - start) / 60), "min")
    return grouped_multiple_smime

@cache_it("grouped_useragent_pgp_attachment", reload = True)
def grouped_encryption_useragent_pgp_attachment(data):
    start = time.time()
    grouped_multiple_pgp_attach = data.groupby(
        ['user', 'client_merged', 'pgp_signature_attachment', 'pgp_encryption_attachment']).agg(
        {'header_date': ['count', 'min', 'max']})
    grouped_multiple_pgp_attach.columns = grouped_multiple_pgp_attach.columns.get_level_values(1)
    end = time.time()
    print("Grouping PGP attachment finished in", ((end - start) / 60), "min")
    return grouped_multiple_pgp_attach

@cache_it("grouped_useragent_pgp_inline", reload = True)
def grouped_encryption_useragent_pgp_inline(data):
    start = time.time()
    grouped_multiple_pgp_inline = data.groupby(
        ['user', 'client_merged', 'pgp_signature_inline', 'pgp_encryption_inline']).agg(
        {'header_date': ['count', 'min', 'max']})
    grouped_multiple_pgp_inline.columns = grouped_multiple_pgp_inline.columns.get_level_values(1)
    end = time.time()
    print("Grouping PGP inline finished in", ((end - start) / 60), "min")
    return grouped_multiple_pgp_inline

@cache_it("cache_grouped_useragent_pgp_merged", reload = True)
def grouped_merged_useragent_pgp(data):
    start = time.time()
    grouped_merged_pgp = data.groupby(
        ['user', 'client_merged', 'pgp_signature_merged', 'pgp_encryption_merged']).agg(
        {'header_date': ['count', 'min', 'max']})
    grouped_merged_pgp.columns = grouped_merged_pgp.columns.get_level_values(1)
    end = time.time()
    print("Grouping merged PGP finished in", ((end - start) / 60), "min")
    return grouped_merged_pgp

@cache_it("cache_smime_users_first_usage_date", reload=True)
def cf_smime_users_first_usage_date(grouped_multiple_smime):
    # Get all Mails for smime users after their first SMIME usage
    smime_users = grouped_multiple_smime.loc[((grouped_multiple_smime["smime_sig_fixed"] == True) | (grouped_multiple_smime["smime_encryption"] == "SMIMEEnc.enc"))]
    smime_users['min'] = pandas.to_datetime(smime_users['min'], format="%Y-%m-%d", errors='coerce')
    smime_users_first_usage_date = smime_users.groupby('user', as_index=False).agg(
        {'min': ['min']})
    smime_users_first_usage_date.columns = smime_users_first_usage_date.columns.get_level_values(0)
    return smime_users_first_usage_date

@cache_it("cache_smime_user_mails", reload=True)
def cf_smime_user_mails(data, smime_users_first_usage_date):
    return data.loc[((data["user"].isin(list(smime_users_first_usage_date["user"]))) & (data["sender"] != "empty"))]

#Non Cached function to be called by cached functions to avoid duplicate code.
def ncf_smime_usage_after_first_use(smime_usage):
    smime_usage_after_first_use = smime_usage.groupby(
        ['user', 'smime_sig_fixed']).agg(
        {'header_date': ['count', 'min', 'max']})
    smime_usage_after_first_use.columns = smime_usage_after_first_use.columns.get_level_values(1)
    return smime_usage_after_first_use

@cache_it("cache_smime_usage_after_first_use", reload=True)
def cf_smime_usage_after_first_use(smime_user_mails, smime_users_first_usage_date):
    smime_usage = pandas.merge(smime_user_mails, smime_users_first_usage_date, on=['user'], how='left')
    smime_usage['min'] = pandas.to_datetime(smime_usage['min'], format="%Y-%m-%d", errors='coerce')
    smime_usage['header_date'] = pandas.to_datetime(smime_usage['header_date'], format="%Y-%m-%d", errors='coerce')
    smime_usage["after_usage"] = ((smime_usage['header_date'] > smime_usage['min']) | (
            (smime_usage['smime_sig_fixed'] == True)))
    smime_usage = smime_usage.loc[smime_usage["after_usage"] == True]
    return ncf_smime_usage_after_first_use(smime_usage)

@cache_it("cache_smime_sent_usage_after_first_use", reload=True)
def cf_smime_sent_usage_after_first_use(smime_usage, smime_users_first_usage_date):
    return ncf_smime_usage_after_first_use(smime_usage)

@cache_it("cache_smime_enc_sent_usage_after_first_use", reload=True)
def cf_smime_enc_sent_usage_after_first_use(smime_usage):
    smime_usage_after_first_use = smime_usage.groupby(
        ['user', 'smime_encryption']).agg(
        {'header_date': ['count', 'min', 'max']})
    smime_usage_after_first_use.columns = smime_usage_after_first_use.columns.get_level_values(1)
    return smime_usage_after_first_use

def ncf_smime_users_percentages_after_first_use(smime_usage_after_first_use):
    smime_users_sums = smime_usage_after_first_use.groupby(['user', 'smime_sig_fixed']).agg({'count': 'sum'})
    smime_users_percentages = smime_users_sums.groupby(level=0).apply(lambda x:
                                                                      100 * x / float(x.sum())).reset_index()
    return smime_users_percentages

@cache_it("cache_smime_users_percentages_after_first_use", reload=True)
def cf_smime_users_percentages_after_first_use(smime_usage_after_first_use):
    return ncf_smime_users_percentages_after_first_use(smime_usage_after_first_use)

@cache_it("cache_smime_sent_users_percentages_after_first_use", reload=True)
def cf_smime_sent_users_percentages_after_first_use(smime_sent_usage_after_first_use):
    return ncf_smime_users_percentages_after_first_use(smime_sent_usage_after_first_use)

@cache_it("cache_smime_enc_sent_users_percentages_after_first_use", reload=True)
def cf_smime_enc_sent_users_percentages_after_first_use(smime_sent_usage_after_first_use):
    return ncf_smime_users_percentages_after_first_use(smime_sent_usage_after_first_use)

def ncf_sent_users_percentages_after_first_use_n(df):
    sent_users_sums = df.groupby(['user', 'smime_sig_fixed']).agg(
        {'count': 'sum'})
    sent_users_percentages = sent_users_sums.groupby(level=0).apply(lambda x:
                                                                                              100 * x / float(
                                                                                                  x.sum())).reset_index()
    return sent_users_percentages

def ncf_sent_enc_users_percentages_after_first_use_n(df):
    sent_users_sums = df.groupby(['user', 'smime_encryption']).agg(
        {'count': 'sum'})
    sent_users_percentages = sent_users_sums.groupby(level=0).apply(lambda x:
                                                                                              100 * x / float(
                                                                                                  x.sum())).reset_index()
    return sent_users_percentages

def ncf_add_users_with_zero_encryption(df, list_of_users):
    #Fill users that have signed but not encrypted with 0 and the other way.
    print(df)
    extension_data = []
    i = 0
    for row in list_of_users.itertuples():
        #print("a", row)
        if not df["user"].str.contains(row.user).any():
            #print("is in", row.user)
            #SMIME
            extension = {}
            extension["user"] = row.user
            extension["count"] = 0
            try:
                if row.smime_sig_fixed:
                    #print("ssig - with value", row.smime_sig_fixed)
                    extension_data.append(extension)
            except:
                pass
            try:
                if row.smime_encryption == "SMIMEEnc.enc":
                    #print("senc - with value", row.smime_encryption)
                    extension_data.append(extension)
            except:
                pass

            #PGP
            try:
                if row.pgp_signature_merged:
                    #("psig - with value", row.pgp_signature_merged)
                    extension_data.append(extension)
            except:
                pass
            try:
                if row.pgp_encryption_merged:
                    #print("penc - with value", row.pgp_encryption_merged)
                    extension_data.append(extension)
            except:
                pass
        else:
            i += 1
    print(f'[Data_Enc] Added {len(extension_data)} entries. {i} were already in the DF.')
    df2 = pandas.DataFrame(extension_data, columns=['user', 'count'])
    df = df.append(df2)
    #print(df)
    return df


#### PGP Helper Functions
@cache_it("cache_pgp_users_first_usage_date", reload=True)
def cf_pgp_users_first_usage_date(grouped_multiple_pgp):
    # Get all Mails for PGP users after their first PGP usage
    pgp_users = grouped_multiple_pgp.loc[((grouped_multiple_pgp["pgp_signature_merged"] == True) | (grouped_multiple_pgp["pgp_encryption_merged"]))]
    pgp_users['min'] = pandas.to_datetime(pgp_users['min'], format="%Y-%m-%d", errors='coerce')
    pgp_users_first_usage_date = pgp_users.groupby('user', as_index=False).agg(
        {'min': ['min']})
    pgp_users_first_usage_date.columns = pgp_users_first_usage_date.columns.get_level_values(0)
    return pgp_users_first_usage_date

@cache_it("cache_pgp_user_mails", reload=True)
def cf_pgp_user_mails(data, pgp_users_first_usage_date):
    return data.loc[((data["user"].isin(list(pgp_users_first_usage_date["user"]))) & (data["sender"] != "empty"))]

#Non Cached function to be called by cached functions to avoid duplicate code.
def ncf_pgp_usage_after_first_use(pgp_usage):
    pgp_usage_after_first_use = pgp_usage.groupby(
        ['user', 'pgp_signature_merged']).agg(
        {'header_date': ['count', 'min', 'max']})
    pgp_usage_after_first_use.columns = pgp_usage_after_first_use.columns.get_level_values(1)
    return pgp_usage_after_first_use

@cache_it("cache_pgp_sent_usage_after_first_use", reload=True)
def cf_pgp_sent_usage_after_first_use(pgp_usage, pgp_users_first_usage_date):
    return ncf_pgp_usage_after_first_use(pgp_usage)

@cache_it("cache_pgp_enc_sent_usage_after_first_use", reload=True)
def cf_pgp_enc_sent_usage_after_first_use(pgp_usage):
    pgp_usage_after_first_use = pgp_usage.groupby(
        ['user', 'pgp_encryption_merged']).agg(
        {'header_date': ['count', 'min', 'max']})
    pgp_usage_after_first_use.columns = pgp_usage_after_first_use.columns.get_level_values(1)
    return pgp_usage_after_first_use

def ncf_pgp_users_percentages_after_first_use(pgp_usage_after_first_use):
    pgp_users_sums = pgp_usage_after_first_use.groupby(['user', 'pgp_signature_merged']).agg({'count': 'sum'})
    pgp_users_percentages = pgp_users_sums.groupby(level=0).apply(lambda x:
                                                                      100 * x / float(x.sum())).reset_index()
    return pgp_users_percentages

@cache_it("cache_pgp_sent_users_percentages_after_first_use", reload=True)
def cf_pgp_sent_users_percentages_after_first_use(pgp_sent_usage_after_first_use):
    return ncf_pgp_users_percentages_after_first_use(pgp_sent_usage_after_first_use)

def ncf_pgp_sent_users_percentages_after_first_use_n(df):
    sent_users_sums = df.groupby(['user', 'pgp_signature_merged']).agg(
        {'count': 'sum'})
    sent_users_percentages = sent_users_sums.groupby(level=0).apply(lambda x:
                                                                                              100 * x / float(
                                                                                                  x.sum())).reset_index()
    return sent_users_percentages

def ncf_sent_pgp_enc_users_percentages_after_first_use_n(df):
    sent_users_sums = df.groupby(['user', 'pgp_encryption_merged']).agg(
        {'count': 'sum'})
    sent_users_percentages = sent_users_sums.groupby(level=0).apply(lambda x:
                                                                                              100 * x / float(
                                                                                                  x.sum())).reset_index()
    return sent_users_percentages

def graph_pgp(data):
    grouped_merged_pgp = grouped_merged_useragent_pgp(data)

    pgp_users_first_usage_date = cf_pgp_users_first_usage_date(grouped_merged_pgp)
    print("[PGP-DEBUG1]", pgp_users_first_usage_date)
    pgp_user_mails = cf_pgp_user_mails(data, pgp_users_first_usage_date)
    print("[PGP-DEBUG2]", pgp_user_mails)

    # Do it for sent mails - Since we switched to per user, we're using sent mails only anyway
    start = time.time()
    pgp_user_sent_mails = pgp_user_mails#data_support.get_sent_mails(pgp_user_mails)

    #pgp_user_sent_mails = pgp_user_mails[
    #                                                      pgp_user_mails['folder'].str.contains('Postbox.sent') |
    #                                                      pgp_user_mails['folder'].str.contains(
    #                                                          'Postbox.sent_sub')]

    print("[PGP-DEBUG3]", pgp_user_sent_mails)
    end = time.time()
    print("Reducing PGP to sent mails finished", ((end - start) / 60), "min")

    pgp_sent_usage = pandas.merge(pgp_user_sent_mails, pgp_users_first_usage_date, on=['user'], how='left')
    pgp_sent_usage['min'] = pandas.to_datetime(pgp_sent_usage['min'], format="%Y-%m-%d", errors='coerce')
    pgp_sent_usage['header_date'] = pandas.to_datetime(pgp_sent_usage['header_date'], format="%Y-%m-%d",
                                                       errors='coerce')
    pgp_sent_usage["after_usage"] = ((pgp_sent_usage['header_date'] > pgp_sent_usage['min']) | (
            #(pgp_sent_usage['header_date'] >= pgp_sent_usage['min']) & (
        (pgp_sent_usage['pgp_signature_merged'] == True)))
    print("[PGP-DEBUG4]", pgp_sent_usage)
    pgp_sent_usage = pgp_sent_usage.loc[pgp_sent_usage["after_usage"] == True]
    print("[PGP-DEBUG5]", pgp_sent_usage)

    pgp_sent_usage_after_first_use = cf_pgp_sent_usage_after_first_use(pgp_sent_usage, pgp_users_first_usage_date)
    print("[PGP-DEBUG6]", pgp_sent_usage_after_first_use)
    pgp_sent_users_percentages = cf_pgp_sent_users_percentages_after_first_use(pgp_sent_usage_after_first_use)
    print("[PGP-DEBUG7]", pgp_sent_users_percentages)

    data_all = pgp_sent_users_percentages.loc[pgp_sent_users_percentages["pgp_signature_merged"] == True]
    generate_pgp_usage_after_plot(data_all,
                                  "PGP Signature usage after first use (Sent only)", "_sent")

    # Do it for sent mails by used clients
    clients_after_first_use = pgp_sent_usage.groupby(
        'user').agg(
        {'client_merged': ['nunique']}).reset_index()
    single_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 1]
    dual_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 2]
    triple_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 3]
    more_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] > 3]
    single_client_res = pgp_sent_usage_after_first_use.loc[
        pgp_sent_usage_after_first_use["user"].isin(list(single_client))]
    dual_client_res = pgp_sent_usage_after_first_use.loc[
        pgp_sent_usage_after_first_use["user"].isin(list(dual_client))]
    triple_client_res = pgp_sent_usage_after_first_use.loc[
        pgp_sent_usage_after_first_use["user"].isin(list(triple_client))]
    more_client_res = pgp_sent_usage_after_first_use.loc[
        pgp_sent_usage_after_first_use["user"].isin(list(more_client))]

    pgp_sent_users_percentages_single = ncf_pgp_sent_users_percentages_after_first_use_n(single_client_res)
    pgp_sent_users_percentages_dual = ncf_pgp_sent_users_percentages_after_first_use_n(dual_client_res)
    pgp_sent_users_percentages_triple = ncf_pgp_sent_users_percentages_after_first_use_n(triple_client_res)
    pgp_sent_users_percentages_more = ncf_pgp_sent_users_percentages_after_first_use_n(more_client_res)

    # Enc sent - Based on first sent pgp mail, no matter if signed/encrypted
    # single_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 1]
    # dual_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 2]
    # triple_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 3]
    # more_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] > 3]

    pgp_enc_sent_usage_after_first_use = cf_pgp_enc_sent_usage_after_first_use(pgp_sent_usage)

    single_client_res_enc = pgp_enc_sent_usage_after_first_use.loc[
        pgp_enc_sent_usage_after_first_use["user"].isin(list(single_client))]
    dual_client_res_enc = pgp_enc_sent_usage_after_first_use.loc[
        pgp_enc_sent_usage_after_first_use["user"].isin(list(dual_client))]
    triple_client_res_enc = pgp_enc_sent_usage_after_first_use.loc[
        pgp_enc_sent_usage_after_first_use["user"].isin(list(triple_client))]
    more_client_res_enc = pgp_enc_sent_usage_after_first_use.loc[
        pgp_enc_sent_usage_after_first_use["user"].isin(list(more_client))]

    pgp_enc_sent_users_percentages_single = ncf_sent_pgp_enc_users_percentages_after_first_use_n(single_client_res_enc)
    pgp_enc_sent_users_percentages_dual = ncf_sent_pgp_enc_users_percentages_after_first_use_n(dual_client_res_enc)
    pgp_enc_sent_users_percentages_triple = ncf_sent_pgp_enc_users_percentages_after_first_use_n(triple_client_res_enc)
    pgp_enc_sent_users_percentages_more = ncf_sent_pgp_enc_users_percentages_after_first_use_n(more_client_res_enc)


    try:
        data1 = pgp_sent_users_percentages_single.loc[pgp_sent_users_percentages_single['pgp_signature_merged'] == True]
        data1_f = ncf_add_users_with_zero_encryption(data1, single_client_res_enc)
        generate_pgp_usage_after_plot(data1_f,
                                  "[1 Client] PGP Signature usage after first use (Sent only)", "_sent_1_client")

    except:
        data1 = {"count": [200, 250]}
        data1_f = {"count": [200, 250]}
        print("[Error] Skipped one - Empty result?")

    try:
        data2 = pgp_sent_users_percentages_dual.loc[pgp_sent_users_percentages_dual['pgp_signature_merged'] == True]
        data2_f = ncf_add_users_with_zero_encryption(data2, dual_client_res_enc)
        generate_pgp_usage_after_plot(data2_f,
                                      "[2 Clients] PGP Signature usage after first use (Sent only)", "_sent_2_client")
    except:
        data2 = {"count": [200, 250]}
        data2_f = {"count": [200, 250]}
        print("[Error] Skipped dual - Empty result?")

    try:
        data3 = pgp_sent_users_percentages_triple.loc[pgp_sent_users_percentages_triple['pgp_signature_merged'] == True]
        data3_f = ncf_add_users_with_zero_encryption(data3, triple_client_res_enc)
        generate_pgp_usage_after_plot(
            data3_f,
            "[3 Clients] PGP Signature usage after first use (Sent only)", "_sent_3_client")
    except:
        data3 = {"count": [200, 250]}
        data3_f = {"count": [200, 250]}
        print("[Error] Skipped three - Empty result?")

    try:
        data4 = pgp_sent_users_percentages_more.loc[pgp_sent_users_percentages_more['pgp_signature_merged'] == True]
        data4_f = ncf_add_users_with_zero_encryption(data4, more_client_res_enc)
        generate_pgp_usage_after_plot(
            data4_f,
            "[4+ Clients] PGP Signature usage after first use (Sent only)", "_sent_4+_client")
    except:
        data4 = {"count": [200, 250]}
        data4_f = {"count": [200, 250]}
        print("[Error] Skipped more - Empty result?")

    try:
        data1_enc = pgp_enc_sent_users_percentages_single.loc[
            pgp_enc_sent_users_percentages_single["pgp_encryption_merged"] == True]
        data1_enc_f = ncf_add_users_with_zero_encryption(data1_enc, single_client_res)
        generate_pgp_usage_after_plot(data1_enc_f,
                                        "[1 Client] PGP Signature usage after first use (Sent only)",
                                        "_sent_enc_1_client")
    except:
        data1_enc = {"count": [200, 250]}
        data1_enc_f = {"count": [200, 250]}
        print("[Error] Skipped dual - Empty result?")

    try:
        data2_enc = pgp_enc_sent_users_percentages_dual.loc[
            pgp_enc_sent_users_percentages_dual["pgp_encryption_merged"] == True]
        data2_enc_f = ncf_add_users_with_zero_encryption(data2_enc, dual_client_res)
        generate_pgp_usage_after_plot(data2_enc_f,
                                        "[2 Clients] PGP Signature usage after first use (Sent only)",
                                        "_sent_enc_2_client")
    except:
        data2_enc = {"count": [200, 250]}
        data2_enc_f = {"count": [200, 250]}
        print("[Error] Skipped dual - Empty result?")

    try:
        data3_enc = pgp_enc_sent_users_percentages_triple.loc[
            pgp_enc_sent_users_percentages_triple["pgp_encryption_merged"] == True]
        data3_enc_f = ncf_add_users_with_zero_encryption(data3_enc, triple_client_res)
        generate_pgp_usage_after_plot(
            data3_enc_f,
            "[3 Clients] PGP Signature usage after first use (Sent only)", "_sent_enc_3_client")
    except:
        data3_enc = {"count": [200, 250]}
        data3_enc_f = {"count": [200, 250]}
        print("[Error] Skipped three - Empty result?")

    try:
        data4_enc = pgp_enc_sent_users_percentages_more.loc[
            pgp_enc_sent_users_percentages_more["pgp_encryption_merged"] == True]
        data4_enc_f = ncf_add_users_with_zero_encryption(data4_enc, more_client_res)
        generate_pgp_usage_after_plot(
            data4_enc_f,
            "[4+ Clients] PGP Signature usage after first use (Sent only)", "_sent_enc_4+_client")
    except:
        data4_enc = {"count": [200, 250]}
        data4_enc_f = {"count": [200, 250]}
        print("[Error] Skipped more - Empty result?")

    pgp_sign_users = len(list(data1["count"])) + len(list(data2["count"])) + len(list(data3["count"])) + len(list(data4["count"]))
    pgp_enc_users = len(list(data1_enc["count"])) + len(list(data2_enc["count"])) + len(list(data3_enc["count"])) + len(list(data4_enc["count"]))

    #print(data_all["count"])
    data_merged_raw = {
        "1": list(data1_f["count"]),
        "1_enc": list(data1_enc_f["count"]),
        "2": list(data2_f["count"]),
        "2_enc": list(data2_enc_f["count"]),
        "3": list(data3_f["count"]),
        "3_enc": list(data3_enc_f["count"]),
        "4+": list(data4_f["count"]),
        "4+_enc": list(data4_enc_f["count"]),
    }
    data_merged = pandas.concat([pandas.Series(v, name=k) for k, v in data_merged_raw.items()], axis=1)
    data_merged.to_csv(load_data.get_cache_path("cache_temp_data_merged_pgp", testrun), sep = ',', quotechar = '\'')
    # data_for_graph = {"all": data_all, "1":data1, "2":data2, "3":data3, "4+":data4}
    # print(pandas.concat(data_all, data1, data2, data3, data4))
    #print(data_merged)
    generate_pgp_usage_after_boxplot(data_merged, "PGP Usage after first PGP use")
    return data_merged_raw

def graph_smime(data):
    grouped_multiple_smime = grouped_encryption_useragent_smime(data)
    smime_users_first_usage_date = cf_smime_users_first_usage_date(grouped_multiple_smime)
    smime_user_mails = cf_smime_user_mails(data, smime_users_first_usage_date)
    smime_usage_after_first_use = cf_smime_usage_after_first_use(smime_user_mails, smime_users_first_usage_date)
    smime_users_percentages = cf_smime_users_percentages_after_first_use(smime_usage_after_first_use)
    generate_smime_usage_after_plot(smime_users_percentages.loc[smime_users_percentages["smime_sig_fixed"] == True],
                                    "SMIME Signature usage after first use")

    # Do it for sent mails - Since we switched to per user, we're using sent mails only anyway
    start = time.time()
    smime_user_sent_mails = smime_user_mails #data_support.get_sent_mails(smime_user_mails)
    #smime_user_sent_mails = smime_user_mails[
    #    smime_user_mails['folder'].str.contains('Postbox.sent') | smime_user_mails['folder'].str.contains(
    #        'Postbox.sent_sub')]
    end = time.time()
    print(f"Reducing SMIME to sent mails finished in {((end - start) / 60):.2f} min, amount of mails before left merge: {smime_user_sent_mails.shape[0]}")

    smime_sent_usage = pandas.merge(smime_user_sent_mails, smime_users_first_usage_date, on=['user'], how='left')
    smime_sent_usage['min'] = pandas.to_datetime(smime_sent_usage['min'], format="%Y-%m-%d", errors='coerce')
    smime_sent_usage['header_date'] = pandas.to_datetime(smime_sent_usage['header_date'], format="%Y-%m-%d",
                                                         errors='coerce')
    smime_sent_usage["after_usage"] = ((smime_sent_usage['header_date'] > smime_sent_usage['min']) | (
            #(smime_sent_usage['header_date'] >= smime_sent_usage['min']) & (
        (smime_sent_usage['smime_sig_fixed'] == True)))
    print(f"[Debug] Length of smime_sent_usage before drop: {smime_sent_usage.shape[0]}, unique users: {smime_sent_usage['user'].unique().shape[0]}")
    smime_sent_usage = smime_sent_usage.loc[smime_sent_usage["after_usage"] == True]
    print(f"[Debug] Length of smime_sent_usage after drop: {smime_sent_usage.shape[0]}, unique users: {smime_sent_usage['user'].unique().shape[0]}")

    smime_sent_usage_after_first_use = cf_smime_sent_usage_after_first_use(smime_sent_usage,
                                                                           smime_users_first_usage_date)
    smime_sent_users_percentages = cf_smime_sent_users_percentages_after_first_use(smime_sent_usage_after_first_use)

    data_all = smime_sent_users_percentages.loc[smime_sent_users_percentages["smime_sig_fixed"] == True]
    generate_smime_usage_after_plot(data_all,
                                    "SMIME Signature usage after first use (Sent only)", "_sent")

    # Do it for sent mails by used clients
    clients_after_first_use = smime_sent_usage.groupby(
        'user').agg(
        {'client_merged': ['nunique']}).reset_index()
    print(clients_after_first_use)

    # Signatures sent
    single_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 1]
    dual_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 2]
    triple_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 3]
    more_client = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] > 3]
    single_client_res = smime_sent_usage_after_first_use.loc[
        smime_sent_usage_after_first_use["user"].isin(list(single_client))]
    dual_client_res = smime_sent_usage_after_first_use.loc[
        smime_sent_usage_after_first_use["user"].isin(list(dual_client))]
    triple_client_res = smime_sent_usage_after_first_use.loc[
        smime_sent_usage_after_first_use["user"].isin(list(triple_client))]
    more_client_res = smime_sent_usage_after_first_use.loc[
        smime_sent_usage_after_first_use["user"].isin(list(more_client))]

    clients_used_after_first_use = smime_sent_usage.groupby(
        ['user', "smime_sig_fixed"]).agg(
        {'client_merged': ['nunique']}).reset_index()

    dual_client_clients_used_for_signature = clients_used_after_first_use.loc[
        clients_used_after_first_use["user"].isin(list(dual_client))]

    triple_client_clients_used_for_signature = clients_used_after_first_use.loc[
        clients_used_after_first_use["user"].isin(list(triple_client))]

    more_client_clients_used_for_signature = clients_used_after_first_use.loc[
        clients_used_after_first_use["user"].isin(list(more_client))]

    dual_client_clients_used_for_signature.columns = ["user", "ssig", "ccount"]
    triple_client_clients_used_for_signature.columns = ["user", "ssig", "ccount"]
    more_client_clients_used_for_signature.columns = ["user", "ssig", "ccount"]
    print("dual_client_clients_used_for_signature", dual_client_clients_used_for_signature)
    print("triple_client_clients_used_for_signature", triple_client_clients_used_for_signature)
    print("more_client_clients_used_for_signature", more_client_clients_used_for_signature)

    dual_client_single_sign = dual_client_clients_used_for_signature.loc[((dual_client_clients_used_for_signature["ccount"] == 1) & (dual_client_clients_used_for_signature["ssig"]))]
    dual_client_single_sign_count = dual_client_single_sign.shape[0]
    dual_client_single_sign_percent = f'{(dual_client_single_sign_count / len(list(dual_client)) * 100):.2f}'

    if testrun:
        #Add fake data, testrun doesn't have enough results
        print("WARNING: Not enough data for triple client, using dual client data.")
        triple_client_single_sign = dual_client_single_sign
        triple_client = dual_client
    else:
        triple_client_single_sign = triple_client_clients_used_for_signature.loc[(
                (triple_client_clients_used_for_signature["ccount"] == 1) & (
            triple_client_clients_used_for_signature["ssig"]))]
    triple_client_single_sign_count = triple_client_single_sign.shape[0]
    triple_client_single_sign_percent = f'{(triple_client_single_sign_count / len(list(triple_client)) * 100):.2f}'

    try:
        more_client_single_sign = more_client_clients_used_for_signature.loc[
            ((more_client_clients_used_for_signature["ccount"] == 1) & (more_client_clients_used_for_signature["ssig"]))]
        more_client_single_sign_count = more_client_single_sign.shape[0]
        more_client_single_sign_percent = f'{(more_client_single_sign_count / len(list(more_client)) * 100):.2f}'
    except:
        more_client_single_sign_count = 0
        more_client_single_sign_percent = '0.00'

    add_tex_result("DualClientSingleSignCount", dual_client_single_sign_count, "MultiClient")
    add_tex_result("DualClientSingleSignPercent", dual_client_single_sign_percent, "MultiClient")
    add_tex_result("TripleClientSingleSignCount", triple_client_single_sign_count, "MultiClient")
    add_tex_result("TripleClientSingleSignPercent", triple_client_single_sign_percent, "MultiClient")
    add_tex_result("MoreClientSingleSignCount", more_client_single_sign_count, "MultiClient")
    add_tex_result("MoreClientSingleSignPercent", more_client_single_sign_percent, "MultiClient")

    multi_client_single_sign_count = dual_client_single_sign_count + triple_client_single_sign_count + more_client_single_sign_count
    total_multi_client_count = len(list(dual_client))+len(list(triple_client))+len(list(more_client))

    add_tex_result("MultiClientSingleSignCount", multi_client_single_sign_count, "MultiClient")
    add_tex_result("MultiClientSingleSignPercent", f'{(multi_client_single_sign_count / total_multi_client_count * 100):.2f}', "MultiClient")
    add_tex_result("MultiClientMultiSignCount", total_multi_client_count - multi_client_single_sign_count, "MultiClient")
    add_tex_result("MultiClientMultiSignPercent", f'{100-(multi_client_single_sign_count / total_multi_client_count * 100):.2f}', "MultiClient")

    smime_sent_users_percentages_single = ncf_sent_users_percentages_after_first_use_n(single_client_res)
    smime_sent_users_percentages_dual = ncf_sent_users_percentages_after_first_use_n(dual_client_res)
    smime_sent_users_percentages_triple = ncf_sent_users_percentages_after_first_use_n(triple_client_res)
    smime_sent_users_percentages_more = ncf_sent_users_percentages_after_first_use_n(more_client_res)

    print(dual_client_single_sign_count, dual_client_single_sign_percent, dual_client_single_sign)
    print(triple_client_single_sign_count, triple_client_single_sign_percent, triple_client_single_sign)

    # Enc sent - Based on first sent smime mail, no matter if signed/encrypted
    # single_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 1]
    # dual_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 2]
    # triple_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] == 3]
    # more_client_enc = clients_after_first_use["user"].loc[clients_after_first_use["client_merged"]["nunique"] > 3]

    smime_enc_sent_usage_after_first_use = cf_smime_enc_sent_usage_after_first_use(smime_sent_usage)

    single_client_res_enc = smime_enc_sent_usage_after_first_use.loc[
        smime_enc_sent_usage_after_first_use["user"].isin(list(single_client))]
    dual_client_res_enc = smime_enc_sent_usage_after_first_use.loc[
        smime_enc_sent_usage_after_first_use["user"].isin(list(dual_client))]
    triple_client_res_enc = smime_enc_sent_usage_after_first_use.loc[
        smime_enc_sent_usage_after_first_use["user"].isin(list(triple_client))]
    more_client_res_enc = smime_enc_sent_usage_after_first_use.loc[
        smime_enc_sent_usage_after_first_use["user"].isin(list(more_client))]

    smime_enc_sent_users_percentages_single = ncf_sent_enc_users_percentages_after_first_use_n(single_client_res_enc)
    smime_enc_sent_users_percentages_dual = ncf_sent_enc_users_percentages_after_first_use_n(dual_client_res_enc)
    smime_enc_sent_users_percentages_triple = ncf_sent_enc_users_percentages_after_first_use_n(triple_client_res_enc)
    smime_enc_sent_users_percentages_more = ncf_sent_enc_users_percentages_after_first_use_n(more_client_res_enc)

    try:
        data1 = smime_sent_users_percentages_single.loc[smime_sent_users_percentages_single["smime_sig_fixed"] == True]
        data1_f = ncf_add_users_with_zero_encryption(data1, single_client_res_enc)
        generate_smime_usage_after_plot(data1_f,
                                        "[1 Client] SMIME Signature usage after first use (Sent only)", "_sent_1_client")
    except:
        data1 = {"count": [200, 250]}
        data1_f = {"count": [200, 250]}
        print("[Error] Skipped one - Empty result?")

    try:
        data2 = smime_sent_users_percentages_dual.loc[smime_sent_users_percentages_dual["smime_sig_fixed"] == True]
        data2_f = ncf_add_users_with_zero_encryption(data2, dual_client_res_enc)
        generate_smime_usage_after_plot(data2_f,
                                        "[2 Clients] SMIME Signature usage after first use (Sent only)", "_sent_2_client")
    except:
        data2 = {"count": [200, 250]}
        data2_f = {"count": [200, 250]}
        print("[Error] Skipped two - Empty result?")

    try:
        data3 = smime_sent_users_percentages_triple.loc[smime_sent_users_percentages_triple["smime_sig_fixed"] == True]
        data3_f = ncf_add_users_with_zero_encryption(data3, triple_client_res_enc)
        print("Data3", data3_f)
        generate_smime_usage_after_plot(
            data3_f,
            "[3 Clients] SMIME Signature usage after first use (Sent only)", "_sent_3_client")
    except:
        data3 = {"count": [200, 250]}
        data3_f = {"count": [200, 250]}
        print("[Error] Skipped three - Empty result?")

    try:
        data4 = smime_sent_users_percentages_more.loc[smime_sent_users_percentages_more["smime_sig_fixed"] == True]
        data4_f = ncf_add_users_with_zero_encryption(data4, more_client_res_enc)
        generate_smime_usage_after_plot(
            data4_f,
            "[4+ Clients] SMIME Signature usage after first use (Sent only)", "_sent_4+_client")
    except:
        data4 = {"count": [200, 250]}
        data4_f = {"count": [200, 250]}
        print("[Error] Skipped more - Empty result?")

    try:
        data1_enc = smime_enc_sent_users_percentages_single.loc[
            smime_enc_sent_users_percentages_single["smime_encryption"] == "SMIMEEnc.enc"]
        data1_enc_f = ncf_add_users_with_zero_encryption(data1_enc, single_client_res)
        generate_smime_usage_after_plot(data1_enc_f,
                                        "[1 Client] SMIME Signature usage after first use (Sent only)",
                                        "_sent_enc_1_client")
    except:
        data1_enc = {"count": [200, 250]}
        data1_enc_f = {"count": [200, 250]}
        print("[Error] Skipped dual - Empty result?")

    try:
        data2_enc = smime_enc_sent_users_percentages_dual.loc[
            smime_enc_sent_users_percentages_dual["smime_encryption"] == "SMIMEEnc.enc"]
        data2_enc_f = ncf_add_users_with_zero_encryption(data2_enc, dual_client_res)
        generate_smime_usage_after_plot(data2_enc_f,
                                        "[2 Clients] SMIME Signature usage after first use (Sent only)",
                                        "_sent_enc_2_client")
    except:
        data2_enc = {"count": [200, 250]}
        data2_enc_f = {"count": [200, 250]}
        print("[Error] Skipped dual - Empty result?")

    try:
        data3_enc = smime_enc_sent_users_percentages_triple.loc[
            smime_enc_sent_users_percentages_triple["smime_encryption"] == "SMIMEEnc.enc"]
        data3_enc_f = ncf_add_users_with_zero_encryption(data3_enc, triple_client_res)
        generate_smime_usage_after_plot(
            data3_enc_f,
            "[3 Clients] SMIME Signature usage after first use (Sent only)", "_sent_enc_3_client")
    except:
        data3_enc = {"count": [200, 250]}
        data3_enc_f = {"count": [200, 250]}
        print("[Error] Skipped three - Empty result?")

    try:
        data4_enc = smime_enc_sent_users_percentages_more.loc[
            smime_enc_sent_users_percentages_more["smime_encryption"] == "SMIMEEnc.enc"]
        data4_enc_f = ncf_add_users_with_zero_encryption(data4_enc, more_client_res)
        generate_smime_usage_after_plot(
            data4_enc_f,
            "[4+ Clients] SMIME Signature usage after first use (Sent only)", "_sent_enc_4+_client")
    except:
        data4_enc = {"count": [200, 250]}
        data4_enc_f = {"count": [200, 250]}
        print("[Error] Skipped more - Empty result?")


    smime_sign_users = len(list(data1["count"])) + len(list(data2["count"])) + len(list(data3["count"])) + len(list(data4["count"]))
    smime_enc_users = len(list(data1_enc["count"])) + len(list(data2_enc["count"])) + len(list(data3_enc["count"])) + len(list(data4_enc["count"]))

    #print(data_all["count"])

    data_merged_raw = {
        "1": list(data1_f["count"]),
        "1_enc": list(data1_enc_f["count"]),
        "2": list(data2_f["count"]),
        "2_enc": list(data2_enc_f["count"]),
        "3": list(data3_f["count"]),
        "3_enc": list(data3_enc_f["count"]),
        "4+": list(data4_f["count"]),
        "4+_enc": list(data4_enc_f["count"]),
    }
    #print(data_merged_raw)
    data_merged = pandas.concat([pandas.Series(v, name=k) for k, v in data_merged_raw.items()], axis=1)
    data_merged.to_csv(load_data.get_cache_path("cache_temp_data_merged_smime", testrun), sep=',', quotechar='\'')
    # data_for_graph = {"all": data_all, "1":data1, "2":data2, "3":data3, "4+":data4}
    # print(pandas.concat(data_all, data1, data2, data3, data4))
    #print(data_merged)
    #print(data2, data2_enc)
    generate_smime_usage_after_boxplot(data_merged,
                                       "SMIME Usage after first SMIME use")
    return data_merged_raw

def grouped_encryption_useragents(data):
    #grouped_multiple_general = grouped_encryption_useragent_general(data)
    #grouped_multiple_pgp_attach = grouped_encryption_useragent_pgp_attachment(data)
    #grouped_multiple_pgp_inline = grouped_encryption_useragent_pgp_inline(data)

    print("SMIMEPGPDEBUG")
    data_smime = graph_smime(data)
    print(data_smime)
    data_pgp = graph_pgp(data)
    print(data_pgp)

    table_smime_pgp_enc(data_merged_smime=data_smime, data_merged_pgp=data_pgp)



def table_smime_pgp_enc(data_merged_smime, data_merged_pgp):
    #ss = smime signed
    #se = smime encrypted
    #ps = pgp signed
    #pe = pgp encrypted
    ss1 = np.array(data_merged_smime['1'])
    ss2 = np.array(data_merged_smime['2'])
    ss3 = np.array(data_merged_smime['3'])
    ss4 = np.array(data_merged_smime['4+'])
    se1 = np.array(data_merged_smime['1_enc'])
    se2 = np.array(data_merged_smime['2_enc'])
    se3 = np.array(data_merged_smime['3_enc'])
    se4 = np.array(data_merged_smime['4+_enc'])
    ps1 = np.array(data_merged_pgp['1'])
    ps2 = np.array(data_merged_pgp['2'])
    ps3 = np.array(data_merged_pgp['3'])
    ps4 = np.array(data_merged_pgp['4+'])
    pe1 = np.array(data_merged_pgp['1_enc'])
    pe2 = np.array(data_merged_pgp['2_enc'])
    pe3 = np.array(data_merged_pgp['3_enc'])
    pe4 = np.array(data_merged_pgp['4+_enc'])

    ts = np.concatenate((ss1, ss2, ss3, ss4))
    tp = np.concatenate((ps1, ps2, ps3, ps4))
    tes = np.concatenate((se1, se2, se3, se4))
    tep = np.concatenate((pe1, pe2, pe3, pe4))

    smime_sign_users = len(ss1[ss1 != 0]) + len(ss2[ss2 != 0]) + len(ss3[ss3 != 0]) + len(ss4[ss4 != 0])
    pgp_sign_users = len(ps1[ps1 != 0]) + len(ps2[ps2 != 0]) + len(ps3[ps3 != 0]) + len(ps4[ps4 != 0])
    smime_enc_users = len(se1[se1 != 0]) + len(se2[se2 != 0]) + len(se3[se3 != 0]) + len(se4[se4 != 0])
    pgp_enc_users = len(pe1[pe1 != 0]) + len(pe2[pe2 != 0]) + len(pe3[pe3 != 0]) + len(pe4[pe4 != 0])

    add_tex_result("SmimeUsersSignedAvg", f'{np.mean(ts):.2f}', "MultiClient")
    add_tex_result("SmimeUsersSignedStd", f'{np.std(ts):.2f}', "MultiClient")
    add_tex_result("SmimeUsersSignedMedian", f'{np.median(ts):.2f}', "MultiClient")
    add_tex_result("SmimeUsersCountVerify", f'{ts.size:.2f}', "MultiClient")
    add_tex_result("PgpUsersSignedAvg", f'{np.mean(tp):.2f}', "MultiClient")
    add_tex_result("PgpUsersSignedStd", f'{np.std(tp):.2f}', "MultiClient")
    add_tex_result("PgpUsersSignedMedian", f'{np.median(tp):.2f}', "MultiClient")
    add_tex_result("PgpUsersCountVerify", f'{tp.size:.2f}', "MultiClient")

    add_tex_result("SmimeUsersEncAvg", f'{np.mean(tes):.2f}', "MultiClient")
    add_tex_result("PgpUsersEncAvg", f'{np.mean(tep):.2f}', "MultiClient")

    add_tex_result("MultiClientSmimeSignedOneMedian", f'{np.median(ss1):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeSignedTwoMedian", f'{np.median(ss2):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeSignedThreeMedian", f'{np.median(ss3):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeSignedFourOrHigherMedian", f'{np.median(ss4):.2f}', "MultiClient")

    add_tex_result("MultiClientSmimeEncryptedOneMedian", f'{np.median(se1):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeEncryptedTwoMedian", f'{np.median(se2):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeEncryptedThreeMedian", f'{np.median(se3):.2f}', "MultiClient")
    add_tex_result("MultiClientSmimeEncryptedFourOrHigherMedian", f'{np.median(se4):.2f}', "MultiClient")

    add_tex_result("MultiClientPgpSignedOneMedian", f'{np.median(ps1):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpSignedTwoMedian", f'{np.median(ps2):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpSignedThreeMedian", f'{np.median(ps3):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpSignedFourOrHigherMedian", f'{np.median(ps4):.2f}', "MultiClient")

    add_tex_result("MultiClientPgpEncryptedOneMedian", f'{np.median(pe1):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpEncryptedTwoMedian", f'{np.median(pe2):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpEncryptedThreeMedian", f'{np.median(pe3):.2f}', "MultiClient")
    add_tex_result("MultiClientPgpEncryptedFourOrHigherMedian", f'{np.median(pe4):.2f}', "MultiClient")

    result = ""
    result += r"\begin{table*}[]" + "\n"
    result += r"\begin{tabular}{lllllllllllllllll}" + "\n"
    result += r"      & \multicolumn{8}{l}{SMIME}                                  & \multicolumn{8}{l}{PGP}                                    \\" + "\n"
    result += r"      & \multicolumn{4}{c|}{Percentage signed} & \multicolumn{4}{c|}{Percentage encrypted} & \multicolumn{4}{c|}{Percentage signed} & \multicolumn{4}{c}{Percentage encrypted}\\" + "\n"
    result += r"      & \multicolumn{4}{c|}{Clients used} & \multicolumn{4}{c|}{Clients used}                               & \multicolumn{4}{c|}{Clients used} & \multicolumn{4}{c}{Clients used}\\" + "\n"
    result += r"      & 1     & 2    & 3    & 4    & 1     & 2     & 3     & 4     & 1     & 2    & 3    & 4    & 1     & 2     & 3     & 4     \\" + "\n"
    result += f"adresses & {len(ss1)} & {len(ss2)} & {len(ss3)} & {len(ss4)} & {len(se1)} & {len(se2)} & {len(se3)} & {len(se4)} & {len(ps1)} & {len(ps2)} & {len(ps3)} & {len(ps4)} & {len(pe1)} & {len(pe2)} & {len(pe3)} & {len(pe4)} \\\\" + "\n"
    result += f"mean     & {np.mean(ss1):.2f} & {np.mean(ss2):.2f} & {np.mean(ss3):.2f} & {np.mean(ss4):.2f} & {np.mean(se1):.2f} & {np.mean(se2):.2f} & {np.mean(se3):.2f} & {np.mean(se4):.2f} & {np.mean(ps1):.2f} & {np.mean(ps2):.2f} & {np.mean(ps3):.2f} & {np.mean(ps4):.2f} & {np.mean(pe1):.2f} & {np.mean(pe2):.2f} & {np.mean(pe3):.2f} & {np.mean(pe4):.2f} \\\\" + "\n"
    result += f"std      & {np.std(ss1):.2f} & {np.std(ss2):.2f} & {np.std(ss3):.2f} & {np.std(ss4):.2f} & {np.std(se1):.2f} & {np.std(se2):.2f} & {np.std(se3):.2f} & {np.std(se4):.2f} & {np.std(ps1):.2f} & {np.std(ps2):.2f} & {np.std(ps3):.2f} & {np.std(ps4):.2f} & {np.std(pe1):.2f} & {np.std(pe2):.2f} & {np.std(pe3):.2f} & {np.std(pe4):.2f} \\\\" + "\n"
    #result += f"min      & {np.min(ss1):.2f} & {np.min(ss2):.2f} & {np.min(ss3):.2f} & {np.min(ss4):.2f} & {np.min(se1):.2f} & {np.min(se2):.2f} & {np.min(se3):.2f} & {np.min(se4):.2f} & {np.min(ps1):.2f} & {np.min(ps2):.2f} & {np.min(ps3):.2f} & {np.min(ps4):.2f} & {np.min(pe1):.2f} & {np.min(pe2):.2f} & {np.min(pe3):.2f} & {np.min(pe4):.2f} \\\\" + "\n"
    result += f"25\%     & {np.quantile(ss1, 0.25):.2f} & {np.quantile(ss2, 0.25):.2f} & {np.quantile(ss3, 0.25):.2f} & {np.quantile(ss4, 0.25):.2f} & {np.quantile(se1, 0.25):.2f} & {np.quantile(se2, 0.25):.2f} & {np.quantile(se3, 0.25):.2f} & {np.quantile(se4, 0.25):.2f} & {np.quantile(ps1, 0.25):.2f} & {np.quantile(ps2, 0.25):.2f} & {np.quantile(ps3, 0.25):.2f} & {np.quantile(ps4, 0.25):.2f} & {np.quantile(pe1, 0.25):.2f} & {np.quantile(pe2, 0.25):.2f} & {np.quantile(pe3, 0.25):.2f} & {np.quantile(pe4, 0.25):.2f} \\\\" + "\n"
    result += f"50\%     & {np.quantile(ss1, 0.50):.2f} & {np.quantile(ss2, 0.50):.2f} & {np.quantile(ss3, 0.50):.2f} & {np.quantile(ss4, 0.50):.2f} & {np.quantile(se1, 0.50):.2f} & {np.quantile(se2, 0.50):.2f} & {np.quantile(se3, 0.50):.2f} & {np.quantile(se4, 0.50):.2f} & {np.quantile(ps1, 0.50):.2f} & {np.quantile(ps2, 0.50):.2f} & {np.quantile(ps3, 0.50):.2f} & {np.quantile(ps4, 0.50):.2f} & {np.quantile(pe1, 0.50):.2f} & {np.quantile(pe2, 0.50):.2f} & {np.quantile(pe3, 0.50):.2f} & {np.quantile(pe4, 0.50):.2f} \\\\" + "\n"
    result += f"75\%     & {np.quantile(ss1, 0.75):.2f} & {np.quantile(ss2, 0.75):.2f} & {np.quantile(ss3, 0.75):.2f} & {np.quantile(ss4, 0.75):.2f} & {np.quantile(se1, 0.75):.2f} & {np.quantile(se2, 0.75):.2f} & {np.quantile(se3, 0.75):.2f} & {np.quantile(se4, 0.75):.2f} & {np.quantile(ps1, 0.75):.2f} & {np.quantile(ps2, 0.75):.2f} & {np.quantile(ps3, 0.75):.2f} & {np.quantile(ps4, 0.75):.2f} & {np.quantile(pe1, 0.75):.2f} & {np.quantile(pe2, 0.75):.2f} & {np.quantile(pe3, 0.75):.2f} & {np.quantile(pe4, 0.75):.2f} \\\\" + "\n"
    result += f"max      & {np.max(ss1):.2f} & {np.max(ss2):.2f} & {np.max(ss3):.2f} & {np.max(ss4):.2f} & {np.max(se1):.2f} & {np.max(se2):.2f} & {np.max(se3):.2f} & {np.max(se4):.2f} & {np.max(ps1):.2f} & {np.max(ps2):.2f} & {np.max(ps3):.2f} & {np.max(ps4):.2f} & {np.max(pe1):.2f} & {np.max(pe2):.2f} & {np.max(pe3):.2f} & {np.max(pe4):.2f}" + "\n"
    result += r"\end{tabular}" + "\n"
    result += r"\caption{Percentage of signed and encrypted emails after first SMIME or PGP usage on each mailaddress}" + "\n"
    result += r"\end{table*}" + "\n"

    with open(load_data.get_cache_path("paper/table_smime_pgp_after_usage.tex", testrun), 'w') as f:
        f.write(result)
    print(result)

def group_message_ids(data):
    cache_name = "message_id_most_common"
    if is_cached(cache_name):
        grouped_multiple = load_from_cache(cache_name)
    else:
        start = time.time()
        grouped_multiple = data.groupby('message_id').agg({'header_date': ['count', 'min', 'max']})#.rename(columns={"count": "count", "min": "min", "max": "max"})
        grouped_multiple.columns = grouped_multiple.columns.get_level_values(1)
        grouped_multiple.sort_values('count', ascending=False)
        store_to_cache(cache_name, grouped_multiple)
        end = time.time()
        print("Grouping message IDs finished in", ((end - start) / 60), "min")
        #TODO: Matplotlib gets stuck here, if not loaded from cache, reload in any case for now
        grouped_multiple = load_from_cache(cache_name)

    #generate_message_id_plot(grouped_multiple["count"], "Message IDs")
    grouped_multiple = grouped_multiple.quantile([0.9, 0.95, 0.99, 0.995, 0.999, 0.9999, 0.99999])
    print(grouped_multiple)

def generate_smime_usage_after_plot(data, title, name_appendix=""):
    generate_usage_after_plot(data, title, name_appendix = name_appendix, type="SMIME")

def generate_pgp_usage_after_plot(data, title, name_appendix=""):
    generate_usage_after_plot(data, title, name_appendix = name_appendix, type="PGP")

def generate_usage_after_plot(data, title, name_appendix="", type="SMIME"):
    plt.clf()
    plt.figure(figsize=(9, 6))

    x = data["count"]
    if (len(x) > 0):
        #print(x)

        q25, q75 = np.percentile(x, [.25, .75])
        #print(q25, q75)
        #bin_width = 2 * (q75 - q25) * len(x) ** (-1 / 3)
        bins = 20#round((x.max() - x.min()) / bin_width)
        #print("Freedman–Diaconis number of bins:", bins)

        plt.hist(x, bins=bins);
        plt.ylabel('Users')
        plt.xlabel('Percentage of signed mails after first '+str(type)+' usage')
        plt.xticks(rotation=45)
        plt.title(title)
        plt.savefig(load_data.get_cache_path('graphs/'+type.lower()+name_appendix+'_after_first_use.png', testrun))

        plt.clf()

        plt.hist(x, density=True, bins=bins, label="Data")
        mn, mx = plt.xlim()
        plt.xlim(mn, mx)
        kde_xs = np.linspace(mn, mx, 300)
        try:
            kde = st.gaussian_kde(x)
            plt.plot(kde_xs, kde.pdf(kde_xs), label="PDF")
        except:
            print("Failed PDF plot - not enough elements?", len(x))
        plt.legend(loc="upper left")
        plt.ylabel('Probability')
        plt.xlabel('Percentage of signed mails after first '+str(type)+' usage')

        plt.xticks(rotation=45)
        plt.title(title+" ("+str(len(x))+" users)")

        plt.savefig(load_data.get_cache_path('graphs/'+type.lower()+name_appendix+'_after_first_use_density.png', testrun))

        print(type+name_appendix+"_after_first_use.png",x.describe())

    else:
        print("[Empty] Skipping "+type+name_appendix+"_after_first_use.png")
        raise Exception

def generate_smime_usage_after_boxplot(data, title):
    plt.clf()
    plt.figure(figsize=(9, 6))

    color1 = "cornflowerblue"
    color2 = "coral"
    mediancolor = "darkgrey"

    plot = data.boxplot(column=['1', '1_enc', '2', '2_enc', '3', '3_enc', '4+', '4+_enc'], grid=False,
                        positions=[1, 1.5, 2.5, 3, 4, 4.5, 5.5, 6], widths=0.4, patch_artist=True,
                        medianprops = dict(color=mediancolor),
                        boxprops = dict(facecolor='r'), return_type="both")

    i = 0
    for patch in plot.lines["boxes"]:
        i += 1
        if i%2 == 0:
            patch.set_facecolor(color1)
        else:
            patch.set_facecolor(color2)

    # Choose the position of each barplots on the x-axis (space=1,4,3,1)
    y_pos = [1.25, 2.75, 4.24, 5.75]

    plt.ylabel('Percentage of signed mails after first S/MIME usage')
    plt.xlabel('Clients used')
    plt.xticks(y_pos, ('1 Client', '2 Clients', '3 Clients', '4 or more Clients'))  # signed / encrypted\n

    custom_lines = [Line2D([0], [0], color=color2, lw=4),
                    Line2D([0], [0], color=color1, lw=4),
                    ]

    plt.legend(custom_lines, ['Signed', 'Encrypted'])
    #plt.title(title)
    plt.savefig(load_data.get_cache_path('graphs/boxplot_smime_after_first_use.png', testrun))

    # Prepare data for seaborn
    data.columns = ['1 Client_Signed', '1 Client_Encrypted', '2 Clients_Signed', '2 Clients_Encrypted', '3 Clients_Signed', '3 Clients_Encrypted', '4 or more Clients_Signed',
                    '4 or more Clients_Encrypted']
    # Modin Bug workaround: Use normal pandas instead of modin, otherwise Seaborn crashes
    melted_data = pd.melt(data._to_pandas())
    # print(melted_data)
    melted_data[['variable', 'type']] = melted_data.variable.str.split("_", expand=True)
    # print(melted_data)
    melted_data.dropna(inplace=True)
    # print(melted_data)
    # print(melted_data.dtypes)
    melted_data['variable'] = melted_data['variable'].astype("category")
    melted_data['type'] = melted_data['type'].astype("category")
    # print(melted_data.dtypes)
    melted_data.columns = ['variable', 'value', 'type']

    # Create paperready graph with seaborn
    graph_support.seaborn_boxplot_paper("sns_boxplot_smime_after_first_use.png", data=melted_data, x_var="variable",
                                        y_var="value", hue="type", hue_order=["Signed", "Encrypted"],
                                        xlabel="# Clients used", ylabel="Percentage of signed or encrypted S/MIME mails",
                                        testrun=testrun)

def generate_pgp_usage_after_boxplot(data, title):
    plt.clf()
    plt.figure(figsize=(9, 6))

    color1 = "cornflowerblue"
    color2 = "coral"


    plot = data.boxplot(column=['1', '1_enc', '2', '2_enc', '3', '3_enc', '4+', '4+_enc'], grid=False, positions=[1,1.5,2.5,3,4,4.5,5.5,6], widths=0.4, patch_artist = True, boxprops=dict(facecolor='r'), return_type="both")

    i = 0
    for patch in plot.lines["boxes"]:
        i+=1
        if i%2 == 0:
            patch.set_facecolor(color1)
        else:
            patch.set_facecolor(color2)

    # Choose the position of each barplots on the x-axis (space=1,4,3,1)
    y_pos = [1.25, 2.75, 4.24, 5.75]

    plt.ylabel('Percentage of signed mails after first PGP usage')
    plt.xlabel('Clients used')
    plt.xticks(y_pos, ('1 Client', '2 Clients', '3 Clients', '4 or more Clients')) #signed / encrypted\n

    custom_lines = [Line2D([0], [0], color=color2, lw=4),
                    Line2D([0], [0], color=color1, lw=4),
                    ]

    plt.legend(custom_lines, ['Signed', 'Encrypted'])

    #plt.title(title)
    plt.savefig(load_data.get_cache_path('graphs/boxplot_pgp_after_first_use.png', testrun))

    # Prepare data for seaborn
    data.columns = ['1 Client_Signed', '1 Client_Encrypted', '2 Clients_Signed', '2 Clients_Encrypted',
                    '3 Clients_Signed', '3 Clients_Encrypted', '4 or more Clients_Signed',
                    '4 or more Clients_Encrypted']
    #Modin Bug workaround: Use normal pandas instead of modin, otherwise Seaborn crashes
    melted_data = pd.melt(data._to_pandas())
    #print(melted_data)
    melted_data[['variable', 'type']] = melted_data.variable.str.split("_", expand=True)
    #print(melted_data)
    melted_data.dropna(inplace = True)
    #print(melted_data)
    #print(melted_data.dtypes)
    melted_data['variable'] = melted_data['variable'].astype("category")
    melted_data['type'] = melted_data['type'].astype("category")
    #print(melted_data.dtypes)
    melted_data.columns = ['variable', 'value', 'type']


    # Create paperready graph with seaborn
    graph_support.seaborn_boxplot_paper("sns_boxplot_pgp_after_first_use.png", data = melted_data, x_var="variable",
                                        y_var="value", hue="type", hue_order=["Signed", "Encrypted"],
                                        xlabel = "# Clients used", ylabel = "Percentage of signed or encrypted PGP mails",
                                        testrun = testrun)

def generate_message_id_plot(data_y, title):
    plt.figure(figsize=(9, 6))


    #ind = np.arange(len(labels))  # the x locations for the groups
    #width = 0.35  # the width of the bars: can also be len(x) sequence

    p1 = plt.plot(data_y)
    # Add some text for labels, title and custom x-axis tick labels, etc.
    plt.ylabel('Message ID')
    #plt.xticks(ind, labels)
    plt.yscale('log')

    plt.xticks(rotation=45)
    plt.title(title)

    plt.savefig(load_data.get_cache_path('graphs/message_id.png', testrun))

def add_tex_result(name, value, category = "general"):
    global tex_results
    if category not in tex_results:
        tex_results[category] = {}
    tex_results[category][name] = value
    print(f"[Debug] Added tex value {name} with value {value} in category {category}.")

def add_tex_var(name, value, category = "general"):
    global tex_var_results
    if category not in tex_var_results:
        tex_var_results[category] = {}
    tex_var_results[category][name] = value
    print(f"[Debug] Added tex variable {name} with value {value} in category {category}.")

def all_tex_results():
    global tex_results
    global tex_var_results
    with open(load_data.get_cache_path("paper/generated_chris.tex", testrun), 'w') as f:
        for category in tex_results:
            f.write('%%% '+category+" %%%"+"\n")
            for entry in tex_results[category]:
                f.write(tex_command(entry, tex_results[category][entry])+"\n")
                print(tex_command(entry, tex_results[category][entry]))
        for category in tex_var_results:
            f.write('%%% '+category+" %%%"+"\n")
            for entry in tex_var_results[category]:
                f.write(tex_var(entry, tex_var_results[category][entry])+"\n")
                print(tex_var(entry, tex_var_results[category][entry]))

def add_table_result(line):
    global table_results
    table_results += line + "\n"
    #print(line)

def all_table_results():
    global table_results
    with open(load_data.get_cache_path("paper/table_demographics.tex", testrun), 'w') as f:
        f.write(table_results)
        print(table_results)

def split_sent_received_mails(data):
    # sent = data[(data['folder'].str.contains('Postbox.sent') | data['folder'].str.contains('Postbox.sent_sub'))]
    # received = data[~(data['folder'].str.contains('Postbox.sent') | data['folder'].str.contains('Postbox.sent_sub'))]
    sent = data_support.get_sent_mails(data)
    received = data_support.get_received_mails(data)
    active_users = set(sent['user'])
    received["active_user"] = received['user'].isin(active_users).astype(int)
    print(sent.shape[0], received.shape[0], received.loc[received["active_user"] == True].shape[0], received.loc[received["active_user"] == False].shape[0])
    return sent, received

def color_code_entry(data, max_value):
    min_gray = 10
    max_gray = 100
    max_percent = 5
    max_value = max(max_value, 1)
    percent = min(data / max_value * 100, max_percent)
    #print(data, max_value, percent, percent*(100/
    if data == 0:
        return f'{data:,}'
    else:
        return r'\cellcolor{gray!'+str(min_gray+round(percent*((max_gray-min_gray)/max_percent)))+'}'+f'{data:,}'

def table_line(entry, sent_count, received_count, smime_sent_sign_count, smime_sent_enc_count, smime_received_sign_count, smime_received_enc_count, pgp_sent_sign_count, pgp_sent_enc_count, pgp_received_sign_count, pgp_received_enc_count):
    entry = entry.replace("[d]","").replace("[i]","").strip()
    smime_sent_sign_count = color_code_entry(smime_sent_sign_count, sent_count)
    smime_sent_enc_count = color_code_entry(smime_sent_enc_count, sent_count)
    smime_received_sign_count = color_code_entry(smime_received_sign_count, received_count)
    smime_received_enc_count = color_code_entry(smime_received_enc_count, received_count)

    pgp_sent_sign_count = color_code_entry(pgp_sent_sign_count, sent_count)
    pgp_sent_enc_count = color_code_entry(pgp_sent_enc_count, sent_count)
    pgp_received_sign_count = color_code_entry(pgp_received_sign_count, received_count)
    pgp_received_enc_count = color_code_entry(pgp_received_enc_count, received_count)

    line = f'& {entry} & {smime_sent_sign_count} & {smime_sent_enc_count} & {smime_received_sign_count} & {smime_received_enc_count} & {pgp_sent_sign_count} & {pgp_sent_enc_count} & {pgp_received_sign_count} & {pgp_received_enc_count} & {sent_count:,} & {received_count:,}\\\\'
    #print(line)
    return line

@cache_it("cache_table_value_counts_usergroup", reload=True)
def table_value_counts_usergroup(full_df):
    return full_df["usergroup"].value_counts().reset_index()

@cache_it("cache_table_value_counts_operating_system", reload=True)
def table_value_counts_operating_system(full_df):
    return full_df["operating_system"].value_counts().reset_index()

@cache_it("cache_table_value_counts_clients", reload=True)
def table_value_counts_clients(full_df):
    return full_df["client_merged_top5"].value_counts().reset_index()

@cache_it("cache_table_value_counts_clients_all", reload=True)
def table_value_counts_clients_all(full_df):
    return full_df["client_merged"].value_counts().reset_index()

@cache_it("cache_tex_client_count_sent_mails", is_df=False, is_json=True)
def tex_client_count_sent_mails(sent_df):
    #Gets MultiClient stats for sending users.

    clients = sent_df.groupby('user').agg({'client_merged': ['nunique']}).reset_index()
    clients.columns = ["user", "nunique"]
    clients_single = clients.loc[clients["nunique"] == 1]
    clients_multi = clients.loc[clients["nunique"] > 1]
    return {"users_client_multi": clients_multi.shape[0], "users_clients_single": clients_single.shape[0], "users_clients": clients.shape[0]}

def remove_empty_sender(data):
    try:
        return data.loc[data["sender"] != "empty"]
    except:
        print("WARNING: remove_empty_sender exception, returning base value")
        return data

def generate_table_data(full_df, sent_mails_df, received_mails_df):
    #os_group = sent_mails_df.groupby('operating_system').agg(
    #    {'header_date': ['count']})
    #print(os_group)

    add_tex_var('MailsIn1994', f'{full_df.loc[full_df["header_year"] == 1994].shape[0]:,}')
    add_tex_var('MailsIn2020', f'{full_df.loc[full_df["header_year"] == 2020].shape[0]:,}')

    multi_client_usage_users = tex_client_count_sent_mails(sent_mails_df)

    add_tex_result("SingleClientUsers", f'{multi_client_usage_users["users_clients_single"]:,}', "MultiClient")
    add_tex_result("MultiClientUsers", f'{multi_client_usage_users["users_client_multi"]:,}', "MultiClient")
    add_tex_result("MultiClientUsersPercent", f'{(multi_client_usage_users["users_client_multi"] / multi_client_usage_users["users_clients"] * 100):.2f}', "MultiClient")
    add_tex_result("ClientUsersTotal", f'{multi_client_usage_users["users_clients"]:,}', "MultiClient")

    sent_smime_sig_mails = sent_mails_df.loc[((sent_mails_df["smime_sig_fixed"] == True))]
    #sent_smime_sig_mails = remove_empty_sender(sent_smime_sig_mails)
    sent_smime_enc_mails = sent_mails_df.loc[sent_mails_df["smime_encryption"] == "SMIMEEnc.enc"]
    #sent_smime_enc_mails = remove_empty_sender(sent_smime_enc_mails)
    sent_pgp_sig_mails = sent_mails_df.loc[((sent_mails_df["pgp_signature_merged"] == True))]
    #sent_pgp_sig_mails = remove_empty_sender(sent_pgp_sig_mails)
    sent_pgp_enc_mails = sent_mails_df.loc[sent_mails_df["pgp_encryption_merged"] == True]
    #sent_pgp_enc_mails = remove_empty_sender(sent_pgp_enc_mails)
    sent_sig_mails = sent_mails_df.loc[((sent_mails_df["smime_sig_fixed"] == True) | (sent_mails_df["pgp_signature_merged"] == True))]
    #sent_sig_mails = remove_empty_sender(sent_sig_mails)
    sent_enc_mails = sent_mails_df.loc[((sent_mails_df["smime_encryption"] == "SMIMEEnc.enc") | (sent_mails_df["pgp_encryption_merged"] == True))]
    #sent_enc_mails = remove_empty_sender(sent_enc_mails)
    received_smime_sig_mails = received_mails_df.loc[received_mails_df["smime_sig_fixed"] == True]
    received_smime_enc_mails = received_mails_df.loc[received_mails_df["smime_encryption"] == "SMIMEEnc.enc"]
    received_pgp_sig_mails = received_mails_df.loc[received_mails_df["pgp_signature_merged"] == True]
    received_pgp_enc_mails = received_mails_df.loc[received_mails_df["pgp_encryption_merged"] == True]

    received_smime_sig_mails_active = received_smime_sig_mails.loc[received_smime_sig_mails["active_user"] == True]
    received_pgp_sig_mails_active = received_pgp_sig_mails.loc[received_pgp_sig_mails["active_user"] == True]

    received_smime_sig_users = set(received_smime_sig_mails_active["user"].unique())
    received_pgp_sig_users = set(received_pgp_sig_mails_active["user"].unique())

    #print("[PGP-DEBUG-L1]", sent_pgp_sig_mails)

    sending_users = int(multi_client_usage_users["users_clients"])
    smime_sig_users_uniq = set(sent_smime_sig_mails["user"].unique())
    pgp_sig_users_uniq = set(sent_pgp_sig_mails["user"].unique())
    sig_users_uniq = set(sent_sig_mails["user"].unique())
    smime_enc_users_uniq = set(sent_smime_enc_mails["user"].unique())
    pgp_enc_users_uniq = set(sent_pgp_enc_mails["user"].unique())
    enc_users_uniq = set(sent_enc_mails["user"].unique())

    smime_sig_users = len(smime_sig_users_uniq)
    pgp_sig_users = len(pgp_sig_users_uniq)
    sig_users = len(sig_users_uniq)
    smime_enc_users = len(smime_enc_users_uniq)
    pgp_enc_users = len(pgp_enc_users_uniq)
    enc_users = len(enc_users_uniq)

    #Exclusive users

    smime_enc_exclusive = len(enc_users_uniq - pgp_enc_users_uniq)
    pgp_enc_exclusive = len(enc_users_uniq - smime_enc_users_uniq)

    smime_sig_exclusive = len(sig_users_uniq - pgp_sig_users_uniq)
    pgp_sig_exclusive = len(sig_users_uniq - smime_sig_users_uniq)

    smime_and_pgp_sig = len(smime_sig_users_uniq.intersection(pgp_sig_users_uniq))
    smime_and_pgp_enc = len(smime_enc_users_uniq.intersection(pgp_enc_users_uniq))

    add_tex_result("TotalSendingUsers", f'{sending_users:,}', "GeneralCrypto")

    add_tex_result("ReceivedSMIMESigUsers", f'{len(received_smime_sig_users):,}', "GeneralCrypto")
    add_tex_result("ReceivedPGPSigUsers", f'{len(received_pgp_sig_users):,}', "GeneralCrypto")
    add_tex_result("ReceivedSigUsers", f'{len(received_smime_sig_users.union(received_pgp_sig_users)):,}', "GeneralCrypto")
    add_tex_result("ReceivedSMIMESigUsersPercent", f'{len(received_smime_sig_users)/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("ReceivedPGPSigUsersPercent", f'{len(received_pgp_sig_users)/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("ReceivedSigUsersPercent", f'{len(received_smime_sig_users.union(received_pgp_sig_users))/sending_users*100:.2f}', "GeneralCrypto")

    add_tex_result("SMIMESigUsers", f'{smime_sig_users:,}', "GeneralCrypto")
    add_tex_result("PGPSigUsers", f'{pgp_sig_users:,}', "GeneralCrypto")
    add_tex_result("TotalSigUsers", f'{sig_users:,}', "GeneralCrypto")
    add_tex_result("SMIMEEncUsers", f'{smime_enc_users:,}', "GeneralCrypto")
    add_tex_result("PGPEncUsers", f'{pgp_enc_users:,}', "GeneralCrypto")
    add_tex_result("TotalEncUsers", f'{enc_users:,}', "GeneralCrypto")

    add_tex_result("SMIMESigUsersExclusive", f'{smime_sig_exclusive:,}', "GeneralCrypto")
    add_tex_result("PGPSigUsersExclusive", f'{pgp_sig_exclusive:,}', "GeneralCrypto")
    add_tex_result("SMIMEEncUsersExclusive", f'{smime_enc_exclusive:,}', "GeneralCrypto")
    add_tex_result("PGPEncUsersExclusive", f'{pgp_enc_exclusive:,}', "GeneralCrypto")
    add_tex_result("SMIMEAndPGPSigUsers", f'{smime_and_pgp_sig:,}', "GeneralCrypto")
    add_tex_result("SMIMEAndPGPEncUsers", f'{smime_and_pgp_enc:,}', "GeneralCrypto")

    add_tex_result("SMIMESigUsersPercent", f'{smime_sig_users/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("PGPSigUsersPercent", f'{pgp_sig_users/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("TotalSigUsersPercent", f'{sig_users/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("SMIMEEncUsersPercent", f'{smime_enc_users/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("PGPEncUsersPercent", f'{pgp_enc_users/sending_users*100:.2f}', "GeneralCrypto")
    add_tex_result("TotalEncUsersPercent", f'{enc_users/sending_users*100:.2f}', "GeneralCrypto")

    add_tex_result("NeverCrypto", f'{sending_users-sig_users:,}', "GeneralCrypto")
    add_tex_result("NeverCryptoPercent", f'{(sending_users-sig_users)/sending_users*100:.2f}', "GeneralCrypto")

    #Get 2020 data
    sending_users_2020 = sent_mails_df.loc[sent_mails_df["header_year"] == 2020]
    smime_sig_users_2020 = sent_smime_sig_mails.loc[sent_smime_sig_mails["header_year"] == 2020]
    pgp_sig_users_2020 = sent_pgp_sig_mails.loc[sent_pgp_sig_mails["header_year"] == 2020]
    crypto_users_2020 = sent_sig_mails.loc[sent_sig_mails["header_year"] == 2020]

    sending_users_2020_count = sending_users_2020["user"].nunique()

    add_tex_var("SendingUsers2020", f'{sending_users_2020_count:,}', "GeneralCrypto")
    add_tex_var("SMIMEUsers2020", f'{smime_sig_users_2020["user"].nunique():,}', "GeneralCrypto")
    add_tex_var("PGPUsers2020", f'{pgp_sig_users_2020["user"].nunique():,}', "GeneralCrypto")
    add_tex_var("CryptoUsers2020", f'{crypto_users_2020["user"].nunique():,}', "GeneralCrypto")
    add_tex_var("SMIMEUsers2020Percent", f'{smime_sig_users_2020["user"].nunique() / sending_users_2020_count * 100:.2f}', "GeneralCrypto")
    add_tex_var("PGPUsers2020Percent", f'{pgp_sig_users_2020["user"].nunique() / sending_users_2020_count * 100:.2f}', "GeneralCrypto")
    add_tex_var("CryptoUsers2020Percent", f'{crypto_users_2020["user"].nunique() / sending_users_2020_count * 100:.2f}', "GeneralCrypto")

    #add_tex_result("UsersSentSMIMESigned", ,"CryptoUsage")

    usergroup = table_value_counts_usergroup(sent_smime_sig_mails)
    operating_systems = table_value_counts_operating_system(sent_smime_sig_mails)
    #clients = table_value_counts_clients(sent_sig_mails)
    #clients_all = table_value_counts_clients_all(sent_sig_mails)

    sent_count = sent_mails_df.shape[0]
    received_count = received_mails_df.shape[0]
    smime_sent_sign_count = sent_smime_sig_mails.shape[0]
    smime_sent_enc_count = sent_smime_enc_mails.shape[0]
    smime_received_sign_count = received_smime_sig_mails.shape[0]
    smime_received_enc_count = received_smime_enc_mails.shape[0]

    pgp_sent_sign_count = sent_pgp_sig_mails.shape[0]
    pgp_sent_enc_count = sent_pgp_enc_mails.shape[0]
    pgp_received_sign_count = received_pgp_sig_mails.shape[0]
    pgp_received_enc_count = received_pgp_enc_mails.shape[0]

    sent_count_unknown_useragent = sent_mails_df.loc[sent_mails_df["client_merged_top5"] == "[d]No Useragent"].shape[0]
    sent_count_known_useragent = sent_mails_df.loc[sent_mails_df["client_merged_top5"] != "[d]No Useragent"].shape[0]
    received_count_unknown_useragent = received_mails_df.loc[received_mails_df["client_merged_top5"] == "[d]No Useragent"].shape[0]
    received_count_known_useragent = received_mails_df.loc[received_mails_df["client_merged_top5"] != "[d]No Useragent"].shape[0]
    add_tex_result("NoUseragentTotal", f'{(sent_count_unknown_useragent+received_count_unknown_useragent):,}' ,"MultiClient")
    add_tex_result("NoUseragentTotalPercent", f'{(sent_count_unknown_useragent+received_count_unknown_useragent)/(sent_count+received_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedTotal", f'{(sent_count_known_useragent+received_count_known_useragent):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedTotalPercent", f'{(sent_count_known_useragent+received_count_known_useragent)/(sent_count+received_count)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentSent", f'{(sent_count_unknown_useragent):,}' ,"MultiClient")
    add_tex_result("NoUseragentSentPercent", f'{(sent_count_unknown_useragent)/(sent_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSent", f'{(sent_count_known_useragent):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSentPercent", f'{(sent_count_known_useragent)/(sent_count)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentReceived", f'{received_count_unknown_useragent:,}' ,"MultiClient")
    add_tex_result("NoUseragentReceivedPercent", f'{(received_count_unknown_useragent)/(received_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedReceived", f'{received_count_known_useragent:,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedReceivedPercent", f'{(received_count_known_useragent)/(received_count)*100:.2f}' ,"MultiClient")

    sent_smime_sig_count_unknown_useragent = sent_smime_sig_mails.loc[sent_smime_sig_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    received_smime_sig_count_unknown_useragent = received_smime_sig_mails.loc[received_smime_sig_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    sent_pgp_sig_count_unknown_useragent = sent_pgp_sig_mails.loc[sent_pgp_sig_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    received_pgp_sig_count_unknown_useragent = received_pgp_sig_mails.loc[received_pgp_sig_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]

    sent_smime_sig_count_known_useragent = sent_smime_sig_mails.loc[sent_smime_sig_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]
    received_smime_sig_count_known_useragent = received_smime_sig_mails.loc[received_smime_sig_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]
    sent_pgp_sig_count_known_useragent = sent_pgp_sig_mails.loc[sent_pgp_sig_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]
    received_pgp_sig_count_known_useragent = received_pgp_sig_mails.loc[received_pgp_sig_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]

    sent_signed_unknown = sent_smime_sig_count_unknown_useragent+sent_pgp_sig_count_unknown_useragent
    received_signed_unknown = received_smime_sig_count_unknown_useragent + received_pgp_sig_count_unknown_useragent

    sent_signed_known = sent_smime_sig_count_known_useragent+sent_pgp_sig_count_known_useragent
    received_signed_known = received_smime_sig_count_known_useragent + received_pgp_sig_count_known_useragent

    signed_sent = smime_sent_sign_count + pgp_sent_sign_count
    signed_received = smime_received_sign_count + pgp_received_sign_count
    signed_total = signed_sent + signed_received

    add_tex_result("NoUseragentSignedTotal", f'{(sent_signed_unknown+received_signed_unknown):,}' ,"MultiClient")
    add_tex_result("NoUseragentSignedTotalPercent", f'{(sent_signed_unknown+received_signed_unknown)/(signed_total)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentSignedSent", f'{(sent_signed_unknown):,}' ,"MultiClient")
    add_tex_result("NoUseragentSignedSentPercent", f'{(sent_signed_unknown)/(signed_sent)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentSignedReceived", f'{received_signed_unknown:,}' ,"MultiClient")
    add_tex_result("NoUseragentSignedReceivedPercent", f'{(received_signed_unknown)/(signed_received)*100:.2f}' ,"MultiClient")

    add_tex_result("UseragentIdentifiedSignedTotal", f'{(sent_signed_known+received_signed_known):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedTotalPercent", f'{(sent_signed_known+received_signed_known)/(signed_total)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedSent", f'{(sent_signed_known):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedSentPercent", f'{(sent_signed_known)/(signed_sent)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedSMIMESent", f'{(sent_smime_sig_count_known_useragent):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedSMIMESentPercent", f'{(sent_smime_sig_count_known_useragent)/(smime_sent_sign_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedPGPSent", f'{(sent_pgp_sig_count_known_useragent):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedPGPSentPercent", f'{(sent_pgp_sig_count_known_useragent)/(pgp_sent_sign_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedReceived", f'{received_signed_known:,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedSignedReceivedPercent", f'{(received_signed_known)/(signed_received)*100:.2f}' ,"MultiClient")

    sent_smime_enc_count_unknown_useragent = sent_smime_enc_mails.loc[sent_smime_enc_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    received_smime_enc_count_unknown_useragent = received_smime_enc_mails.loc[received_smime_enc_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    try:
        sent_pgp_enc_count_unknown_useragent = sent_pgp_enc_mails.loc[sent_pgp_enc_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]
    except:
        sent_pgp_enc_count_unknown_useragent = 0
    received_pgp_enc_count_unknown_useragent = received_pgp_enc_mails.loc[received_pgp_enc_mails["client_merged_top5"] == "[d]No Useragent"].shape[0]

    sent_smime_enc_count_known_useragent = sent_smime_enc_mails.loc[sent_smime_enc_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]
    try:
        sent_pgp_enc_count_known_useragent = sent_pgp_enc_mails.loc[sent_pgp_enc_mails["client_merged_top5"] != "[d]No Useragent"].shape[0]
    except:
        sent_pgp_enc_count_known_useragent = 0

    sent_enc_unknown = sent_smime_enc_count_unknown_useragent+sent_pgp_enc_count_unknown_useragent
    sent_enc_known = sent_smime_enc_count_known_useragent+sent_pgp_enc_count_known_useragent
    received_enc_unknown = received_smime_enc_count_unknown_useragent + received_pgp_enc_count_unknown_useragent
    enc_sent = smime_sent_enc_count + pgp_sent_enc_count
    enc_received = smime_received_enc_count + pgp_received_enc_count
    enc_total = enc_sent + enc_received

    add_tex_result("NoUseragentEncryptionTotal", f'{(sent_enc_unknown+received_enc_unknown):,}' ,"MultiClient")
    add_tex_result("NoUseragentEncryptionTotalPercent", f'{(sent_enc_unknown+received_enc_unknown)/(enc_total)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentEncryptionSent", f'{(sent_enc_unknown):,}' ,"MultiClient")
    add_tex_result("NoUseragentEncryptionSentPercent", f'{(sent_enc_unknown)/(enc_sent)*100:.2f}' ,"MultiClient")
    add_tex_result("NoUseragentEncryptionReceived", f'{received_enc_unknown:,}' ,"MultiClient")
    add_tex_result("NoUseragentEncryptionReceivedPercent", f'{(received_enc_unknown)/(enc_received)*100:.2f}' ,"MultiClient")

    add_tex_result("UseragentIdentifiedEncryptionSent", f'{(sent_enc_known):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedEncryptionSentPercent", f'{(sent_enc_known)/(enc_sent)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedEncryptionSMIMESent", f'{(sent_smime_enc_count_known_useragent):,}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedEncryptionSMIMESentPercent", f'{(sent_smime_enc_count_known_useragent)/(smime_sent_enc_count)*100:.2f}' ,"MultiClient")
    add_tex_result("UseragentIdentifiedEncryptionPGPSent", f'{(sent_pgp_enc_count_known_useragent):,}' ,"MultiClient")
    try:
        add_tex_result("UseragentIdentifiedEncryptionPGPSentPercent", f'{(sent_pgp_enc_count_known_useragent)/(pgp_sent_enc_count)*100:.2f}' ,"MultiClient")
    except:
        if not testrun:
            print("[Error] Division by zero -> UseragentIdentifiedEncryptionPGPSentPercent")
        add_tex_result("UseragentIdentifiedEncryptionPGPSentPercent", f'{0:.2f}' ,"MultiClient")

    add_table_result(table_line(r"\multicolumn{2}{l}{\textbf{Total}}", sent_count, received_count, smime_sent_sign_count, smime_sent_enc_count, smime_received_sign_count, smime_received_enc_count, pgp_sent_sign_count, pgp_sent_enc_count, pgp_received_sign_count, pgp_received_enc_count))
    add_table_result(r"\\[-0.2cm]")
    add_table_result(r"\multicolumn{2}{l}{\textbf{Client}}\\")
    item = "client_merged_top5"
    clients = table_value_counts_clients(full_df)
    for client in ("[d]Thunderbird", "[d]Outlook", "[d]Ticketsystem", "[d]AppleMail", "[d]Evolution", "[d]Mutt/NeoMutt"
            , "[d]Outlook-Express", "[d]Claws Mail", "[d]iPhone/iPad-Mail", "[d]MailMate", "[d]Other", "[d]No Useragent"):
                   #, "[d]mu4e",
                   #"[d]KMail", "[d]eM_Client", "[d]Netscape", "[d]iPadMail",
                   #"[d]No Useragent", "[d]Other", "[d]Unknown Gecko/Mozilla", "[d]SOGoMail", "[d]Outlook-Mac", "[d]K9-Android",
                   #, "[d]OTRS Mail Service Ticketsystem", "[d]Internet Messaging Program", "[d]PHPMailer"):#clients.itertuples():
        item_value = client
        if item_value is not None:
            sent_count_client = sent_mails_df.loc[sent_mails_df[item] == item_value].shape[0]
            received_count_client = received_mails_df.loc[received_mails_df[item] == item_value].shape[0]
            try:
                smime_sent_sign_count_client = sent_smime_sig_mails.loc[sent_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_sent_sign_count_client = -1
            try:
                smime_sent_enc_count_client = sent_smime_enc_mails.loc[sent_smime_enc_mails[item] == item_value].shape[0]
            except:
                smime_sent_enc_count_client = -1
            try:
                smime_received_sign_count_client = \
                received_smime_sig_mails.loc[received_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_received_sign_count_client = -1
            try:
                smime_received_enc_count_client = received_smime_enc_mails.loc[received_smime_enc_mails[item] == item_value].shape[
                0]
            except:
                smime_received_enc_count_client = -1

            try:
                pgp_sent_sign_count_client = sent_pgp_sig_mails.loc[sent_pgp_sig_mails[item] == item_value].shape[0]
            except:
                pgp_sent_sign_count_client = -1
            try:
                pgp_sent_enc_count_client = sent_pgp_enc_mails.loc[sent_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_sent_enc_count_client = -1
            try:
                pgp_received_sign_count_client = received_pgp_sig_mails.loc[received_pgp_sig_mails[item] == item_value].shape[0]
            except:
                pgp_received_sign_count_client = -1
            try:
                pgp_received_enc_count_client = received_pgp_enc_mails.loc[received_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_received_enc_count_client = -1

            sent_smime_client = smime_sent_sign_count_client + smime_sent_enc_count_client
            sent_pgp_client = pgp_sent_sign_count_client + pgp_sent_enc_count_client
            sent_crypto_client = sent_smime_client + sent_pgp_client

            add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentSMIME', f'{sent_smime_client:,}')
            add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentPGP', f'{sent_pgp_client:,}')
            add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentCrypto', f'{sent_crypto_client:,}')

            try:
                add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentSMIMEPercent', f'{sent_smime_client/(smime_sent_sign_count+smime_sent_enc_count)*100:.2f}')
            except:
                add_tex_var(f'Client-{item_value.replace("[d]", "").strip()}-SentSMIMEPercent', f'{0:.2f}')
                
            try:
                add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentPGPPercent', f'{sent_pgp_client/(pgp_sent_sign_count+pgp_sent_enc_count)*100:.2f}')
            except:
                add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentPGPPercent', f'{0:.2f}')

            try:
                add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentCryptoPercent', f'{sent_crypto_client/(sent_sig_mails.shape[0]+sent_enc_mails.shape[0])*100:.2f}')
            except:
                add_tex_var(f'Client-{item_value.replace("[d]","").strip()}-SentCryptoPercent', f'{0:.2f}')


            # Print table line
            add_table_result(table_line(item_value, sent_count_client, received_count_client,
                                        smime_sent_sign_count_client, smime_sent_enc_count_client,
                                        smime_received_sign_count_client, smime_received_enc_count_client,
                                        pgp_sent_sign_count_client, pgp_sent_enc_count_client,
                                        pgp_received_sign_count_client,
                                        pgp_received_enc_count_client))


    add_table_result(r"\\[-0.2cm]")
    add_table_result(r"\multicolumn{2}{l}{\textbf{Operating System}}\\")
    item = "operating_system"
    for os in operating_systems.itertuples():
        item_value = os[2]
        if item_value is not None:
            sent_count = sent_mails_df.loc[sent_mails_df[item] == item_value].shape[0]
            received_count = received_mails_df.loc[received_mails_df[item] == item_value].shape[0]
            try:
                smime_sent_sign_count = sent_smime_sig_mails.loc[sent_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_sent_sign_count = -1
            try:
                smime_sent_enc_count = sent_smime_enc_mails.loc[sent_smime_enc_mails[item] == item_value].shape[0]
            except:
                smime_sent_enc_count = -1
            try:
                smime_received_sign_count = \
                    received_smime_sig_mails.loc[received_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_received_sign_count = -1
            try:
                smime_received_enc_count = \
                received_smime_enc_mails.loc[received_smime_enc_mails[item] == item_value].shape[
                    0]
            except:
                smime_received_enc_count = -1

            try:
                pgp_sent_sign_count = sent_pgp_sig_mails.loc[sent_pgp_sig_mails[item] == item_value].shape[0]
            except:
                pgp_sent_sign_count = -1
            try:
                pgp_sent_enc_count = sent_pgp_enc_mails.loc[sent_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_sent_enc_count = -1
            try:
                pgp_received_sign_count = received_pgp_sig_mails.loc[received_pgp_sig_mails[item] == item_value].shape[
                    0]
            except:
                pgp_received_sign_count = -1
            try:
                pgp_received_enc_count = received_pgp_enc_mails.loc[received_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_received_enc_count = -1

            # Print table line
            add_table_result(table_line(item_value.replace("Win", "Windows"), sent_count, received_count, smime_sent_sign_count, smime_sent_enc_count,
                       smime_received_sign_count,
                       smime_received_enc_count, pgp_sent_sign_count, pgp_sent_enc_count, pgp_received_sign_count,
                       pgp_received_enc_count))

    item = "usergroup"
    add_table_result(r"\\[-0.2cm]")
    add_table_result(r"\multicolumn{2}{l}{\textbf{Usergroup}}\\")
    for group in usergroup.itertuples():
        item_value = group[2]
        if item_value is not None:
            sent_count = sent_mails_df.loc[sent_mails_df[item] == item_value].shape[0]
            received_count = received_mails_df.loc[received_mails_df[item] == item_value].shape[0]
            try:
                smime_sent_sign_count = sent_smime_sig_mails.loc[sent_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_sent_sign_count = -1
            try:
                smime_sent_enc_count = sent_smime_enc_mails.loc[sent_smime_enc_mails[item] == item_value].shape[0]
            except:
                smime_sent_enc_count = -1
            try:
                smime_received_sign_count = \
                    received_smime_sig_mails.loc[received_smime_sig_mails[item] == item_value].shape[0]
            except:
                smime_received_sign_count = -1
            try:
                smime_received_enc_count = \
                received_smime_enc_mails.loc[received_smime_enc_mails[item] == item_value].shape[
                    0]
            except:
                smime_received_enc_count = -1

            try:
                pgp_sent_sign_count = sent_pgp_sig_mails.loc[sent_pgp_sig_mails[item] == item_value].shape[0]
            except:
                pgp_sent_sign_count = -1
            try:
                pgp_sent_enc_count = sent_pgp_enc_mails.loc[sent_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_sent_enc_count = -1
            try:
                pgp_received_sign_count = received_pgp_sig_mails.loc[received_pgp_sig_mails[item] == item_value].shape[
                    0]
            except:
                pgp_received_sign_count = -1
            try:
                pgp_received_enc_count = received_pgp_enc_mails.loc[received_pgp_enc_mails[item] == item_value].shape[0]
            except:
                pgp_received_enc_count = -1

            # Print table line
            add_table_result(table_line(usergroup_mapping[item_value], sent_count, received_count, smime_sent_sign_count, smime_sent_enc_count,
                       smime_received_sign_count,
                       smime_received_enc_count, pgp_sent_sign_count, pgp_sent_enc_count, pgp_received_sign_count,
                       pgp_received_enc_count))

            add_tex_result("Table"+usergroup_mapping[item_value].replace(" ", "")+"SMIMESignPercent", f'{smime_sent_sign_count / sent_count * 100:.2f}', "TableUseragentNumbers")
            add_tex_result("Table"+usergroup_mapping[item_value].replace(" ", "")+"PGPSignPercent", f'{pgp_sent_sign_count / sent_count * 100:.2f}', "TableUseragentNumbers")
            add_tex_result("Table"+usergroup_mapping[item_value].replace(" ", "")+"SMIMEEncPercent", f'{smime_sent_enc_count / sent_count * 100:.2f}', "TableUseragentNumbers")
            add_tex_result("Table"+usergroup_mapping[item_value].replace(" ", "")+"PGPEncPercent", f'{pgp_sent_enc_count / sent_count * 100:.2f}', "TableUseragentNumbers")

    add_table_result(r"\\[-0.2cm]")
    all_table_results()

@cache_it("cache_unique_postboxes", is_df=False)
def get_unique_postboxes(data):
    return data["user"].unique().shape[0]

@cache_it("cache_unique_domaincount", is_df=False)
def get_unique_domaincount(data):
    return data["domain"].unique().shape[0]

def get_postboxes_sent_count(sent_mails_df):
    return sent_mails_df["user"].unique().shape[0]

@cache_it("cache_mailheader_date", is_df=False)
def get_broken_mailheader_date(data):
    return data["header_date"].isnull().sum()

@cache_it("cache_missing_useragent_all_mails", reload=True)
def get_missing_useragent(data):
    return data["client_merged"].value_counts()[""]

@cache_it("cache_pgp_misuse_keys", is_df=True, reload=True)
def get_pgp_misuse_keys(data):
    pgp_key_data = data.loc[(data["pgp_key_data_attachment"] != "[]")]
    pgp_key_data = pgp_key_data["pgp_key_data_attachment"].fillna('empty')
    unique_pgp_keys = pgp_key_data.value_counts().reset_index()
    unique_pgp_keys.columns = ["json", "count"]
    return unique_pgp_keys
    #print(unique_pgp_keys.to_json(orient="split"))
    #print(unique_pgp_keys.to_json(orient="records"))
    #return json.loads(unique_pgp_keys.to_json(orient="split"))

def analyse_pgp_misuse_keys(data):
    private = data[data['json'].str.contains('"type": "priv*')]
    public = data[data['json'].str.contains('"type": "pub"', regex=False)]
    sub = data[data['json'].str.contains('"type": "sub"', regex=False)]
    compliance_8 = data[data['json'].str.contains('"compliance": "8"', regex=False)]
    compliance_23 = data[data['json'].str.contains('"compliance": "23"', regex=False)]
    compliance_6001 = data[data['json'].str.contains('"compliance": "6001"', regex=False)]
    print("Priv Keys", 0 if private.empty else private.shape[0])
    print("Pub Keys", 0 if public.empty else public.shape[0])
    print("Sub Keys", 0 if sub.empty else sub.shape[0])

    print("compliance_8", 0 if compliance_8.empty else compliance_8.shape[0])
    print("compliance_23", 0 if compliance_23.empty else compliance_23.shape[0])
    print("compliance_6001", 0 if compliance_6001.empty else compliance_6001.shape[0])

    json_data = data["json"].astype(str)

    compliance1 = json_data.str.replace('.*"compliance":\s*"([0-9]*)".*', r'[f]\1', regex=True)
    compliance1 = compliance1.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    type1 = json_data.str.replace('"type":\s*"([0-9a-zA-Z]*)"', r'[f]\1[e]', regex=True)
    type1 = type1.str.replace('[\S\s.\[\{]*\[f\]([a-zA-Z0-9]*)\[e\].*', r'[f]\1', regex=True)
    type1 = type1.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    algo = json_data.str.replace('.*"length":\s*"([0-9]*)".*"algo":\s*"([0-9a-zA-Z]*)".*', r'[f]\2-\1', regex=True)
    algo = algo.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    print("\n-----\ncompliance\n", compliance1.value_counts().to_dict())
    print("\n-----\ntype\n", type1.value_counts().to_dict())
    print("\n-----\nalgo\n", algo.value_counts().to_dict())
    compliance2 = json_data.str.replace('.*"compliance":\s*"([0-9]*)".*"compliance":\s*"([0-9]*)".*', r'[f]\2', regex=True)
    compliance2 = compliance2.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    type2 = json_data.str.replace('[\S.]*"type":\s*"([0-9a-zA-Z]*)".*"type":\s*"([0-9a-zA-Z]*)".*', r'[f]\2', regex=True)
    type2 = type2.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    algo2 = json_data.str.replace('.*"length":\s*"([0-9]*)".*"algo":\s*"([0-9a-zA-Z]*)".*"length":\s*"([0-9]*)".*"algo":\s*"([0-9a-zA-Z]*)".*', r'[f]\4-\3', regex=True)
    algo2 = algo2.str.replace('^(?!\[f\]).*', '[f]None', regex=True)
    print("\n-----\ncompliance2\n", compliance2.value_counts().to_dict())
    print("\n-----\ntype2\n", type2.value_counts().to_dict())
    print("\n-----\nalgo2\n", algo2.value_counts().to_dict())



@cache_it("cache_pgp_misuse_autocrypt", is_df=True, reload=True)
def get_pgp_misuse_autocrypt(data):
    autocrypt_pgp = data.loc[(data["autocrypt_data"] != "empty")]
    autocrypt_pgp = autocrypt_pgp["autocrypt_data"].fillna('empty')
    unique_pgp_keys = autocrypt_pgp.value_counts().reset_index()
    unique_pgp_keys.columns = ["json", "count"]
    return unique_pgp_keys

@cache_it("cache_internal_mail", is_df=False)
def get_internal_mail(data):
    #Remove all mails that contain an external mail
    internal_data = data[~((data["sender"].str.contains("@ext")) | (data['receiver_list'].str.contains('@ext')) | (data['cc_list'].str.contains('@ext')) | (data['bcc_list'].str.contains('@ext')))]
    #Remove mails where receiver or sender is empty
    internal_data = internal_data[~((internal_data["sender"].str.contains("empty")) | (internal_data['receiver_list'].str.contains('empty')))]
    return internal_data.shape[0]

@cache_it("cache_parsing_errors_smime", reload = True)
def parse_smime_error(data):
    return data['smime_error'].value_counts()

@cache_it("cache_parsing_errors_pgp", reload = True)
def parse_pgp_error(data):
    return data['pgp_error'].value_counts()

@cache_it("cache_parsing_errors_other", reload = True)
def parse_other_error(data):
    return data['other_error'].value_counts()

@cache_it("chris_smime_errors_by_year", reload = True)
def parse_smime_error_by_year(data):
    return data

@cache_it("chris_smime_errors_by_year_client", reload = True)
def parse_smime_error_by_year_client(data):
    return data

@cache_it("chris_smime_errors_by_year_user", reload = True)
def parse_smime_error_by_year_user(data):
    return data

@cache_it("chris_smime_errors_by_user", reload = True)
def parse_smime_error_by_user(data):
    return data

@cache_it("chris_pgp_errors_by_year", reload = True)
def parse_pgp_error_by_year(data):
    return data

@cache_it("chris_pgp_errors_by_year_user", reload = True)
def parse_pgp_error_by_year_user(data):
    return data

@cache_it("chris_pgp_errors_by_user", reload = True)
def parse_pgp_error_by_user(data):
    return data

@cache_it("chris_pgp_errors_by_user_client", reload = True)
def parse_pgp_error_by_user_client(data):
    return data

@cache_it("chris_pgp_errors_by_year_client", reload = True)
def parse_pgp_error_by_year_client(data):
    return data

@cache_it("chris_other_errors_by_year", reload = True)
def parse_other_error_by_year(data):
    return data

@cache_it("chris_other_errors_by_year_client", reload = True)
def parse_other_error_by_year_client(data):
    return data

@cache_it("chris_other_errors_by_error_year_client", reload = True)
def parse_other_error_by_error_year_client(data):
    return data

@cache_it("chris_other_errors_by_error_year", reload = True)
def parse_other_error_by_error_year(data):
    return data

def parse_errors(data):
    mailsTotal = data.shape[0]

    smime_error_mails = data.loc[data["smime_error"] != "empty"]
    pgp_error_mails = data.loc[data["pgp_error"] != "empty"]
    other_error_mails = data.loc[data["other_error"] != "empty"]

    #Merge mailadress parsing errors
    other_error_mails["other_error"] = other_error_mails["other_error"].str.replace('pseudo-method-mail|pseudo-method-mail|pseudo-method-mail', 'pseudo-method-mail', regex=False)
    other_error_mails["other_error"] = other_error_mails["other_error"].str.replace('pseudo-method-mail|pseudo-method-mail', 'pseudo-method-mail', regex=False)

    smime_error_mails = data_support.replace_useragents(smime_error_mails)
    pgp_error_mails = data_support.replace_useragents(pgp_error_mails)
    other_error_mails = data_support.replace_useragents(other_error_mails)

    print("------------PARSING-ERRORS-START--------------")
    print(parse_smime_error_by_year(smime_error_mails.groupby(['header_year']).agg({'smime_error': ['count']})))
    print(parse_smime_error_by_user(smime_error_mails.groupby(['sender']).agg({'smime_error': ['count']})))
    print(parse_smime_error_by_year_user(smime_error_mails.groupby(['header_year', 'sender']).agg({'smime_error': ['count']})))
    print(parse_pgp_error_by_year(pgp_error_mails.groupby(['header_year']).agg({'pgp_error': ['count']})))
    print(parse_pgp_error_by_year_user(pgp_error_mails.groupby(['header_year', 'sender']).agg({'pgp_error': ['count']})))
    print(parse_pgp_error_by_user(pgp_error_mails.groupby(['sender']).agg({'pgp_error': ['count']})))
    print(parse_pgp_error_by_user_client(pgp_error_mails.groupby(['sender', 'client_merged']).agg({'pgp_error': ['count']})))
    print(parse_other_error_by_year(other_error_mails.groupby(['header_year']).agg({'other_error': ['count']})))

    print(parse_smime_error_by_year_client(smime_error_mails.groupby(['header_year', 'client_merged']).agg({'smime_error': ['count']})))
    print(parse_pgp_error_by_year_client(pgp_error_mails.groupby(['header_year', 'client_merged']).agg({'pgp_error': ['count']})))
    print(parse_other_error_by_year_client(other_error_mails.groupby(['header_year', 'client_merged']).agg({'other_error': ['count']})))

    print(parse_other_error_by_error_year_client(other_error_mails.groupby(['other_error', 'header_year', 'client_merged']).agg({'other_error': ['count']})))
    print(parse_other_error_by_error_year(other_error_mails.groupby(['other_error', 'header_year']).agg({'other_error': ['count']})))
    print("------------PARSING-ERRORS-END--------------")


    errors = 0

    smime_parsing_error = parse_smime_error(data)
    smime_parsing_error.columns = ['errorcode', 'count']
    if smime_parsing_error.shape[0] > 0:
        smime_parsing_error = smime_parsing_error.loc[smime_parsing_error["errorcode"] != "empty"]
        errors += smime_parsing_error["count"].sum()
        add_tex_result("SmimeDecodingErrors", f'{smime_parsing_error["count"].sum():,}', "errors")
        add_tex_result("SmimeDecodingErrorsPercent", f'{(smime_parsing_error["count"].sum() / mailsTotal * 100):.2f}\%', "errors")

    pgp_parsing_error = parse_pgp_error(data)
    pgp_parsing_error.columns = ['errorcode', 'count']
    if pgp_parsing_error.shape[0] > 0:
        print('PGP Parsing Error:', pgp_parsing_error)
        pgp_parsing_error = pgp_parsing_error.loc[pgp_parsing_error["errorcode"] != "empty"]
        errors += pgp_parsing_error["count"].sum()
        add_tex_result("PGPDecodingErrors", f'{pgp_parsing_error["count"].sum():,}', "errors")
        add_tex_result("PGPDecodingErrorsPercent", f'{(pgp_parsing_error["count"].sum() / mailsTotal * 100):.2f}\%', "errors")

    other_parsing_error = parse_other_error(data)
    other_parsing_error.columns = ['errorcode', 'count']
    if other_parsing_error.shape[0] > 0:
        other_parsing_error = other_parsing_error.to_dict('split')
        errors += other_parsing_error["data"][1][1]
        print('Other Parsing Error:', other_parsing_error)
        add_tex_result("PseudoFailed", f'{other_parsing_error["data"][1][1]:,}', "errors")
        add_tex_result("PseudoFailedPercent", f'{(other_parsing_error["data"][1][1] / mailsTotal * 100):.2f}\%', "errors")
        try:
            priv_found = other_parsing_error["data"][2][1]
            errors += priv_found
        except:
            priv_found = 0
        add_tex_result("PseudoPrivFound", f'{priv_found:,}', "errors")
        add_tex_result("PseudoPrivFoundPercent", f'{(priv_found / mailsTotal * 100):.2f}\%', "errors")


    add_tex_result("ParsingErrorsTotal", f'{errors:,}', "errors")
    add_tex_result("ParsingErrorsTotalPercent", f'{(errors / mailsTotal * 100):.2f}\%', "errors")
    add_tex_result("ParsingSuccessfulTotal", f'{(mailsTotal - errors):,}', "errors")
    add_tex_result("ParsingSuccessfulTotalPercent", f'{((mailsTotal - errors) / mailsTotal * 100):.2f}\%', "errors")


@cache_it("cache_folder_max_user", is_df = False, is_json = True)
def get_max_folder_counts(received_mails_df, sent_mails_df):
    recv_max = received_mails_df["user"].value_counts()[0].item()
    sent_max = sent_mails_df["user"].value_counts()[0].item()

    print(recv_max)
    print(sent_max)

    result = {}
    result["recv"] = recv_max
    result["sent"] = sent_max
    return result

@cache_it("cache_active_users", reload = True)
def get_active_users(sent_mails_df):
    list_of_users = sent_mails_df["user"].unique()
    return pandas.Series(list_of_users)

@cache_it("cache_received_active_users", is_df = False, is_json = True)
def get_active_received(data, sent_mails_df):
    list_of_users = get_active_users(sent_mails_df)["0"]
    data_of_active_users = data.loc[data["user"].isin(set(list_of_users))]
    #print(list_of_users)
    #print(set(list_of_users))
    #print(data_of_active_users)
    received_mails_of_active_users = data_support.get_received_mails(data_of_active_users)
    grouped_data = received_mails_of_active_users.groupby(['user']).agg({'header_date': ['count']}).reset_index()
    grouped_data.columns = ['user', 'count']
    print(grouped_data)
    result = {}
    result["usersToVerify"] = f'{grouped_data.shape[0]:,}'
    result["total"] = f'{grouped_data["count"].sum():,}'
    result["mean"] = f'{grouped_data["count"].mean():,.2f}'
    result["std"] = f'{grouped_data["count"].std():,.2f}'
    result["min"] = f'{grouped_data["count"].min():,}'
    result["max"] = f'{grouped_data["count"].max():,}'
    result["median"] = f'{grouped_data["count"].median():,.2f}'
    print(result)
    return result

@cache_it("cache_folder_counts_sent", is_df = False, is_json = True)
def get_folder_counts(data, sent_mails_df):
    list_of_users = get_active_users(sent_mails_df)["0"]
    data_of_active_users = data.loc[data["user"].isin(set(list_of_users))]
    grouped_by_user_and_folder = data_of_active_users.groupby(['user', 'folder']).agg({'header_date': ['count']}).reset_index()
    grouped_by_user_and_folder.columns = ['user', 'folder', 'count']
    grouped_by_user_and_folder = grouped_by_user_and_folder.set_index(["user", "folder"])
    grouped_by_user_and_folder = grouped_by_user_and_folder.unstack(fill_value=0)  # .stack().reset_index()
    if len(grouped_by_user_and_folder.columns) == 8: #Postbox undefined doesn't appear in dataset
        grouped_by_user_and_folder.columns = ['Postbox.Other', 'Postbox.inb', 'Postbox.inb_sub', 'Postbox.sent',
                                              'Postbox.sent_sub', 'Postbox.spam', 'Postbox.trash',
                                              'Postbox.trash_30d']
    else:
        grouped_by_user_and_folder.columns = ['Postbox.Other', 'Postbox.inb', 'Postbox.inb_sub', 'Postbox.sent',
        'Postbox.sent_sub', 'Postbox.spam', 'Postbox.trash',
        'Postbox.trash_30d', 'Postbox.undefined']

    result = {}
    for folder in list(['Postbox.Other', 'Postbox.inb', 'Postbox.inb_sub', 'Postbox.sent',
     'Postbox.sent_sub', 'Postbox.spam', 'Postbox.trash',
     'Postbox.trash_30d', 'Postbox.undefined']):
        if folder in grouped_by_user_and_folder:
            folder_data = grouped_by_user_and_folder[folder]
            result[folder] = {}
            result[folder]["mean"] = f'{folder_data.mean():,.2f}'
            result[folder]["std"] = f'{folder_data.std():,.2f}'
            result[folder]["min"] = f'{folder_data.min():,}'
            result[folder]["max"] = f'{folder_data.max():,}'
            result[folder]["median"] = f'{folder_data.median():,.2f}'

    return result


#Store all sent PGP related mails, to be called with already reduced to PGP set
@cache_it("chris_cache_pgp_data_sent", is_df=True, reload=True)
def get_pgp_data_sent(data):
    #pgp_key_data = data.loc[((data['folder'].str.contains('Postbox.sent')) | (data['folder'].str.contains('Postbox.sent_sub')))]
    pgp_key_data = data_support.get_sent_mails(data)
    return pgp_key_data

#Store all PGP related mails
@cache_it("chris_cache_pgp_data", is_df=True, reload=True)
def get_pgp_data(data):
    data["autocrypt_data"] = data["autocrypt_data"].fillna('empty')
    pgp_key_data = data.loc[((data["pgp_key_data_attachment"] != "[]") | (data["pgp_signature_merged"] == True) | (data["pgp_encryption_merged"] == True) | (data["autocrypt_data"] != "empty"))]
    get_pgp_data_sent(pgp_key_data)
    return pgp_key_data

#get Mails sent by usergroup.
# Example results:
# sc.h,50064455
# sa.h,20774631
@cache_it("chris_cache_usergroups", is_df=True, reload=True)
def get_usergroup_data(data):
    return data["usergroup"].value_counts()

@cache_it("chris_cache_usergroups_user", is_df=True, reload=True)
def get_usergroup_data_users(data):
    return data.groupby(['usergroup']).agg({'user': ['nunique']})


@cache_it("chris_table_sign_enc_per_year", is_df=True, reload=True)
def get_sign_by_year(data):
    signed_mail = data.loc[data['is_signed'] == True]
    smime_signed_mail = signed_mail.loc[signed_mail["smime_sig_fixed"] == True]
    pgp_signed_mail = signed_mail.loc[signed_mail["pgp_signature_merged"] == True]
    encrypted_mail = data.loc[data['is_encrypted'] == True]
    smime_encrypted_mail = encrypted_mail.loc[encrypted_mail["smime_encryption"] == "SMIMEEnc.enc"]
    pgp_encrypted_mail = encrypted_mail.loc[encrypted_mail["pgp_encryption_merged"] == True]

    #group per year
    mails_per_year = data.groupby(['header_year']).agg({'user': ['count']})
    signed_mails_per_year = signed_mail.groupby(['header_year']).agg({'user': ['count']})
    smime_signed_mails_per_year = smime_signed_mail.groupby(['header_year']).agg({'user': ['count']})
    pgp_signed_mails_per_year = pgp_signed_mail.groupby(['header_year']).agg({'user': ['count']})
    encrypted_mails_per_year = encrypted_mail.groupby(['header_year']).agg({'user': ['count']})
    smime_encrypted_mails_per_year = smime_encrypted_mail.groupby(['header_year']).agg({'user': ['count']})
    pgp_encrypted_mails_per_year = pgp_encrypted_mail.groupby(['header_year']).agg({'user': ['count']})

    #Set column names
    mails_per_year.columns = ["mails"]
    signed_mails_per_year.columns = ["signed_count"]
    smime_signed_mails_per_year.columns = ["smime_signed_count"]
    pgp_signed_mails_per_year.columns = ["pgp_signed_count"]
    encrypted_mails_per_year.columns = ["encrypted_count"]
    smime_encrypted_mails_per_year.columns = ["smime_encrypted_count"]
    pgp_encrypted_mails_per_year.columns = ["pgp_encrypted_count"]
    print("##########################################ABCDEFG########")
    all_data = pandas.concat([mails_per_year, signed_mails_per_year, smime_signed_mails_per_year, pgp_signed_mails_per_year, encrypted_mails_per_year, smime_encrypted_mails_per_year, pgp_encrypted_mails_per_year], axis=1)
    all_data.fillna(0, inplace=True)

    #Calculate percentages
    all_data["signed"] = all_data["signed_count"] / all_data["mails"] * 100
    all_data["smime_signed"] = all_data["smime_signed_count"] / all_data["mails"] * 100
    all_data["pgp_signed"] = all_data["pgp_signed_count"] / all_data["mails"] * 100
    all_data["encrypted"] = all_data["encrypted_count"] / all_data["mails"] * 100
    all_data["smime_encrypted"] = all_data["smime_encrypted_count"] / all_data["mails"] * 100
    all_data["pgp_encrypted"] = all_data["pgp_encrypted_count"] / all_data["mails"] * 100

    #Round to 2 digits
    all_data["signed"] = all_data["signed"].round(2).apply(str)
    all_data["smime_signed"] = all_data["smime_signed"].round(2).apply(str)
    all_data["pgp_signed"] = all_data["pgp_signed"].round(2).apply(str)
    all_data["encrypted"] = all_data["encrypted"].round(2).apply(str)
    all_data["smime_encrypted"] = all_data["smime_encrypted"].round(2).apply(str)
    all_data["pgp_encrypted"] = all_data["pgp_encrypted"].round(2).apply(str)
    reorder = all_data[['mails', 'signed', 'smime_signed', 'pgp_signed', 'encrypted', 'smime_encrypted', 'pgp_encrypted', 'signed_count', 'smime_signed_count', 'pgp_signed_count', 'encrypted_count', 'smime_encrypted_count', 'pgp_encrypted_count']]
    return reorder

def percentile(n):
    def percentile_(x):
        return np.percentile(x, n)
    percentile_.__name__ = '%s' % n
    return percentile_


#@cache_it("chris_debug_user_sender_matching1", is_df=True, reload=True)
#def get_debug_user_mails1(data):
#    print("DEBUG-USER-MAILS")
#    return data.groupby(['user']).agg(
#        {'sender': ['nunique']})

def get_distribution(data):
    grouped_multiple_smime = data.groupby(['folder']).agg(
        {'mail_relay_count': ['count', 'min', 'max', 'mean', 'median', percentile(1), percentile(5), percentile(10), percentile(25)]})
    print(grouped_multiple_smime)
    data["return_path"].fillna('empty', inplace=True)
    data['return_path_set'] = (data["return_path"] == "empty")
    grouped_multiple_smime = data.groupby(['folder', 'return_path_set']).agg(
        {'mail_relay_count': ['count', 'min', 'mean', percentile(1), percentile(5), percentile(25), percentile(50), percentile(75)]})
    print(grouped_multiple_smime)

    grouped_multiple_smime = data.loc[data['mail_relay_count'] == 1].groupby(['folder', 'return_path_set']).agg(
        {'mail_relay_count': ['count', 'min', 'mean', percentile(1), percentile(5), percentile(25), percentile(50), percentile(75)]})
    print(grouped_multiple_smime)

    grouped_multiple_smime = data.loc[data['mail_relay_count'] == 2].groupby(['folder', 'return_path_set']).agg(
        {'mail_relay_count': ['count', 'min', 'mean', percentile(1), percentile(5), percentile(25), percentile(50), percentile(75)]})
    print(grouped_multiple_smime)

    grouped_multiple_smime = data.groupby(['folder', 'is_sent']).agg(
        {'mail_relay_count': ['count', 'min', 'mean', percentile(1), percentile(5), percentile(50), percentile(75)]})
    print(grouped_multiple_smime)

#Don't cache this, saves a few minutes but eats to much RAM during transformation -> sometimes OOM killed
def slim_down():
    data = load_data.load_all_mails(testrun, testrows=1000000)

    mailsTotal = data.shape[0]
    add_tex_result("mailsTotal", f'{mailsTotal:,}')

    parse_errors(data)

    data = load_data.drop_error_mails(data)

    data = data_support.replace_useragents(data)
    data = data_support.replace_useragents_top5(data)
    print("Useragents identified")
    data = data_support.replace_operating_system(data)

    # print(data[["cert_data","smime_signature", 'smime_sig_fixed', "smime_level"]].loc[data["smime_signature"].isin(["SMIMESig.sig", "SMIMESig.unknown"])])
    # print(data[["cert_data","smime_signature", 'smime_sig_fixed', "smime_level"]].loc[(data["smime_signature"].isin(["SMIMESig.sig", "SMIMESig.unknown"]) & data["smime_level"] <= 1)])
    # print(data[["cert_data","smime_signature", 'smime_sig_fixed', "smime_level"]].loc[data["smime_sig_fixed"] == True])

    # print("\n".join(data["operating_system"].unique()))

    print("OS identified")
    data['pgp_signature_merged'] = (
                (data["pgp_sig_attachment_fixed"] == True) | (data["pgp_signature_inline"] != "PGPSig.un"))
    data['pgp_encryption_merged'] = (
                (data["pgp_enc_attachment_fixed"] == True) | (data["pgp_encryption_inline"] != "PGPEnc.un"))
    data['pgp_sign_or_encrypt'] = (
                (data["pgp_signature_merged"] == True) | (data["pgp_encryption_merged"] != True))
    print("PGP Signature+Encryption merged")
    data['is_signed'] = (
            (data["pgp_signature_merged"] == True) | (data["smime_sig_fixed"] == True))
    data['is_encrypted'] = (
                (data["pgp_encryption_merged"] == True) | (data["smime_encryption"] == "SMIMEEnc.enc"))
    print(get_sign_by_year(data))
    usergroup = data['user'].astype(str)
    usergroup = usergroup.str.split('@', expand=True)[1]
    #Merge unclear and unknown internal with NX internal
    usergroup = usergroup.str.replace('^u.h', 'NX.h', regex=True)
    usergroup = usergroup.str.replace('^unc.h', 'NX.h', regex=True)
    data["usergroup"] = usergroup
    print("Usergroup splitted")
    #data.drop(columns=['ID', 'cc_list', 'bcc_list', 'reply_to', 'return_path' , 'dkim', 'pgp_inline_forward'], inplace = True)
    return data

def main():
    #data = read_csv("../../data/")
    #store_all_mails(data)

    #Reread files to ensure we catch empty domains and mailboxes for total counts, since they are removed from the dataframe otherwise
    postboxes_total, domains_total = get_domain_and_mailcount("../../data/")
    add_tex_result("postboxesTotal", f'{postboxes_total:,}')
    add_tex_result("domainsTotal", f'{domains_total:,}')

    data = slim_down()

    #Mailcounts by usergroup
    usergroup_data = get_usergroup_data(data)
    total_users = 0
    for ugroup in usergroup_data.itertuples():
        total_users += ugroup[2]
    for ugroup in usergroup_data.itertuples():
        add_tex_var(f'Usergroup{usergroup_mapping[ugroup[1]].replace(" ", "")}MailCount', f'{ugroup[2]:,}')
        add_tex_var(f'Usergroup{usergroup_mapping[ugroup[1]].replace(" ", "")}MailPercent', f'{ugroup[2]/total_users*100:.2f}')

    #Users by usergroup
    usergroup_data_users = get_usergroup_data_users(data)
    total_users = 0
    print(usergroup_data_users)
    for ugroup in usergroup_data_users.itertuples():
        print('Usergroup-debug:', ugroup)
        if ugroup[1] in usergroup_mapping:
            total_users += int(ugroup[2])
    for ugroup in usergroup_data_users.itertuples():
        if ugroup[1] in usergroup_mapping:
            add_tex_var(f'Usergroup{usergroup_mapping[ugroup[1]].replace(" ", "")}UserCount', f'{int(ugroup[2]):,}')
            add_tex_var(f'Usergroup{usergroup_mapping[ugroup[1]].replace(" ", "")}UserPercent', f'{int(ugroup[2])/total_users*100:.2f}')

    get_distribution(data)

    pgp_misuse = get_pgp_misuse_keys(data)
    print("PGP Misuse")
    analyse_pgp_misuse_keys(pgp_misuse)
    autocrypt_misuse = get_pgp_misuse_autocrypt(data)
    print("Autocrypt Misuse")
    analyse_pgp_misuse_keys(autocrypt_misuse)

    sent_mails_df, received_mails_df = split_sent_received_mails(data)

    #Debug some data
    #get_debug_user_mails1(sent_mails_df)

    mailsTotal = data.shape[0]
    add_tex_result("mailsTotalAfterRemoval", f'{mailsTotal:,}')

    crypto_mails_count = data.loc[((data["is_signed"] == True) | (data["is_encrypted"] == True))].shape[0]
    add_tex_result("mailsTotalSignedOrEncrypted", f'{crypto_mails_count:,}')
    add_tex_result("mailsTotalSignedOrEncryptedPercent", f'{crypto_mails_count/mailsTotal*100:.2f}')

    postboxes_count = int(get_unique_postboxes(data))
    domain_count = int(get_unique_domaincount(data))
    postboxes_sent_count = int(get_postboxes_sent_count(sent_mails_df))
    add_tex_result("postboxesWithZeroMails", f'{postboxes_total - postboxes_count:,}')
    add_tex_result("domainsWithZeroMails", f'{domains_total - domain_count:,}')
    add_tex_result("postboxesTotalWithAtleastOneMail", f'{postboxes_count:,}')
    add_tex_result("postboxesTotalWithAtleastOneMailButNoSent", f'{postboxes_count-postboxes_sent_count:,}')
    add_tex_result("inactivePostboxes", f'{(postboxes_total - postboxes_count) + (postboxes_count-postboxes_sent_count):,}')
    add_tex_result("postboxesAtleastOneSentMail", f'{postboxes_sent_count:,}')
    add_tex_result("domainsWithAtleastOneMail", f'{domain_count:,}')
    add_tex_result("postboxesNoSendMail", f'{postboxes_total - postboxes_sent_count:,}')
    add_tex_result("mailsOfPostboxesAtleastOneSentMail", f'{sent_mails_df.shape[0]:,}')

    #Set date for all mails before xxxx and after xxxx to NaT to get numbers for broken dates.
    add_tex_result("mailsWithBrokenDate", f'{int(get_broken_mailheader_date(data)):,}')
    add_tex_result("mailsWithBrokenDatePercent", f'{int(get_broken_mailheader_date(data)) / mailsTotal * 100:.2f}')
    print("General tex commands done")

    completely_internal_mail_count = int(get_internal_mail(data))
    add_tex_result("completelyInternalMail", f'{completely_internal_mail_count:,}')
    add_tex_result("completelyInternalMailPercent", f'{(completely_internal_mail_count/mailsTotal*100):.2f}')

    active_received = get_active_received(data, sent_mails_df)
    for method in ["usersToVerify", "total", "mean", "std", "median", "min", "max"]:
        add_tex_result(f'ActiveUsersReceived{method}', active_received[method])

    mails_per_year(received_mails_df)
    mails_per_year_sent(sent_mails_df)
    #group_message_ids(data)

    #print("First sent mail:")
    #_print_df(data[data['folder'].str.contains('Postbox.sent') | data['folder'].str.contains('Postbox.sent_sub')].groupby("header_date", group_keys=False).agg({'header_date': ['min']}))

    max_user = get_max_folder_counts(received_mails_df, sent_mails_df)
    print(max_user)

    add_tex_result(f'MaxSent', f'{max_user["sent"]:,}', "folder demographics")
    add_tex_result(f'MaxReceived', f'{max_user["recv"]:,}', "folder demographics")

    folder_results = get_folder_counts(data, sent_mails_df)
    folderMapping = {}
    folderMapping["Postbox.inb"] = "Inbox"
    folderMapping["Postbox.inb_sub"] = "InboxSubfolder"
    folderMapping["Postbox.sent"] = "Outbox"
    folderMapping["Postbox.sent_sub"] = "OutboxSubfolder"
    folderMapping["Postbox.trash"] = "Trash"
    folderMapping["Postbox.trash_30d"] = "TrashThirty"
    folderMapping["Postbox.spam"] = "Spam"
    folderMapping["Postbox.Other"] = "Other"
    for folder in folderMapping:
        for method in ["mean", "std", "median", "min", "max"]:
            add_tex_result(f'{method}Folder{folderMapping[folder]}', folder_results[folder][method], "folder demographics")

    #sent_mails_df = sent_mails_df[sent_mails_df['sender'] != "empty"]

    grouped_encryption_useragents(sent_mails_df)
    generate_table_data(data, sent_mails_df, received_mails_df)

    all_tex_results()

    #Creates cache file for analysis_pgp
    if recreate_analysis_pgp:
        get_pgp_data(data)


if __name__ == "__main__":
    main()
