import pandas as pd
import modin.pandas as pandas
import numpy as np
from pathlib import Path
from collections import namedtuple
from statistics import mean
import datetime
import time
import numbers
import re
import os
from pprint import pprint
import helper, data_support, load_data, graph_support
import matplotlib.pyplot as plt
import scipy.stats as st
import seaborn as sns
import functools
import pickle
import json
from matplotlib.lines import Line2D
import flat_table
import ast

# Set to false to recreate numbers. Warning - long runtime. Does not recreate base csv
# Recreates graphs anyway
use_cache = True

#Use to skip loading after 1000 entries, will use different cache than non-testrun
testrun = False

base_path = load_data.base_path

tex_results = {}
table_results = ""

key_data_final = {}

algo_mapping = {}
algo_mapping["1"] = "RSA"#/EncryptOrSign(1)"
algo_mapping["2"] = "RSA"#/EncryptOnly(2)"
algo_mapping["3"] = "RSA"#/SignOnly(3)"
algo_mapping["16"] = "Elgamal"#/EncryptOnly(16)"
algo_mapping["17"] = "DSA"#(17)"
algo_mapping["18"] = "ECDH"#(18)"
algo_mapping["19"] = "ECDSA"#(19)"
algo_mapping["20"] = "Reserved/FormerlyElgamalEncryptOrSign(20)"
algo_mapping["21"] = "Reserved/DH(21)"
algo_mapping["22"] = "EdDSA"#(22)"
algo_mapping["23"] = "Reserved/AEDH(23)"
algo_mapping["24"] = "Reserved/AEDSA(24)"
algo_mapping["100"] = "Private/Experimental(100)"
algo_mapping["101"] = "Private/Experimental(101)"
algo_mapping["102"] = "Private/Experimental(102)"
algo_mapping["103"] = "Private/Experimental(103)"
algo_mapping["104"] = "Private/Experimental(104)"
algo_mapping["105"] = "Private/Experimental(105)"
algo_mapping["106"] = "Private/Experimental(106)"
algo_mapping["107"] = "Private/Experimental(107)"
algo_mapping["108"] = "Private/Experimental(108)"
algo_mapping["109"] = "Private/Experimental(109)"
algo_mapping["110"] = "Private/Experimental(110)"

sym_algo_mapping = {}
sym_algo_mapping["0"] = "Plaintext"
sym_algo_mapping["1"] = "IDEA"
sym_algo_mapping["2"] = "TripleDES"
sym_algo_mapping["3"] = "CAST5"
sym_algo_mapping["4"] = "Blowfish"
sym_algo_mapping["5"] = "Reserved(5)"
sym_algo_mapping["6"] = "Reserved(6)"
sym_algo_mapping["7"] = "AES128(7)"
sym_algo_mapping["8"] = "AES192(8)"
sym_algo_mapping["9"] = "AES256(9)"
sym_algo_mapping["10"] = "Twofish256(10)"
sym_algo_mapping["11"] = "Camelia128(11)"
sym_algo_mapping["12"] = "Camelia192(12)"
sym_algo_mapping["13"] = "Camelia256(13)"
sym_algo_mapping["100"] = "Private/Experimental(100)"
sym_algo_mapping["101"] = "Private/Experimental(101)"
sym_algo_mapping["102"] = "Private/Experimental(102)"
sym_algo_mapping["103"] = "Private/Experimental(103)"
sym_algo_mapping["104"] = "Private/Experimental(104)"
sym_algo_mapping["105"] = "Private/Experimental(105)"
sym_algo_mapping["106"] = "Private/Experimental(106)"
sym_algo_mapping["107"] = "Private/Experimental(107)"
sym_algo_mapping["108"] = "Private/Experimental(108)"
sym_algo_mapping["109"] = "Private/Experimental(109)"
sym_algo_mapping["110"] = "Private/Experimental(110)"

sign_algo_mapping = {}
sign_algo_mapping["1"] = "MD5(1)"
sign_algo_mapping["2"] = "SHA1(2)"
sign_algo_mapping["3"] = "RIPE-MD160(4)"
sign_algo_mapping["4"] = "Reserved(4)"
sign_algo_mapping["5"] = "Reserved(5)"
sign_algo_mapping["6"] = "Reserved(6)"
sign_algo_mapping["7"] = "Reserved(7)"
sign_algo_mapping["8"] = "SHA2-256(8)"
sign_algo_mapping["9"] = "SHA2-384(9)"
sign_algo_mapping["10"] = "SHA2-512(10)"
sign_algo_mapping["11"] = "SHA2-224(11)"
sign_algo_mapping["12"] = "SHA3-256(12)"
sign_algo_mapping["13"] = "Reserved(13)"
sign_algo_mapping["14"] = "SHA3-512(14)"
sign_algo_mapping["100"] = "Private/Experimental(100)"
sign_algo_mapping["101"] = "Private/Experimental(101)"
sign_algo_mapping["102"] = "Private/Experimental(102)"
sign_algo_mapping["103"] = "Private/Experimental(103)"
sign_algo_mapping["104"] = "Private/Experimental(104)"
sign_algo_mapping["105"] = "Private/Experimental(105)"
sign_algo_mapping["106"] = "Private/Experimental(106)"
sign_algo_mapping["107"] = "Private/Experimental(107)"
sign_algo_mapping["108"] = "Private/Experimental(108)"
sign_algo_mapping["109"] = "Private/Experimental(109)"
sign_algo_mapping["110"] = "Private/Experimental(110)"

#columns = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "header_date",
#                "client", "folder",
#                "mail_relay_count", "mail_virus_scans",
#                "message_id", "dkim",
#                "other_error"]

def cache_it(cache_name, is_df = True, reload = False, is_json = False):
    def real_decorator(function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            if is_cached(cache_name):
                if is_df:
                    result = load_from_cache(cache_name)
                else:
                    with open(load_data.get_cache_path(cache_name, testrun), 'r') as fp:
                        result = fp.read()
                        if is_json:
                            result = json.loads(result)
            else:
                result = function(*args, **kwargs)
                if is_df:
                    store_to_cache(cache_name, result)
                    if reload:
                        #Fix for some index problems on some DFs after grouping
                        result = load_from_cache(cache_name)
                else:
                    with open(load_data.get_cache_path(cache_name), 'w') as fp:
                        if is_json:
                            fp.write(json.dumps(result))
                        else:
                            fp.write(str(result))
            return result
        return wrapper
    return real_decorator

def store_to_cache(name, data):
    global testrun
    return load_data.store_to_cache(name, data, testrun)

def is_cached(name):
    global use_cache
    global testrun
    return load_data.is_cached(name, testrun, use_cache)

def load_from_cache(name):
    global testrun
    return load_data.load_from_cache(name, testrun)


@cache_it("chris_cache_pgp_data", is_df=True, reload=True)
def get_pgp_data(data):
    #Throw exception if no cached data is available
    raise ValueError("No cached data")
    return pgp_key_data

def insert_assert_ok(type_top, keyid, type, value):
    '''if type in key_data_final[type_top][keyid]:
        print(type, key_data_final[type_top][keyid][type], value)
    else:
        print(type, value)'''
    assert(type not in key_data_final[type_top][keyid] or key_data_final[type_top][keyid][type] == value)
    key_data_final[type_top][keyid][type] = value

'''def insert_set(type_top, keyid, type, value, date):
    if type not in key_data_final[type_top][keyid]:
        key_data_final[type_top][keyid][type] = set()
    key_data_final[type_top][keyid][type].add(value)'''

def insert_set(type_top, keyid, type, value, date):
    if type not in key_data_final[type_top][keyid]:
        key_data_final[type_top][keyid][type] = {}
        key_data_final[type_top][keyid][type+"_set"] = set()

    if value not in key_data_final[type_top][keyid][type]:
        key_data_final[type_top][keyid][type][value] = {}
        key_data_final[type_top][keyid][type+"_set"].add(value)
        if date == None:
            key_data_final[type_top][keyid][type][value]["earliest"] = datetime.datetime.min
            key_data_final[type_top][keyid][type][value]["highest"] = datetime.datetime.max
        else:
            key_data_final[type_top][keyid][type][value]["earliest"] = date
            key_data_final[type_top][keyid][type][value]["highest"] = date

    if date != None:
        if date < key_data_final[type_top][keyid][type][value]["earliest"]:
            key_data_final[type_top][keyid][type][value]["earliest"] = date

        if date > key_data_final[type_top][keyid][type][value]["highest"]:
            key_data_final[type_top][keyid][type][value]["highest"] = date

def parse_key_infos(data, type_top = "main", from_field = "autocrypt", autocrypt_state = None, mail_date = None, type_top2="main"):
    #print(data)
    if type_top == "main":
        if "ParserKeyWord.AUTOCRYPT_STATE" in data:
            autocrypt_state = data["ParserKeyWord.AUTOCRYPT_STATE"]

        if "ParserKeyWord.AUTOCRYPT_KEY" in data:
            data = data["ParserKeyWord.AUTOCRYPT_KEY"]
            if len(data) == 0:
                #print("Broken AutoCrypt", mail_date)
                #print("[Info] Empty autocrypt key data")
                return 1
            else:
                pass
        elif type_top2=="main":
            data = data[0]
            if len(data) == 0:
                #print("[Info] Empty key data")
                return 2
            else:
                pass

    for keydata in data:
        if isinstance(keydata, dict):
            key = keydata
        else:
            key = data[keydata]
        keyid = key["keyid"]
        ktype = key["type"] if "type" in key else None
        length = key["length"] if "length" in key else None
        algo = key["algo"] if "algo" in key else None
        cap = key["cap"] if "cap" in key else None
        compliance = key["compliance"] if "compliance" in key else None
        curve = key["curve"] if "curve" in key else None
        date = key["date"] if "date" in key else None
        subkeys = key["subkey_info"] if "subkey_info" in key else None
        uids = key["uids"] if "uids" in key else None
        expires = key["expires"] if "expires" in key else None

        algo_merged = ""

        if algo in algo_mapping:
            algo_merged = algo_mapping[algo]
        else:
            algo_merged = f"Undefined({algo})"
        if length is not None:
            algo_merged += f"-{length}"
        if curve is not None:
            algo_merged += f"-{curve}"

        if keyid not in key_data_final[type_top]:
            key_data_final[type_top][keyid] = {}
            key_data_final[type_top][keyid]["autocrypt"] = 0
            key_data_final[type_top][keyid]["key_data_attachment"] = 0
        key_data_final[type_top][keyid]["is_"+type_top2] = True
        key_data_final[type_top][keyid][from_field] += 1
        insert_set(type_top, keyid, "type", ktype, mail_date)
        insert_assert_ok(type_top, keyid, "length", length)
        insert_assert_ok(type_top, keyid, "algo", algo)
        insert_assert_ok(type_top, keyid, "compliance", compliance)
        insert_set(type_top, keyid, "cap", cap, mail_date)
        insert_assert_ok(type_top, keyid, "curve", curve)
        insert_assert_ok(type_top, keyid, "date", date)
        insert_set(type_top, keyid, "uids", uids, mail_date)
        insert_set(type_top, keyid, "expires", expires, mail_date)
        insert_set(type_top, keyid, "autocrypt_state", autocrypt_state, mail_date)
        insert_assert_ok(type_top, keyid, "algo_merged", algo_merged)
        if subkeys is not None:
            for key in subkeys:
                insert_set(type_top, keyid, "subkeys", key, mail_date)
            parse_key_infos(subkeys, type_top = "main", from_field=from_field, autocrypt_state=autocrypt_state, mail_date = mail_date, type_top2 = "sub")
    return 0



def insert_count(type_top, keyid, type, value, date):
    if type not in key_data_final[type_top][keyid]:
        key_data_final[type_top][keyid][type] = {}
        key_data_final[type_top][keyid][type+"_set"] = set()

    if value not in key_data_final[type_top][keyid][type]:
        key_data_final[type_top][keyid][type][value] = {}
        key_data_final[type_top][keyid][type+"_set"].add(value)
        if date == None:
            key_data_final[type_top][keyid][type][value]["count"] = 0
            key_data_final[type_top][keyid][type][value]["earliest"] = datetime.datetime.min
            key_data_final[type_top][keyid][type][value]["highest"] = datetime.datetime.max
        else:
            key_data_final[type_top][keyid][type][value]["count"] = 0
            key_data_final[type_top][keyid][type][value]["earliest"] = date
            key_data_final[type_top][keyid][type][value]["highest"] = date

    if date != None:
        key_data_final[type_top][keyid][type][value]["count"] += 1
        if date < key_data_final[type_top][keyid][type][value]["earliest"]:
            key_data_final[type_top][keyid][type][value]["earliest"] = date

        if date > key_data_final[type_top][keyid][type][value]["highest"]:
            key_data_final[type_top][keyid][type][value]["highest"] = date

        year = date.year
        if type not in key_data_final[type_top+"_year"]:
            key_data_final[type_top+"_year"][type] = {}
        if year not in key_data_final[type_top+"_year"][type]:
            key_data_final[type_top+"_year"][type][year] = {}
        if value not in key_data_final[type_top+"_year"][type][year]:
            key_data_final[type_top + "_year"][type][year][value] = 0
        key_data_final[type_top+"_year"][type][year][value] += 1

def add_sign_data(data, mail_date = None, type_top = "sig"):
    keyid = data["keyid"].strip()
    if keyid == "-1":
        return 2
    try:
        algo = sign_algo_mapping[data["digest algo "].strip()]
    except:
        try:
            algo = f'Unknown({data["digest algo "].strip()})'
        except:
            algo = f'Unknown(NoData)'
    version = data["version"].strip()
    if "sigclass" in data:
        sigclass = data["sigclass"].strip()
    else:
        sigclass = "empty"
    if keyid in key_data_final[type_top]:
        insert_count(type_top, keyid, "sig-algo", algo, mail_date)
        insert_count(type_top, keyid, "sig-version", version, mail_date)
        insert_count(type_top, keyid, "sig-class", sigclass, mail_date)
    else:
        key_data_final[type_top][keyid] = {}
        insert_count(type_top, keyid, "sig-algo", algo, mail_date)
        insert_count(type_top, keyid, "sig-version", version, mail_date)
        insert_count(type_top, keyid, "sig-class", sigclass, mail_date)
        return 1
    return 0

def add_enc_data(data, mail_date = None, type_top = "enc"):
    #print(data, mail_date, type_top)
    keyid = data["keyid"].strip()
    if keyid == "-1":
        return 2
    try:
        algo = sym_algo_mapping[data["algo"].strip()]
    except:
        algo = f'Unknown({data["algo"].strip()})'
    version = data["version"].strip()
    if keyid in key_data_final[type_top]:
        insert_count(type_top, keyid, "enc-algo", algo, mail_date)
        insert_count(type_top, keyid, "enc-version", version, mail_date)
    else:
        key_data_final[type_top][keyid] = {}
        insert_count(type_top, keyid, "enc-algo", algo, mail_date)
        insert_count(type_top, keyid, "enc-version", version, mail_date)
        return 1
    return 0

def generate_plot_algo(df, filename):
    plt.clf()
    plt.figure(figsize=(6, 3))
    sns.set(rc={'figure.figsize': (6, 3)})
    sns.set_theme()
    sns.set_style("white")
    sns.set_style("ticks")
    df.columns = ["year", "Algorithm", "count"]
    print(df.columns, df)
    print(df.unstack(fill_value=0))

    colors1 = sns.color_palette()
    colors2 = sns.color_palette("muted")
    colors3 = sns.color_palette("pastel")
    print(colors1)
    print(colors2)
    print(colors3)

    df.unstack(fill_value=0).plot(legend=True, xlim=(2000,2020), xticks=[2000, 2005, 2010, 2015, 2020], kind="area", y=["DSA", "EdDSA", "RSA-2048", "RSA-3072", "RSA-4096"], color=[colors1[0], colors1[1], colors1[2], colors2[2], colors3[2]])
    L = plt.legend(title=None)
    #L.get_texts()[0].set_text('Mails')
    #L.get_texts()[1].set_text('PGP Encrypted')
    #L.get_texts()[2].set_text('S/MIME Encrypted')
    #L.get_texts()[3].set_text('PGP Signed')
    #L.get_texts()[4].set_text('S/MIME Signed')
    plt.xlabel("Year")
    plt.ylabel("# PGP Keys Created Per Year")
    sns.despine()
    plt.savefig(load_data.get_cache_path('graphs/pgp_algos_used'+filename+'.png', testrun))

    #graph_support.seaborn_stackplot_paper('sns_pgp_algos_used'+filename+'.png', data = df, x_var = "year", y_var = "count", hue = "Algorithm", hue_order = ["DSA", "RSA-2048", "RSA-3072", "RSA-4096", "EdDSA"], xlabel = "Year", ylabel = "# Keys Created Per Year", testrun = testrun)

def generate_plot_hash(df, filename):
    plt.clf()
    plt.figure(figsize=(9, 6))
    print(df.unstack(fill_value=0))
    df.unstack(fill_value=0).plot(legend=True, xlim=(1997,2020), kind="area")
    #df.unstack(fill_value=0).plot.line(legend=True, xlim=(1997,2020))
    plt.xlabel("Year")
    plt.ylabel("# Signatures")
    plt.savefig(load_data.get_cache_path('graphs/pgp_hash_used'+filename+'.png', testrun))

def main():
    key_data = get_pgp_data(None)
    key_data["header_date"] = pandas.to_datetime(key_data["header_date"] , format="%Y-%m-%d", errors='coerce')

    autocrypt = {}
    key_data_dict = key_data.to_dict(orient="records")

    key_data_final["main"] = {}
    key_data_final["sig"] = {}
    key_data_final["enc"] = {}
    key_data_final["sig_year"] = {}
    key_data_final["enc_year"] = {}

    result_codes = {}
    result_codes["parse"] = {}
    result_codes["parse"][0] = 0
    result_codes["parse"][1] = 0
    result_codes["parse"][2] = 0
    result_codes["sign"] = {}
    result_codes["sign"][0] = 0
    result_codes["sign"][1] = 0
    result_codes["sign"][2] = 0
    result_codes["enc"] = {}
    result_codes["enc"][0] = 0
    result_codes["enc"][1] = 0
    result_codes["enc"][2] = 0

    senders = set()

    #print(key_data_dict)
    for row in key_data_dict:
        date = row["header_date"]
        senders.add(row["sender"])
        if row["autocrypt_data"] != "empty":
            autocrypt_data = json.loads(row["autocrypt_data"])
            #if("sec" in row["autocrypt_data"]):
            #    print("autocrypt", row)
            result_codes["parse"][parse_key_infos(autocrypt_data, "main", "autocrypt", mail_date = date)] += 1
            #if row["pgp_signature_attachment"] != "PGPSig.un":
            #    print(row)
        if row["pgp_key_data_attachment"] != "[]":
            key_data_attachment = ast.literal_eval(row["pgp_key_data_attachment"])
            #if("sec" in row["pgp_key_data_attachment"]):
                #print("key_attachment", row["pgp_key_data_attachment"], row)
            result_codes["parse"][parse_key_infos(key_data_attachment, "main", "key_data_attachment", mail_date = date)] += 1

    for row in key_data_dict:
        date = row["header_date"]
        if row["pgp_signature_attachment"] != "PGPSig.un":
            signature_attachment = ast.literal_eval(row["pgp_signature_attachment"])
            result_codes["sign"][add_sign_data(signature_attachment, mail_date=date)] += 1
        if row["pgp_encryption_attachment"] != "PGPEnc.un":
            encryption_attachment = ast.literal_eval(row["pgp_encryption_attachment"])
            result_codes["enc"][add_enc_data(encryption_attachment, mail_date=date)] += 1
        if row["pgp_signature_inline"] != "PGPSig.un":
            signature_attachment = ast.literal_eval(row["pgp_signature_inline"])
            result_codes["sign"][add_sign_data(signature_attachment, mail_date=date)] += 1
        if row["pgp_encryption_inline"] != "PGPEnc.un":
            encryption_attachment = ast.literal_eval(row["pgp_encryption_inline"])
            result_codes["enc"][add_enc_data(encryption_attachment, mail_date=date)] += 1

    for key in key_data_final["main"]:
        #print(key)
        key_data_final["main"][key]["type_len"] = len(key_data_final["main"][key]["type"]) if "type" in key_data_final["main"][key] else 0
        key_data_final["main"][key]["cap_len"] = len(key_data_final["main"][key]["cap"]) if "cap" in key_data_final["main"][key] else 0
        key_data_final["main"][key]["uids_len"] = len(key_data_final["main"][key]["uids"]) if "uids" in key_data_final["main"][key] else 0
        key_data_final["main"][key]["expires_len"] = len(key_data_final["main"][key]["expires"]) if "expires" in key_data_final["main"][key] else 0
        key_data_final["main"][key]["autocrypt_state_len"] = len(key_data_final["main"][key]["autocrypt_state"]) if "autocrypt_state" in key_data_final["main"][key] else 0
        key_data_final["main"][key]["subkeys_len"] = len(key_data_final["main"][key]["subkeys"]) if "subkeys" in key_data_final["main"][key] else 0
        run = "main"
        for key_latest in ('cap', 'uids', 'expires', 'autocrypt_state'):
            date = datetime.datetime.min
            value = None
            for cap in key_data_final[run][key][key_latest]:
                current_date = key_data_final[run][key][key_latest][cap]["highest"]
                if current_date >= date:
                    date = current_date
                    value = cap
            key_data_final[run][key][key_latest + "_latest"] = value

            date = datetime.datetime.max
            value = None
            for cap in key_data_final[run][key][key_latest]:
                current_date = key_data_final[run][key][key_latest][cap]["earliest"]
                if current_date <= date:
                    date = current_date
                    value = cap
            key_data_final[run][key][key_latest + "_earliest"] = value


        '''#Odd PGP Subkeys that use PGP primary as a subkey
        pub_sub = set()
        pub_sub.add("pub")
        pub_sub.add("sub")
        if key_data_final["main"][key]["type_set"] == pub_sub:
            print(key)
            '''

    run = "sig"
    for key in key_data_final[run]:
        #print(key)
        key_data_final[run][key]["sig-algo_len"] = len(key_data_final[run][key]["sig-algo"]) if "sig-algo" in key_data_final[run][key] else 0
        key_data_final[run][key]["sig-version_len"] = len(key_data_final[run][key]["sig-version"]) if "sig-version" in key_data_final[run][key] else 0
        key_data_final[run][key]["sig-class_len"] = len(key_data_final[run][key]["sig-class"]) if "sig-class" in key_data_final[run][key] else 0

        for key_latest in ('sig-algo', 'sig-version', 'sig-class'):
            date = datetime.datetime.min
            value = None
            for cap in key_data_final[run][key][key_latest]:
                current_date = key_data_final[run][key][key_latest][cap]["highest"]
                if current_date >= date:
                    date = current_date
                    value = cap
            key_data_final[run][key][key_latest+"_latest"] = value

    run = "enc"
    for key in key_data_final[run]:
        # print(key)
        key_data_final[run][key]["enc-algo_len"] = len(key_data_final[run][key]["enc-algo"]) if "enc-algo" in key_data_final[run][key] else 0
        key_data_final[run][key]["enc-version_len"] = len(key_data_final[run][key]["enc-version"]) if "enc-version" in key_data_final[run][key] else 0

        for key_latest in ('enc-algo', 'enc-version'):
            date = datetime.datetime.min
            value = None
            for cap in key_data_final[run][key][key_latest]:
                current_date = key_data_final[run][key][key_latest][cap]["highest"]
                if current_date >= date:
                    date = current_date
                    value = cap
            key_data_final[run][key][key_latest + "_latest"] = value



    result = pd.DataFrame.from_dict(key_data_final["main"], orient="index")
    result_sig = pd.DataFrame.from_dict(key_data_final["sig"], orient="index")
    result_enc = pd.DataFrame.from_dict(key_data_final["enc"], orient="index")
    print(key_data_final["sig_year"])
    result_sig_year = pd.DataFrame.from_dict(key_data_final["sig_year"]["sig-algo"], orient="index")
    result_enc_year = pd.DataFrame.from_dict(key_data_final["enc_year"]["enc-algo"], orient="index")
    result.to_csv("/home/mailanalysis/results/chris_results_pgp.csv", sep = ',', quotechar = '\'', encoding='utf-8')
    result_sig.to_csv("/home/mailanalysis/results/chris_results_pgp_sig.csv", sep = ',', quotechar = '\'', encoding='utf-8')
    result_enc.to_csv("/home/mailanalysis/results/chris_results_pgp_enc.csv", sep = ',', quotechar = '\'', encoding='utf-8')
    result_sig_year.to_csv("/home/mailanalysis/results/chris_results_pgp_sig_year.csv", sep = ',', quotechar = '\'', encoding='utf-8')
    result_enc_year.to_csv("/home/mailanalysis/results/chris_results_pgp_enc_year.csv", sep = ',', quotechar = '\'', encoding='utf-8')
    print("result_codes", result_codes)
    print(result["type_set"].value_counts())
    print(result["algo_merged"].value_counts())
    print(result_enc["enc-algo_set"].value_counts())
    print(result_enc["enc-version_set"].value_counts())
    print(result_sig["sig-algo_set"].value_counts())
    print(result_sig["sig-version_set"].value_counts())
    print(result_sig["sig-class_set"].value_counts())

    result["year"] = result["date"].str.split(':', expand=True)[1].astype(int)
    result["expires_latest"].fillna('01:2222', inplace=True)
    result["expires_earliest"].fillna('01:2222', inplace=True)
    result["expires_year"] = result["expires_latest"].str.split(':', expand=True)[1].astype(int)

    result_old = result

    for type in ("main", "sub"):
        result = result_old.loc[result_old["is_"+type] == True]
        total_count = result["algo_merged"].shape[0]
        output = []
        algo_odd = {}
        for row in result["algo_merged"].value_counts().reset_index().itertuples():
            algo = row[1]
            count = row[2]
            algo_only = algo.split("-")[0]
            algo_bit = algo.split("-")[1]
            if algo_bit in ("256", "521", "768", "1024", "2048", "3072", "4096", "8192"):
                output.append(load_data.tex_var("PGP-"+type+"-"+algo, count, ","))
                output.append(load_data.tex_var("PGP-"+type+"-"+algo+"-Percent", (count/total_count*100), ".2f"))
            else:
                if (type+"-"+algo_only) not in algo_odd:
                    algo_odd[(type+"-"+algo_only)] = 0
                algo_odd[(type+"-"+algo_only)] += count
        for algo in algo_odd:
            output.append(load_data.tex_var("PGP-"+algo+"-non-standard", algo_odd[algo], ","))
            output.append(load_data.tex_var("PGP-"+algo+"-non-standard-Percent", algo_odd[algo]/total_count*100, ".2f"))



        pd.set_option('display.max_rows', None)

        DSA_only = result.loc[result["algo_merged"].str.slice(0, 3) == "DSA"]
        output.append(load_data.tex_var("PGP-"+type+"-DSA", DSA_only.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-DSA-Percent", DSA_only.shape[0]/total_count*100, ".2f"))
        print("-----------DSA---------")
        print(DSA_only.groupby(["year", "expires_year"], dropna=False).agg(
            {'year': ['count', 'min', 'max', 'mean', 'median']}))

        output.append(load_data.tex_var("PGP-"+type+"-DSA-Until-2010", DSA_only.loc[DSA_only["year"] <= 2010].shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-DSA-Until-2010-Percent", DSA_only.loc[DSA_only["year"] <= 2010].shape[0]/DSA_only.shape[0]*100, ".2f"))

        output.append(load_data.tex_var("PGP-"+type+"-DSA-No-Expire", DSA_only.loc[DSA_only["expires_year"] == 2222].shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-DSA-No-Expire-Percent", DSA_only.loc[DSA_only["expires_year"] == 2222].shape[0]/DSA_only.shape[0]*100, ".2f"))


        RSA_only = result.loc[result["algo_merged"].str.slice(0, 3) == "RSA"]
        output.append(load_data.tex_var("PGP-"+type+"-RSA", RSA_only.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-RSA-Percent", RSA_only.shape[0]/total_count*100, ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-RSA-No-Expire", RSA_only.loc[RSA_only["expires_year"] == 2222].shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-RSA-No-Expire-Percent", RSA_only.loc[RSA_only["expires_year"] == 2222].shape[0]/RSA_only.shape[0]*100, ".2f"))
        print("-----------RSA---------")
        print(RSA_only.groupby(["year", "algo_merged"], dropna=False).agg(
            {'year': ['count', 'min', 'max', 'mean', 'median']}))

        Elgamal_only = result.loc[result["algo_merged"].str.slice(0, 3) == "Elg"]
        output.append(load_data.tex_var("PGP-"+type+"-Elgamal", Elgamal_only.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-Elgamal-Percent", Elgamal_only.shape[0]/total_count*100, ".2f"))
        print("-----------Elg---------")
        print(Elgamal_only.groupby(["year", "expires_year"], dropna=False).agg(
            {'year': ['count', 'min', 'max', 'mean', 'median']}))


        print("-----------EdDSA---------")
        EdDSA_only = result.loc[result["algo_merged"].str.slice(0, 5) == "EdDSA"]
        output.append(load_data.tex_var("PGP-"+type+"-EdDSA", EdDSA_only.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-EdDSA-Percent", EdDSA_only.shape[0]/total_count*100, ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-EdDSA-No-Expire", EdDSA_only.loc[EdDSA_only["expires_year"] == 2222].shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-EdDSA-No-Expire-Percent", EdDSA_only.loc[EdDSA_only["expires_year"] == 2222].shape[0]/EdDSA_only.shape[0]*100, ".2f"))
        print(EdDSA_only.groupby(["year", "expires_year"], dropna=False).agg(
            {'year': ['count', 'min', 'max', 'mean', 'median']}))

        graph_data = result.loc[((result["algo_merged"].str.slice(0, 3) == "DSA") | (result["algo_merged"].str.slice(0, 5) == "EdDSA") | (result["algo_merged"] == "RSA-2048") | (result["algo_merged"] == "RSA-4096") | (result["algo_merged"] == "RSA-3072"))]

        graph_data["algo_merged"] = graph_data["algo_merged"].str.replace('DSA.*', 'DSA', regex=True)
        graph_data["algo_merged"] = graph_data["algo_merged"].str.replace('EdDSA.*', 'EdDSA', regex=True)
        #Remove single outlier with fake date
        print("-----------GRAPH DATA---------")
        graph_data_sum = graph_data.groupby(["year", "algo_merged"]).size()
        print(graph_data_sum)
        generate_plot_algo(graph_data_sum, type)



        expired_only = result.loc[result["expires_year"] <= 2021]
        output.append(load_data.tex_var("PGP-"+type+"-Expired", expired_only.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-Expired-Percent", expired_only.shape[0]/total_count*100, ".2f"))

        more_than_one_expire = result.loc[result["expires_len"] > 1]
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpire", more_than_one_expire.shape[0], ","))
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpire-Percent", more_than_one_expire.shape[0]/total_count*100, ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpireAvg", more_than_one_expire["expires_len"].mean(), ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpireMedian", more_than_one_expire["expires_len"].median(), ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpireStd", more_than_one_expire["expires_len"].std(), ".2f"))
        output.append(load_data.tex_var("PGP-"+type+"-MoreThanOneExpireMax", more_than_one_expire["expires_len"].max(), ","))

        keys_never_expire_never_changed = result.loc[((result["expires_earliest"] == "01:2222") & (result["expires_latest"] == "01:2222"))]
        output.append(load_data.tex_var("PGP-"+type+"-keys_never_expire_never_changed", keys_never_expire_never_changed.shape[0], ","))

        keys_never_expire_but_shortened = result.loc[((result["expires_earliest"] == "01:2222") & (result["expires_latest"] != "01:2222"))]
        output.append(load_data.tex_var("PGP-"+type+"-keys_never_expire_but_shortened", keys_never_expire_but_shortened.shape[0], ","))

        keys_expire_changed_to_never_expire = result.loc[((result["expires_earliest"] != "01:2222") & (result["expires_latest"] == "01:2222"))]
        output.append(load_data.tex_var("PGP-"+type+"-keys_expire_changed_to_never_expire", keys_expire_changed_to_never_expire.shape[0], ","))
        print(keys_expire_changed_to_never_expire["expires"])
        keys_expire_changed_to_never_expire.to_csv("/home/mailanalysis/results/chris_results_pgp_keys_changed_to_never_expire.csv", sep = ',', quotechar = '\'', encoding='utf-8')

        print(load_data.tex_var("PGP-"+type+"-Count", result.shape[0], ","))
        print(load_data.tex_var("PGP-"+type+"-Unlimited-Lifetime", result.loc[(result['expires_latest'] == "01:2222")].shape[0], ","))
        print(load_data.tex_var("PGP-"+type+"-Unlimited-Lifetime-Percent", (result.loc[(result['expires_latest'] == "01:2222")].shape[0]/result.shape[0]*100), ".2f"))
        print(load_data.tex_var("PGP-"+type+"-subkeys", result['subkeys_len'].mean(), ".2f"))
        print(load_data.tex_var("PGP-"+type+"-subkeys-max", result['subkeys_len'].max(), ","))
        result = result.loc[(result['expires_latest'] != "01:2222")]
        result['lifetime'] = data_support.time_distance(result['date'], result['expires_latest'])
        print(load_data.tex_var("PGP-"+type+"-Lifetime-Mean-Years", result['lifetime'].mean() / 365, ".2f"))
        print(load_data.tex_var("PGP-"+type+"-Lifetime-Median-Years", result['lifetime'].median() / 365, ".2f"))
        print(load_data.tex_var("PGP-"+type+"-Lifetime-STD-Years", result['lifetime'].std() / 365, ".2f"))
        print(load_data.tex_var("PGP-"+type+"-Lifetime-Max-Years", result['lifetime'].max() / 365, ".2f"))
        print(load_data.tex_var("PGP-"+type+"-Lifetime-Min-Weeks", result['lifetime'].min() / 7, ".2f"))

        #print(result['date'])
        for line in sorted(output):
            print(line)

    result_sig_year_sorted = result_sig_year.sort_index().fillna(0).reset_index().dropna()
    result_sig_year_sorted["SHA1(2)"] = result_sig_year_sorted["SHA1(2)"].astype(int)
    result_sig_year_sorted["SHA2-256(8)"] = result_sig_year_sorted["SHA2-256(8)"].astype(int)
    result_sig_year_sorted["MD5(1)"] = result_sig_year_sorted["MD5(1)"].astype(int)
    result_sig_year_sorted["SHA2-512(10)"] = result_sig_year_sorted["SHA2-512(10)"].astype(int)
    result_sig_year_sorted["RIPE-MD160(4)"] = result_sig_year_sorted["RIPE-MD160(4)"].astype(int)
    result_sig_year_sorted["Unknown(NoData)"] = result_sig_year_sorted["Unknown(NoData)"].astype(int)
    result_sig_year_sorted["SHA2-224(11)"] = result_sig_year_sorted["SHA2-224(11)"].astype(int)
    #result_sig_year_sorted["SHA2-384(9)"] = result_sig_year_sorted["SHA2-384(9)"].astype(int)
    result_sig_year_sorted["total"] = result_sig_year_sorted["SHA1(2)"] + result_sig_year_sorted["SHA2-256(8)"] \
                                      + result_sig_year_sorted["MD5(1)"] + result_sig_year_sorted["SHA2-512(10)"] \
                                      + result_sig_year_sorted["RIPE-MD160(4)"] +  result_sig_year_sorted["Unknown(NoData)"] \
                                      +  result_sig_year_sorted["SHA2-224(11)"]# +  result_sig_year_sorted["SHA2-384(9)"]

    print(result_sig_year_sorted)
    generate_plot_hash(result_sig_year_sorted, "")
    #print(result_enc_year.sort_index())

    md5_until_2000 = result_sig_year_sorted.loc[result_sig_year_sorted["index"] <= 2000]
    print(load_data.tex_var("PGP-MD5-Before2000", md5_until_2000["MD5(1)"].sum(), ","))
    print(load_data.tex_var("PGP-MD5-Before2000-Percent", md5_until_2000["MD5(1)"].sum() / md5_until_2000["total"].sum()*100, ".2f"))

    sha1_between_2004_2015 = result_sig_year_sorted.loc[((result_sig_year_sorted["index"] >= 2004) & (result_sig_year_sorted["index"] <= 2015))]
    print(load_data.tex_var("PGP-SHA1-Between-2004-2015", sha1_between_2004_2015["SHA1(2)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA1-Between-2004-2015-Percent", sha1_between_2004_2015["SHA1(2)"].sum() / sha1_between_2004_2015["total"].sum()*100, ".2f"))


    sha2_after_2016 = result_sig_year_sorted.loc[((result_sig_year_sorted["index"] >= 2016))]
    print(load_data.tex_var("PGP-SHA2-256-After-2016", sha2_after_2016["SHA2-256(8)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA2-256-After-2016-Percent", sha2_after_2016["SHA2-256(8)"].sum() / sha2_after_2016["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-SHA2-512-After-2016", sha2_after_2016["SHA2-512(10)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA2-512-After-2016-Percent", sha2_after_2016["SHA2-512(10)"].sum() / sha2_after_2016["total"].sum()*100, ".2f"))


    print(load_data.tex_var("PGP-MD5", result_sig_year_sorted["MD5(1)"].sum(), ","))
    print(load_data.tex_var("PGP-MD5-Percent", result_sig_year_sorted["MD5(1)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-SHA1", result_sig_year_sorted["SHA1(2)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA1-Percent", result_sig_year_sorted["SHA1(2)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-SHA2-256", result_sig_year_sorted["SHA2-256(8)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA2-256-Percent", result_sig_year_sorted["SHA2-256(8)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-SHA2-512", result_sig_year_sorted["SHA2-512(10)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA2-512-Percent", result_sig_year_sorted["SHA2-512(10)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    #print(load_data.tex_var("PGP-SHA2-384", result_sig_year_sorted["SHA2-384(9)"].sum(), ","))
    #print(load_data.tex_var("PGP-SHA2-384-Percent", result_sig_year_sorted["SHA2-384(9)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-RIPE-MD160", result_sig_year_sorted["RIPE-MD160(4)"].sum(), ","))
    print(load_data.tex_var("PGP-RIPE-MD160-Percent", result_sig_year_sorted["RIPE-MD160(4)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))
    print(load_data.tex_var("PGP-SHA2-224", result_sig_year_sorted["SHA2-224(11)"].sum(), ","))
    print(load_data.tex_var("PGP-SHA2-224-Percent", result_sig_year_sorted["SHA2-224(11)"].sum() / result_sig_year_sorted["total"].sum()*100, ".2f"))


if __name__ == "__main__":
    main()