import pandas as pandas
import numpy as np
from pathlib import Path
from collections import namedtuple
from statistics import mean
import datetime
import time
import numbers
import re
from pprint import pprint
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
matplotlib.rcParams['hatch.linewidth'] = 0.2  # previous pdf hatch linewidth


def generate_plot(data_x, data_y, title):
    plt.figure(figsize=(9, 6))
    fig, ax = plt.subplots()

    labels = ['2005', '2006', '2007', '2008', '2009', '2010', '2011', '2012', '2013', '2014', '2015', '2016', '2017', '2018', '2019', '2020']

    x = np.arange(len(labels))  # the label locations
    width = 0.35  # the width of the bars

    rects1 = ax.bar(x - width / 2, data_y[0], width, label='G1')
    rects2 = ax.bar(x + width / 2, data_y[1], width, label='G2')
    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel('# Certificates')
    ax.set_xticks(x)
    ax.set_xticklabels(labels)
    ax.legend(loc="upper left")

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = rect.get_height()
            if height > 0:
                ax.annotate('{}'.format(height),
                        xy=(rect.get_x() + rect.get_width() / 2, height),
                        xytext=(0, 3),  # 3 points vertical offset
                        textcoords="offset points",
                        ha='center', va='bottom')

    autolabel(rects1)
    autolabel(rects2)

    plt.xticks(rotation=45)
    plt.title(title)

    fig.tight_layout()

    plt.savefig('graphs/ca_certs.png')

def generate_stacked_plot(data_x, data_y, title):
    plt.figure(figsize=(9, 6))
    plt.grid(axis="y", which="both")

    print(data_y)

    labels = ['2005', '2006', '2007', '2008', '2009', '2010', '2011', '2012', '2013', '2014', '2015', '2016', '2017', '2018', '2019', '2020']

    ind = np.arange(len(labels))  # the x locations for the groups
    width = 0.35  # the width of the bars: can also be len(x) sequence

    p1 = plt.bar(ind, data_y[0], width)
    p2 = plt.bar(ind, data_y[1], width, bottom=data_y[0])
    # Add some text for labels, title and custom x-axis tick labels, etc.
    plt.ylabel('# Certificates')
    plt.xticks(ind, labels)
    plt.legend((p1[0], p2[0]), ('CA G1', 'CA G2'), loc="upper left")

    plt.xticks(rotation=45)

    plt.savefig('graphs/ca_certs_stacked.png')

def generate_stacked_plot_by_type(data_x, data_y, title):
    plt.figure(figsize=(9, 6))
    plt.grid(axis="y", which="both")

    print(data_y)

    labels = ['2005', '2006', '2007', '2008', '2009', '2010', '2011', '2012', '2013', '2014', '2015', '2016', '2017', '2018', '2019', '2020', '2021*']

    ind = np.arange(len(labels))  # the x locations for the groups
    width = 0.35  # the width of the bars: can also be len(x) sequence

    p1 = plt.bar((ind - width / 2) - 0.1*width, data_y[0], width, edgecolor = 'black')
    p2 = plt.bar((ind - width / 2) - 0.1*width, data_y[1], width, bottom=data_y[0], edgecolor = 'black')
    p1color = list(p1[0].get_facecolor())
    p1color[3] = 0.4
    p2color = list(p2[0].get_facecolor())
    p2color[3] = 0.4
    p1u = plt.bar((ind + width / 2) + 0.1*width, data_y[2], width, color = p1color, edgecolor = 'black', hatch = 'xxx')
    p2u = plt.bar((ind + width / 2) + 0.1*width, data_y[3], width, bottom=data_y[2], color = p2color, edgecolor = 'black', hatch = 'xxx')

    #rects1 = ax.bar(x - width / 2, data_y[0], width, label='G1')
    #rects2 = ax.bar(x + width / 2, data_y[1], width, label='G2')
    # Add some text for labels, title and custom x-axis tick labels, etc.
    plt.ylabel('# Certificates')
    plt.xticks(ind, labels)
    plt.legend((p1[0], p2[0], p1u[0], p2u[0]), ('Server - CA G1', 'Server - CA G2', 'User - CA G1', 'User - CA G2'), loc="upper left")

    plt.xticks(rotation=45)

    plt.savefig('graphs/ca_certs_stacked_by_type.png')


read_df = pandas.read_csv("../../ldata/CA_Certs.csv", sep=',')
generate_plot(read_df["year"], [read_df["G1"], read_df["G2"]], "Certificates signed by CA Generation")
generate_stacked_plot(read_df["year"], [read_df["G1"], read_df["G2"]], "Certificates signed by CA Generation")

read_df = pandas.read_csv("../../ldata/CA_CertsByType.csv", sep=',')
generate_stacked_plot_by_type(read_df["year"], [read_df["G1"].loc[read_df["type"] == "S"], read_df["G2"].loc[read_df["type"] == "S"], read_df["G1"].loc[read_df["type"] == "N"], read_df["G2"].loc[read_df["type"] == "N"]], "Certificates signed by CA Generation")