import modin.pandas as pandas
import numpy as np
#import matplotlib.pyplot as plt
from pathlib import Path
#from dataprep.eda import plot as prepplot
from collections import namedtuple
from statistics import mean
import datetime
import time

#Headerdata for CSV Files from 2019-07-email-luis
headers = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "return_path",
                "header_date",
                #"receiver_date",
                "client", "folder",
                "mail_relay_count", "mail_virus_scans",
                #"return_path_equals_sender",
                #"user_equals_to_sender",
                #"incoming_mail",
                "message_id", "dkim",
                "cert_data", "smime_signature", "smime_encryption", "autocrypt_data",
                "pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment",
                "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline",
                "pgp_inline_forward",
                "smime_error", "pgp_error", "other_error"]


def _print_df(df):
    with pandas.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(df)


def read_csv(path='../data/'):
    df = []

    files_read = 0
    
    filenames = Path(path).rglob('*.csv')
    for f in filenames:
        filename = str(f).rsplit("/", 1)[1]
        if(len("ff5c96d866cad16908d4a7cc6484fa2c2e3e41e3.csv") == len(filename)):
            print(filename)
            read_df = pandas.read_csv(f, sep = ',', quotechar = '\'')
            read_df["ID"] = filename
            read_df.fillna("empty", inplace=True)
            #print(read_df)
            df.append(read_df)
            files_read += 1

    filenames = Path(path).rglob('*.csv.gz')
    for f in filenames:
        filename = str(f).rsplit("/", 1)[1]
        print(filename)
        read_df = pandas.read_csv(f, sep = ',', quotechar = '\'', compression='gzip')
        read_df["ID"] = filename
        read_df.fillna("empty", inplace=True)
        #print(read_df)
        df.append(read_df)
        files_read += 1
        if files_read == 5:
            break

    data = pandas.concat(df, ignore_index=True)
    data["header_date"] = pandas.to_datetime(data["header_date"], format="%Y-%m-%d", errors='coerce')
    _parse_df(data)
    print("Read", files_read, "files for data analysis.")
    return data

def _print_avg_min_max(data):
    print(" - Mean:", data.mean())
    print(" - Median:", data.median())
    print(" - Min:", data.min())
    print(" - Max:", data.max())
    print("")

def _parse_df(df):
    #Enhance some data
    df.loc[df['folder'] == 'Postbox.inb', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.inb_sub', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.trash_30d', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.trash', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.spam', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.junk', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.Other', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.sent', 'direction'] = 1
    df.loc[df['folder'] == 'Postbox.sent_sub', 'direction'] = 1

    df['header_year'] = df['header_date'].str[:4]

def get_summaries(data):

    print("Postbox summaries:")

    print("Mails Total:")
    print(data.count())
    print("Mails Per Year:")
    print(data.groupby("header_year").agg({'header_date': ['count', 'min', 'max']}))

    print("First mail:")
    print(data.groupby("header_date").agg({'header_date': ['min']}))


    print("E-Mails in postbox:")
    _print_avg_min_max(data.groupby("ID").count()["user"])

    #folder_data = data.groupby("ID")["folder"].value_counts()
    #_print_df(folder_data)

    print("E-Mails distributed over folders:")
    folder_distribution = data["folder"].value_counts(normalize = True) * 100

    try:
        count_inbox = folder_distribution["Postbox.inb"]
    except:
        count_inbox = 0

    try:
        count_inbox_sub = folder_distribution["Postbox.inb_sub"]
    except:
        count_inbox_sub = 0

    try:
        count_trash_30d = folder_distribution["Postbox.trash_30d"]
    except:
        count_trash_30d = 0

    try:
        count_trash = folder_distribution["Postbox.trash"]
    except:
        count_trash = 0

    try:
        count_spam = folder_distribution["Postbox.spam"]
    except:
        count_spam = 0

    try:
        count_junk = folder_distribution["Postbox.junk"]
    except:
        count_junk = 0

    try:
        count_other = folder_distribution["Postbox.Other"]
    except:
        count_other = 0

    try:
        count_sent = folder_distribution["Postbox.sent"]
    except:
        count_sent = 0

    try:
        count_sent_sub = folder_distribution["Postbox.sent_sub"]
    except:
        count_sent_sub = 0

    try:
        count_other = folder_distribution["Postbox.Other"]
    except:
        count_other = 0

    print(f'Inbox: {float(count_inbox + count_inbox_sub + count_trash_30d + count_trash + count_spam + count_junk + count_other):.2f}%')
    print(f' - No subfolder: {count_inbox:.2f}%')
    print(f' - Subfolder: {count_inbox_sub:.2f}%')
    print(f' - 30dTrash: {count_trash_30d:.2f}%')
    print(f' - Trash: {count_trash:.2f}%')
    print(f' - Spam: {count_spam:.2f}%')
    print(f' - Junk: {count_junk:.2f}%')
    print(f' - Other: {count_other:.2f}%')
    print(f'')
    print(f'Outbox: {count_sent + count_sent_sub:.2f}%')
    print(f' - No subfolder: {count_sent_sub:.2f}%')
    print(f' - Subfolder: {count_sent_sub:.2f}%')
    print(f'')

    signature_data_smime = data["smime_signature"].value_counts(normalize=True) * 100
    signature_data_smime_count = data["smime_signature"].value_counts()
    encryption_data_smime = data["smime_encryption"].value_counts(normalize=True) * 100
    encryption_data_smime_count = data["smime_encryption"].value_counts()
    print(f'SMIME')
    print(f' - signed: {signature_data_smime["SMIMESig.sig"]:.2f}%')
    print(f' - unsigned: {signature_data_smime["SMIMESig.un"]:.2f}%')
    if "SMIMEEnc.enc" in encryption_data_smime:
        print(f' - encrypted: {encryption_data_smime["SMIMEEnc.enc"]:.2f}%')
    else:
        print(f' - encrypted: {0:.2f}%')
    print(f' - unencrypted: {encryption_data_smime["SMIMEEnc.un"]:.2f}%')
    print(f'')
    print("signature_data_smime_count")
    print(signature_data_smime_count)
    print("encryption_data_smime_count")
    print(encryption_data_smime_count)

    signature_data_pgp = data["pgp_signature_attachment"].value_counts(normalize=True) * 100
    signature_data_pgp_count = data["pgp_signature_attachment"].value_counts()
    signature_data_pgp_inline = data["pgp_signature_inline"].value_counts(normalize=True) * 100
    signature_data_pgp_inline_count = data["pgp_signature_inline"].value_counts()
    encryption_data_pgp = data["pgp_encryption_attachment"].value_counts(normalize=True) * 100
    encryption_data_pgp_count = data["pgp_encryption_attachment"].value_counts()
    encryption_data_pgp_inline = data["pgp_encryption_inline"].value_counts(normalize=True) * 100
    encryption_data_pgp_inline_count = data["pgp_encryption_inline"].value_counts()
    autocrypt_pgp = data["autocrypt_data"].fillna('empty').value_counts(normalize=True) * 100
    print(f'PGP')
    print(f' - unsigned: {signature_data_pgp["PGPSig.un"]:.2f}%')
    print(f' - unencrypted: {encryption_data_pgp["PGPEnc.un"]:.2f}%')
    print(f' - signed (attachment): {100-signature_data_pgp["PGPSig.un"]:.2f}%')
    print(f' - signed (inline): {100-signature_data_pgp_inline["PGPSig.un"]:.2f}%')
    print(f' - encrypted (attachment)(: {100-encryption_data_pgp["PGPEnc.un"]:.2f}%')
    print(f' - encrypted (inline): {100-encryption_data_pgp_inline["PGPEnc.un"]:.2f}%')
    print(f' - Autocrypt data available: {100-autocrypt_pgp["empty"]:.2f}%')
    print("signature_data_pgp_count")
    print(signature_data_pgp_count)
    print("signature_data_pgp_inline_count")
    print(signature_data_pgp_inline_count)
    print("encryption_data_pgp_count")
    print(encryption_data_pgp_count)
    print("encryption_data_pgp_inline_count")
    print(encryption_data_pgp_inline_count)


    
def plot_graphs(data):
    plt.figure()
    data.groupby("direction").count()["smime_signature"].plot(x="header_date")
