import modin.pandas as pandas
import numpy as np
from pathlib import Path
from collections import namedtuple
from statistics import mean
import datetime
import time
import numbers
import json
import os
import data_support
import re
import matplotlib.pyplot as plt


#Headerdata for CSV Files from 2019-07-email-luis
headers = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "return_path",
                "header_date",
                "client", "folder",
                "mail_relay_count", "mail_virus_scans",
                "message_id", "dkim",
                "cert_data", "smime_signature", "smime_encryption", "autocrypt_data",
                "pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment",
                "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline",
                "pgp_inline_forward",
                "smime_error", "pgp_error", "other_error"]

base_path = "/home/mailanalysis"

def write_string(string, path):
    f = open(path, "w")
    f.write(string)
    f.close()

def latex_table(data, path, caption, label):
    if len(data) == 0:
        # No data...
        return
    with open(path, 'w') as target:
        target.write("\n%%%% TABLE FOR " f"{path}" " %%%%\n")
        target.write("\\begin{table}\n")
        target.write("\t\\centering\n")
        target.write("\t\\small\n")
        target.write(f"\t\\caption{{{caption}}}\n")
        target.write(f"\t\\label{{{label}}}\n")
        ## Column layout
        columns = 1
        if isinstance(list(data)[0], list):
            columns = len(list(data)[0])
        layout = layout = "l|" + "".join(['c' for i in range(columns)])
        target.write("\t\\begin{tabular}{"+layout+"}\n")
        for key, value in data.items():
            row = key
            if isinstance(value, list):
                for v in value:
                    row += " & " + v
            else:
                row += " & " + str(value)
            row += "\\\\\n"
            target.write("\t\t"+row)
        target.write("\t\\end{tabular}\n")
        target.write("\\end{table}\n")


def set_pandas_display_options() -> None:
    """Set pandas display options."""
    # Ref: https://stackoverflow.com/a/52432757/
    display = pandas.options.display

    display.max_columns = 10
    display.max_rows = 100
    display.max_colwidth = 100
    display.width = None
    display.precision = 2  # set as needed

def _print_avg_min_max(data):
    print(" - Mean:", data.mean())
    print(" - Median:", data.median())
    print(" - Min:", data.min())
    print(" - Max:", data.max())
    print("")

def _parse_df(df):
    #Enhance some data
    df.loc[df['folder'] == 'Postbox.inb', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.inb_sub', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.trash_30d', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.trash', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.spam', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.junk', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.Other', 'direction'] = 0
    df.loc[df['folder'] == 'Postbox.sent', 'direction'] = 1
    df.loc[df['folder'] == 'Postbox.sent_sub', 'direction'] = 1

    df['header_year'] = df['header_date'].str[:4]
    

def filter_boxes_with_crypto(path='../data/'):
    """
        Goes through each file and counts the number of mailboxes with at least one crypto mail.
        A crypto mail contains at least one entry in the fields: "cert_data", "smime_signature", "smime_encryption","autocrypt_data","pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment", "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline", "pgp_inline_forward", "smime_error", "pgp_error", "other_error"
    """
    files_read = 0
    relevant_boxes = 0
    total_boxes = 0
    relevant_users = 0
    
    output_file_name = 'relevantFiles.txt'
    output = open(output_file_name, 'w')
    
    log_file = 'log_filter.txt'
    log = open(log_file, 'w')

    filenames = Path(path).rglob('*.csv')
    for f in filenames:
        (total, relevant) = read_file(f)
        if relevant > 0:
            output.write(str(f)+"\n")
            relevant_users += 1
        if files_read % 100 == 0:
            print(datetime.datetime.now(), "#files: ", files_read, 'total mails:', total_boxes, 'relevant mails:', relevant_boxes, 'relevant users: ', relevant_users)
        if files_read % 1000 == 0:
            log.write(str(datetime.datetime.now())+" #files: "+ str(files_read)+ 'total mails:'+ str(total_boxes)+ 'relevant mails:'+ str(relevant_boxes)+ 'relevant users: '+ str(relevant_users))
        relevant_boxes = relevant_boxes + relevant
        total_boxes = total_boxes + total
        files_read += 1


    filenames = Path(path).rglob('*.csv.gz')
    for f in filenames:
        read_file(f)
        (total, relevant) = read_file(f)
        if relevant > 0:
            output.write(str(f)+"\n")
        if files_read % 100 == 0:
            print(datetime.datetime.now(), "#files: ", files_read, 'total mails:', total_boxes, 'relevant mails:', relevant_boxes)
        if files_read % 1000 == 0:
                        log.write(str(datetime.datetime.now())+" #files: "+ str(files_read)+ 'total mails:'+ str(total_boxes)+ 'relevant mails:'+ str(relevant_boxes)+ 'relevant users: '+ str(relevant_users))
        relevant_boxes = relevant_boxes + relevant
        total_boxes = total_boxes + total
        files_read += 1
        
    output.close()
    print(datetime.datetime.now(), "Read", files_read, "files for data analysis.")
    print('Total mails: ',total_boxes, 'relevant mails:', relevant_boxes, 'relevant users: ', relevant_users)
    log.write(str(datetime.datetime.now())," Read"+ str(files_read)+ "files for data analysis.")
    log.write(str(datetime.datetime.now()), ' Total mails: '+str(total_boxes)+'relevant mails:'+ str(relevant_boxes)+ 'relevant users: '+ str(relevant_users))
    log.write(str(datetime.datetime.now()), ' FINISH')
    log.close()

def read_file(file):
    filename = str(file).rsplit("/", 1)[1]
    read_df = pandas.read_csv(file, sep = ',', quotechar = '\'')
    read_df["ID"] = filename
    total_mails = read_df.shape[0]
    read_df.fillna("empty", inplace=True)
    df = read_df[["user","cert_data", "smime_signature", "smime_encryption","autocrypt_data","pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment", "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline", "pgp_inline_forward", "smime_error", "pgp_error", "other_error"]]
   
    df = filter_crypto_mails(df, remote = remote)
    relevant_mails = df.shape[0]
    return (total_mails, relevant_mails)

def filter_crypto_mails(df):
    """
    A mail is a crypto mail iff it contains a SMIME/PGP signature/encryption part or some certifacte or key.
    TODO: Diskussion about errors and known type
    """
    df = df.loc[(df["smime_signature"] != "SMIMESig.un") | (df["smime_encryption"] != "SMIMEEnc.un") | (df["pgp_signature_attachment"] != "PGPSig.un") | (df["pgp_key_data_attachment"].str.contains("type")) | (df["autocrypt_data"] != "empty") | (df["cert_data"] != "empty") | (df["pgp_signature_inline"] != "PGPSig.un") |(df["pgp_encryption_inline"] != "PGPEnc.un") | (df["pgp_encryption_attachment"] != "PGPEnc.un")]
    return df

  
  
def load_relevant_users(test, columns = None, crypto_only = False):
    df = []
    files_read = 0
    total_mails = []
    crypto_mails = []
    max = 0
    maxMails = 0
    maxName = ""
    start = time.time()
    mergedDF = []
   
    with open('relevantFiles.txt', 'r') as f:
        for line in f:
            line = line.replace('\r', '').replace('\n', '')
            if (columns is None):
                read_df = pandas.read_csv(line, sep = ',', quotechar = '\'', low_memory=False)
            else:
                read_df = pandas.read_csv(line, sep = ',', quotechar = '\'', low_memory=False, usecols=columns)
            read_df["ID"] = line
            read_df.fillna("empty", inplace=True)
            mails = read_df.shape[0]
            total_mails.append(mails)
            if crypto_only:
                read_df = filter_crypto_mails(read_df)
                
            df.append(read_df)
            crypto_df = filter_crypto_mails(read_df)
            crypto_mails.append(crypto_df.shape[0])
            if crypto_df.shape[0] > max:
                max = crypto_df.shape[0]
                maxMails = mails
                maxName = line
            if files_read % 100 == 0:
                print("Current file:", files_read, "total mails:", mails, "crypto mails:", crypto_df.shape[0])
            if files_read % 1000 == 0:
                data = pandas.concat(df, ignore_index=True)
                mergedDF.append(data)
                df = []
            files_read += 1
            if test and files_read > 10:
                print(df)
                break
                
    end = time.time()
    print("Read", files_read, "files for data analysis in time", ((end-start)/60), "min")
    print("Crypto mails: ", sum(crypto_mails), "mean: ", mean(crypto_mails), "total mails:", sum(total_mails), "mean: ", mean(total_mails), "max", max , "with total mails: ", maxMails, "of ", maxName)
    data = pandas.concat(df, ignore_index=True)
    mergedDF.append(data)
    end = time.time()
    print("Read", files_read, "files done!", ((end-start)/60), "min")
    return mergedDF

def store_crypto_mails(test):
    dfs = load_relevant_users(test, crypto_only = True)
    memory = 0
    for df in dfs:
        memory += df.memory_usage(index=True).sum()
    print('Total memory: ', memory)
    data = pandas.concat(dfs, ignore_index=True)
    
    filename = "crypto_only.csv"
    if test:
        filename = "test_" + filename
    filename = "../../../data/" + filename
    
    data.to_csv(filename, sep = ',', quotechar = '\'', encoding='utf-8')

def load_crypto_mails(test, columns = None):
    start = time.time()
    filename = "crypto_only.csv"
    if test:
        filename = "test_" + filename
    filename = "../../../data/" + filename
    print("Read file...", filename)
    
    if (columns is None):
        read_df = pandas.read_csv(filename, sep = ',', quotechar = '\'', low_memory=False)
    else:
        read_df = pandas.read_csv(filename, sep = ',', quotechar = '\'', low_memory=False, usecols=columns)
    end = time.time()
    print("Read", read_df.shape, "mails for data analysis in time", ((end-start)/60), "min")
    return read_df
    
def load_certificates(test):
    start = time.time()
    filename = "certificates.csv"
    if test:
        filename = "test_" + filename
    filename = "../../../data/" + filename
   
    read_df = pandas.read_csv(filename, sep = ',', quotechar = '\'', low_memory=False)
    
    end = time.time()
    print("Read", read_df.shape, "mails for data analysis in time", ((end-start)/60), "min")
    return read_df
  
    
def collect_and_store_certificates(certificates, test):
    # TODO: Add key usage
    new_df = pandas.DataFrame(columns=['serialnumber', 'user issuer', 'user issuer country', 'root ca', 'root country', 'ca chain length', 'sign algo', 'enc algo', 'size', 'creation date', 'expire date', 'lifetime', 'isExpired','certificate'])
    for cert in certificates:
        org = cert
        cert = json.loads(cert)
        cert_record = {}
        length = len(cert)
        if length > 0:
            root_cert = cert[0]
            analysis_root_ca(root_cert, cert_record)
            i = 1
            user_cert = cert[len(cert)-i]
            # Often the last entry is the level, e.g. {'level': 3}.
            while not ('sn' in user_cert):
                i += 1
                user_cert = cert[len(cert)-i]
            length = length - i + 1
            analysis_user_cert(user_cert, cert_record)
        cert_record['ca chain length'] = length
        cert_record['certificate'] = org
        new_df = new_df.append(cert_record, ignore_index=True)
        
    filename = "certificates.csv"
    if test:
        filename = "test_" + filename
    filename = "../../../data/" + filename
    print("Write file to: ", filename)
    new_df.to_csv(filename, sep = ',', quotechar = '\'', encoding='utf-8')

    return new_df

def get_cache_path(name, testrun):
    if testrun:
        return base_path+"/results_test/test_"+name
    else:
        return base_path+"/results/"+name

def store_to_cache(name, data, testrun):
    data.to_csv(get_cache_path(name, testrun), sep = ',', quotechar = '\'')

def is_cached(name, testrun, use_cache):
    if testrun:
        run_type = "[Test] "
    else:
        run_type = "[Prod] "
    if use_cache:
        exists = os.path.isfile(get_cache_path(name, testrun))
        if not exists:
            print(run_type+"Cache miss for "+name+". Recreating.")
        return exists
    else:
        # We're not using cache, so return false in any case.
        print(run_type+"Not using cached files for "+name+". Recreating.")
        return False

def load_from_cache(name, testrun):
    if testrun:
        run_type = "[Test] "
    else:
        run_type = "[Prod] "
    start = time.time()
    df = pandas.read_csv(get_cache_path(name, testrun), sep=',', quotechar='\'', low_memory=True)
    print(run_type+"Loading from cache:", name, "-", ((time.time() - start) / 60), "min")
    return df

def drop_error_mails(data):
    return data.loc[((data['other_error'] != "pseudo") & (data['other_error'] != "pseudo-priv-found"))]

def load_all_mails(test, columns=None, testrows = 200000):
    start = time.time()
    filename = "all_mails_all_fields.csv"
    filename = "../../../data/" + filename

    if test:
        nrows = testrows
    else:
        nrows = None

    if (columns is None):
        read_df = pandas.read_csv(filename, sep=',', quotechar='\'', low_memory=True, nrows=nrows)
    else:
        read_df = pandas.read_csv(filename, sep=',', quotechar='\'', low_memory=True, usecols=columns, nrows=nrows)
    end = time.time()

    
    #Replace header_dates and add years.
   

    print("Read", read_df.shape, "mails for data analysis in time", ((end - start) / 60), "min")

    if columns is None or 'header_date' in columns:
        read_df["header_date"] = pandas.to_datetime(read_df["header_date"], format="%Y-%m-%d", errors='coerce')
        read_df['header_date'] = read_df['header_date'].apply(lambda header_date_: pandas.NaT if header_date_.year < 1994 or header_date_.year > 2021 else header_date_)
        read_df["header_year"] = read_df["header_date"].dt.year
    start = time.time()
    if columns is None:
        read_df = data_support.add_level_info_fix_crypto(read_df)
    print("Fixed crypto in time", ((time.time() - start) / 60), "min")

    start = time.time()
    if columns is None or 'folder' in columns:
        read_df["return_path"].fillna('empty', inplace=True)
        # If a return path is set, then a mailserver has handled this mail -> received
        # If we received a mail_relay_count of 0 then a local client sent it, if it's 1, then Horde Webmail was used -> Adds 1 client
        read_df['sender_equals_user'] = (read_df["user"] == read_df["sender"])
        read_df['is_sent'] = (
                (   # Sent folder -> sent mail
                        (
                                ((read_df["folder"].str.contains("Postbox.sent")) | (read_df["folder"].str.contains("Postbox.sent_sub")))
                                &
                                (read_df["return_path"] == "empty") & (read_df["mail_relay_count"] <= 1)
                        )
                )
                    |
                    # Not inbox and not inbox-subfolder AND return_path == empty AND mail_relay_count <= 1 AND user==sender.
                    # We'll miss some mails sent via alias, but causes to much false positives otherwise
                (
                    ((~read_df["folder"].str.contains("Postbox.inb")) & (~read_df["folder"].str.contains("Postbox.inb_sub")))
                        &
                    (read_df["return_path"] == "empty") & (read_df["mail_relay_count"] <= 1) & (read_df['sender_equals_user'])
                )
        )
    print("Fixed crypto in time", ((time.time() - start) / 60), "min")
    return read_df

under_pat = re.compile(r'_([a-z])')
def underscore_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)
    

def tex_var(name, value, formating=""):
    if formating==",":
        return "\\definevar{"+str(name)+"}{"+f'{value:,}'+"}"
    elif formating==".2f":
        return "\\definevar{"+str(name)+"}{"+f'{value:.2f}'+"}"
    return "\\definevar{"+str(name)+"}{"+f'{value}'+"}"

def tex_command(name, value):
    name = underscore_to_camel(name)
    return "\\newcommand{\\"+str(name)+"}{"+str(value)+"}"

def write_commands(commands, file):
    f = open(file, "w")
    for key, value in commands.items():
        f.write(tex_command(key, value)+"\n")
    f.close()
    
def write_var(commands, file):
    f = open(file, "w")
    for key, value in commands.items():
        f.write(tex_var(key, value)+"\n")
    f.close()

def write_certs(test):
    df = load_crypto_mails(test)
    cert_data = df.loc[(df["cert_data"] != "empty")]
    cert_data = cert_data["cert_data"].fillna('empty')
    unique_certs = pandas.unique(cert_data)
    df_cert = collect_and_store_certificates(unique_certs, test)
    
    
def generate_plot_line(data_x, data_y, title, filename, log = False):
    plt.figure(figsize=(9, 6))
    if isinstance(data_y, pandas.DataFrame):
        for column_name, item in data_y.iteritems():
              plt.plot(data_x, item, label = column_name)
    else:
        plt.plot(data_x, data_y)
    plt.legend(loc='upper left')
    
    if log:
        plt.yscale('log')
    #plt.title(title)
    plt.savefig('/home/mailanalysis/results/graphs/'+filename)

def main():
    overall = input("Run overall data? y/n\n")
    test = True
    if overall == 'y':
        test = False
    df = load_crypto_mails(test)
    years = crypto_per_year(df)
    print("############################")
    senders = crypto_mails_sender_receiver(df)
    print(df.shape[0])


    
def test():
    set_pandas_display_options()
    overall = input("Run overall data? y/n\n")
    test = True
    if overall == 'y':
        test = False
    analyze_certs(test)
    #write_certs(test)


if __name__ == "__main__":
    test()

