# Readme

The analysis scripts consist of several files. Some of the scripts have to be executed in a particular order as some files are created by other scripts.

## Global variables

There are some variables that are available in all scripts:

* use_cache: This variable allows to configure if results are to be cached, set this to false if methods or the data changed. Setting this to true saves a lot of runtime.
* testrun: This variable limits the loading of data to a few million entries. This is useful during development, however, not all analysis work with this, as some occurences require a lot of emails to appear. Caching also works with testrun and doesn’t interfere with the results of the normal run.

## Files

All analysis scripts start with analysis_.
Helper.py, data_support.py and loaddata.py include functions to preprocess the data(e.g. formatting the table) and other helpful functions (e.g. storing or loading files).

### analysis_general_metadata.py

Additional variables:
* recreate_analysis_pgp: Set this to true to store all PGP relevant mails in a separate file. Only needs to be done once.
Calculates:

* General usage of Mail/SMIME/PGP
* MultiClient use

Generates (only paper relevant files):

* paper/generated_chris.tex
* paper/table_demographics.tex
* paper/table_smime_pgp_after_usage.tex
* graphs/paper/sns_boxplot_smime_after_first_use.png

### analysis_pgp.py

Additional variables:

* use_cache: This must be true for this script, as the caching mechanism is used to load the PGP data. Requires a previous run of analysis_general_metadata.py with recreate_analysis_pgp = True.

Calculates the usage of PGP algos and keys over the years.
Generates PGP related graphs.

### analysis_crypto_overview.py

In this file we  analyze crypto usage per year
Generates:

* communication_per_year_table.tex
* communication_per_year.tex
* general_crypto_overview_table.tex
* general_crypto_overview.tex
* sent_emails_per_year.tex
* sent_emails_per_year_table.tex



### analysis_crypto_certificates.py

In this file we analyze S/MIME certificates

Some CAs (e.g. DFN) is grouped by name. This depends strongly on the data set. We recommend revisiting the grouping when using another data set. The functions are Analysis_certificates_issuer and analysis_certificates_root
Generates:

* keyrollover.tex
* numbers_key_table_smime.tex
* numbers_smime_certificates.tex
* numbers_smime_certificatesuniversity_members.tex
* numbers_smime_key_exchange.tex
* numbers_smime_multiple_certificates.tex
* sns_boxplot_key_exchange.png
* hist_key_rollover_with_outlier.png
* hist_key_rollover.png

### graph_usage.py

This script generates graphs/communication_per_year.png
Requires:

* chris_table_sign_enc_per_year generated by analysis_general_metadata.py

## Dependencies

For analyzing the data we use pandas, modin (for multi threading pandas), Numpy,
For plotting we use Matplotlib, seaborn.
