from parser.parser import read_json_file, json_data_to_mails
from parser import Pseudomizer, KeyWords
from pathlib import Path
import bcrypt
import csv
import sys, getopt
import json
import gzip
import shutil
import os
import gc
from records import MailRecord

def main(json_input = "data/fetch_json_new.json", csv_output = "results/results.csv", salt = None):
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hu:d:t:", ["user=", "domain=", "threadnum="])
    except getopt.GetoptError:
        #Run with testdata
        print("No parameters, attempting to run with test data: "+json_input)
        pass

    user = None
    domain = None
    threadnum = 0
    for opt, arg in opts:
        if opt == '-h':
            print('main.py -u <user> -d <domain> -t <threadnum>')
            sys.exit()
        elif opt in ("-u", "--user"):
            user = arg
        elif opt in ("-d", "--domain"):
            domain = arg
        elif opt in ("-t", "--threadnum"):
            threadnum = arg

    if salt is None: #Salt is prefilled for tests, to ensure unit tests are working correctly.
        #Attempt to read salt generated on this system, to ensure that reruns use the same salt.
        try:
            with open("data/salt", 'rb') as saltfile:
                salt = saltfile.read()
        except FileNotFoundError:
            #No salt on system yet, create one
            with open("data/salt", 'wb') as saltfile:
                salt = bcrypt.gensalt()
                saltfile.write(salt)

    if user is None or domain is None:
        data = read_json_file(json_input)
        print("Reading: "+json_input)
        user = "testuser"
        domain = "example.org"
        pytest_run = True
    else:
        pytest_run = False
        csv_output = "results/" + Pseudomizer.hash(domain, salt) + "/" + Pseudomizer.hash(user, salt) + ".csv"
        data = json.loads(sys.stdin.read())

        # Write down hash of mailaddress for later mapping
        Path("results/").mkdir(parents=True, exist_ok=True)
        with open("results/mailmapping"+threadnum+".csv", 'a') as csvfile:
            csvwriter = csv.writer(csvfile, delimiter=',', quotechar='\'')
            csvwriter.writerow([user, Pseudomizer.hash(user, salt)])

    result = json_data_to_mails(data)

    del data
    gc.collect()

    username = user.lower().split("@")[0]

    #Ensure results path exists
    Path(csv_output.rsplit("/",1)[0]).mkdir(parents=True, exist_ok=True)
    with open(csv_output, 'w') as csvfile:
        csvwriter = csv.writer(csvfile, delimiter=',', quotechar='\'')
        if csvfile.tell() == 0:
            csvwriter.writerow(MailRecord.headers)
            pass
        for mail in result:
            try:
                pseudomail = Pseudomizer.pseudomize(mail, salt, username=username, domain=domain)
            except Exception as e:
                print("[main-Pseudo] Exception catched: ", e.__class__.__name__)
                pseudomail = []
                pseudomail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo"))
            finally:
                mail_entry = MailRecord.MailRecord(pseudomail)
                result_row = mail_entry.get_list()
                result_row_for_asserts = "".join([str(elem) for elem in mail_entry.get_list()]).lower()
                if(username not in result_row_for_asserts and domain not in result_row_for_asserts):
                    #Check if the list contains the userdata, only write if this is NOT the case.
                    csvwriter.writerow(result_row)
                else:
                    print("[main-Pseudo] Found private data in result, skipping entry")

                    #Search for the field in which this appeared:
                    i = 0
                    for elem in mail_entry.get_list():
                        str_elem = str(elem)
                        if (str_elem.find(username) != -1 or str_elem.find(domain) != -1):
                            print("[main-Pseudo] The entry was found in the field", MailRecord.headers[i])
                        i+=1

                    pseudomail = []
                    pseudomail.append(KeyWords.ParsedItem(KeyWords.ParsedData.OTHER_ERROR, "pseudo-priv-found"))
                    mail_entry = MailRecord.MailRecord(pseudomail)
                    result_row = mail_entry.get_list()
                    csvwriter.writerow(result_row)





    if not(pytest_run):
        # Not in test mode, compress output
        with open(csv_output, 'rb') as csvfile:
            with gzip.open(csv_output + '.gz', 'wb') as gzip_output:
                shutil.copyfileobj(csvfile, gzip_output)
        #Check if gzipped file exists and delete original.
        try:
            if os.stat(csv_output + '.gz').st_size > 0:
                os.remove(csv_output)
        except FileNotFoundError as e:
            print("[main] Gzipped file not found, not deleting CSV")

    print(len(result))

if __name__ == "__main__":
    main()
