from enum import Enum
from parser import CertHelper

class JsonLevelKeyWord(Enum):
    USER = "user"
    MAILBOX = "mailbox"
    HEADER = "hdr"
    BODY = "body"
    BODY_STRUCTURE = "imap.bodystructure"

class ParserKeyWord(Enum):
    USER = "user"
    MAILBOX = 'mailbox'

    SENDER = "from"
    RETURN_PATH = "return-path"
    TO = "to"
    DELIVERED_TO = "delivered-to"
    CC = "cc"
    BCC = "bcc"
    REPLY_TO = "reply-to"

    RECEIVED = "received"
    MSG_ID = "message-id"
    DKIM = "dkim-signature"
    HEADER_DATE = "date"
    RECEIVED_DATE = "date.received"
    Virus_SCAN = "x-dfn-virus-scanned"

    BODY_STRUCTURE = "imap.bodystructure"
    FLAGS = "flags"
    USER_AGENT = "user-agent"
    XMAILER = "x-mailer"

    BODY = 'body'

    AUTOCRYPT = 'Autocrypt'
    AUTOCRYPT_KEY = 'key'
    AUTOCRYPT_STATE = 'state'

    #Mailinglist headers, according to RFC list-help is the only required field, we'll search for subscribe+unsubscribe aswell
    MAILINGLIST_LIST_HELP = 'list-help'
    MAILINGLIST_LIST_SUBSCRIBE = 'list-subscribe'
    MAILINGLIST_LIST_UNSUBSCRIBE = 'list-unsubscribe'

    #ARC_MESSAGE_SIGNATURE = "ARC-Message-Signature"
    #ARC_SEAL = "ARC-Seal"
    #ARC_AUTHENTICATION_RESULTS = "ARC-Authentication-Results"


class ParsedData(Enum):
    IS_MAILINGLIST = "is_mailinglist"
    CERT_DATA = "cert"
    SMIME_ENCRYPTION = "smime_encryption"
    SMIME_SIGNATURE = "smime_signature"
    PGP_KEY_DATA_ATTACHMENT = "pgp_key_attachment"
    PGP_SIGNATURE_ATTACHMENT = "pgp_signature_attachment"
    PGP_ENCRYPTION_ATTACHMENT = "pgp_encryption_attachment"
    PGP_KEY_DATA_INLINE = "pgp_key_inline"
    PGP_SIGNATURE_INLINE = "pgp_signature_inline"
    PGP_ENCRYPTION_INLINE = "pgp_encryption_inline"
    PGP_INLINE_FORWARD = "pgp_inline_forward"
    SMIME_ERROR = "smime_error"
    PGP_ERROR = "pgp_error"
    OTHER_ERROR = "other_error"

class ParsedItem:
    def __init__(self, key, value, meta = None):
        self.key = key
        if isinstance(value, str):
            self.value = value.strip()
        else:
            self.value = value
        self.meta = meta

    def merge(self):
        if self.meta != None:
            if isinstance(self.value, str):
                self.value = self.value + " | meta: " + str(self.meta)
            elif isinstance(self.value, list):
                append: False
                for item in self.value:
                    if isinstance(item, ParsedItem):
                        item.merge()
                    else:
                        append = True
                        break
                if append:
                    if isinstance(self.value[0], CertHelper.Cert):
                        tempmeta = CertHelper.MetaObject()
                        tempmeta.level = self.meta
                    else:
                        tempmeta = self.meta
                    self.value.append(tempmeta)

            elif isinstance(self.value, dict):
                self.value["level"] = self.meta

    def __repr__(self):
        return "<ParsedItem: %s | %s | %s>" % (self.key, self.value, self.meta)