# Overview
This folder contains the scripts to collect the mail metadata. This script is run on the datacenter side, researchers should **not** get access to raw data or the server that this script is run on.
The script also creates files called "mailmapping{threadnumber}.csv", these contain a mapping of users to IDs. This is intended for a potential follow up survey, however this would require deanonymization and thus further consent by the users themselfs. This file may also **never** leave the mail system.

## Calls

pytest (--runslow) (-n 4) - unit tests, should be called once in any case to test if all basic functions work as expected
cat input.json | python main.py -u <user> -d <domain> - The analysis script, u and d are hashed and used as file names, thus no conclusions can be drawn from the file names, but the results are distributed and provide unique mapping. Without input of u and d the input file "data/fetch_json_new.json" is used. The unit tests pass other file names to the main method to run their tests.
 
Input: Json file as transfer from Dovecot
Output: CSV file
Salt: A random salt is generated the first time the script is called (main.py), which is used for hashes and reused on subsequent calls. The salt should **never** leave the system. Otherwise the data can be de-pseudomized.

## main.py - Sequence

Results are stored in results/<pseudomizer.hash(domain, salt)>/<pseudomizer.hash(user, salt)>.csv.

### Program Flow overview:
main
* results/mailmapping.csv
       
    * Used to store hash and mail addresses

* parser.json_data_to_mails splits JSON into single mails
    * parser.parse_json_mail:  
    Header, body and the attachments are processed separately. Processing searches for keywords, such as To, CC, ------ Begin PGP. The list of keywords is stored in a single file (KeyWords). After necessary information is retrieved from the e-mail, there is a list of these entries.
Depending on encountered keywords in the body a corresponding method is called in PGPHelper.
The header and body data are thrown into mailparser.parse_from_string(), from the result the user agent is extracted and iterated over the attachments.
    * parsers.parse_attachment():
    Depending on the encountered keywords or file names in attachments PGPHelper or CertHelper is called.

* Result CSV is opened
* Headers are written
* MailRecord.MailRecord serves as a model that we use to output the desired data. The fields to be output are defined in MailRecord.headers.
Pseudomizer.pseudomize(mail, salt): Called to pseudonymize personal data. Names, addresses, ids etc. are hashed with SHA256 and salt which is **only** available on this system. Other data, such as folders or dates, are assigned to specific categories. 
* Finally, the result is written to the csv file.

 
## Files:

* Pseudomizer: Here information is pseudonymized if necessary. There are certain functions to pseudonymize DKIM, PGP-Keys, certificates, body structure, date and timestamp, folders, addresses etc.. For this there is also a hash function using SHA256, which expects an additional salt. The salt must not be stored together with the data.
* Parser: An e-mail is parsed here.
* CertHelper processes S/MIME certificates.
* PGPHelper: Here PGP signatures, keys etc. are processed by gnupg. Unfortunately the error output must always be processed. 
* KeyWords: A collection of all possible keywords.
* main
* dovecot.sh: Calls the dovecot server to output json for a single user to the main script

## Dependencies and requirements for the system

We assume that a Dovecot mail server is evaluated. For other systems the json parsing has to be adapted.
Furthermore GnuPG must be available. Further dependencies can be found in the Poetry file.
