#from sqlalchemy import Column, Boolean, Integer, String, ARRAY, DateTime, ForeignKey, Enum
#from sqlalchemy.orm import relationship
#from records.BaseModel import Base
from records.Types import Postbox, MIME, find_postbox, SMIMESig, SMIMEEnc, PGPSig, PGPEnc
from parser.KeyWords import ParserKeyWord, ParsedItem, JsonLevelKeyWord, ParsedData
from parser.Pseudomizer import extract_address
import datetime
import json

''' Original plan (missing):
	- (optional) PGP-Header ID Array
	- (optional) PGP-Inline ID Array
	- (optional) S/MIME Cert ID Array
	- Nachrichtenverlauf (Reference, RandomID)
	- Relays (Array)

 Open question:
 	- Attached keys?
 	- How much DKIM, PGP stuff?
 '''

headers = ["user","sender", "receiver_list", "cc_list", "bcc_list", "reply_to", "return_path",
                "header_date",
                #"receiver_date",
                "client", #"xmailer",
                "folder",
                "mail_relay_count", "mail_virus_scans",
                #"return_path_equals_sender",
                #"user_equals_to_sender",
                #"incoming_mail",
                "message_id", "is_list", #"dkim",
                "cert_data", "smime_signature", "smime_encryption", "autocrypt_data",
                "pgp_key_data_attachment", "pgp_signature_attachment", "pgp_encryption_attachment",
                "pgp_key_data_inline", "pgp_signature_inline", "pgp_encryption_inline",
                "pgp_inline_forward",
                "smime_error", "pgp_error", "other_error"]

class MailRecord():
    user = None # Internal Mailbox id
    sender = "empty"
    receiver_list = set()
    cc_list = set()
    bcc_list = set()  # Probably only for sent mails?
    reply_to = None
    return_path = None # Set to eq if it equals to sender mail
    header_date = None
    receiver_date = None
    client = -1
    folder = Postbox.undefined
    mail_relay_count = -1
    mail_virus_scans = -1
    #return_path_equals_sender = None
    #user_equals_to_sender = None
    incoming_mail = -1
    message_id = None
    dkim = None
    cert_data = None
    smime_signature = SMIMESig.un
    smime_encryption = SMIMEEnc.un
    autocrypt_data = None
    pgp_key_data_attachment = None
    pgp_signature_attachment = PGPSig.un
    pgp_encryption_attachment = PGPEnc.un
    pgp_key_data_inline = None
    pgp_signature_inline = PGPSig.un
    pgp_encryption_inline = PGPEnc.un
    pgp_inline_forward = ""
    is_list = 0

    smime_error = ""
    pgp_error = ""
    other_error = ""

    def __init__(self, pseudomized_email):
        #Reinit values - caused some weird behaviour otherwise
        self.user = None  # Internal Mailbox id
        self.sender = "empty"
        self.receiver_list = set()
        self.cc_list = set()
        self.bcc_list = set()  # Probably only for sent mails?
        self.reply_to = None
        self.return_path = None  # Set to eq if it equals to sender mail
        self.header_date = None
        self.receiver_date = None
        self.client = -1
        self.folder = Postbox.undefined
        self.mail_relay_count = -1
        self.mail_virus_scans = -1
        # return_path_equals_sender = None
        # user_equals_to_sender = None
        self.incoming_mail = -1
        self.message_id = None
        self.dkim = None
        self.cert_data = None
        self.smime_signature = SMIMESig.un
        self.smime_encryption = SMIMEEnc.un
        self.autocrypt_data = None
        self.pgp_key_data_attachment = None
        self.pgp_signature_attachment = PGPSig.un
        self.pgp_encryption_attachment = PGPEnc.un
        self.pgp_key_data_inline = None
        self.pgp_signature_inline = PGPSig.un
        self.pgp_encryption_inline = PGPEnc.un
        self.pgp_inline_forward = ""
        self.is_list = 0

        self.smime_error = ""
        self.pgp_error = ""
        self.other_error = ""

        for item in pseudomized_email:
            try:
                item.merge()
                if item.key == JsonLevelKeyWord.USER:
                    self.user = item.value
                elif item.key == ParserKeyWord.RECEIVED_DATE:
                    try:
                        self.receiver_date = datetime.datetime.strptime(item.value+" 0", "%U:%Y  %w").date()
                    except:
                        print("[MailRecord] invalid r Date found")
                        self.receiver_date = "invalid"
                        if len(self.other_error) > 0:
                            self.other_error += "|invalid_date"
                        else:
                            self.other_error = "invalid_date"
                elif item.key == ParserKeyWord.SENDER:
                    self.sender = extract_address(item.value)[0]
                elif item.key == ParserKeyWord.TO:
                    self.receiver_list.add(item.value)
                elif item.key == ParserKeyWord.CC:
                    self.cc_list.add(item.value)
                elif item.key == ParserKeyWord.BCC:
                    self.bcc_list.add(item.value)
                elif item.key == ParserKeyWord.REPLY_TO:
                    self.reply_to = extract_address(item.value)[0]
                elif item.key == ParserKeyWord.RETURN_PATH:
                    self.return_path = extract_address(item.value)[0]
                elif item.key == ParserKeyWord.HEADER_DATE:
                    try:
                        #Convert KW:Year-Weekday(Always set to 1 for pseudo) to datetime:
                        self.header_date = datetime.datetime.strptime(item.value+" 0", "%U:%Y  %w").date()
                    except:
                        print("[MailRecord] invalid h Date found")
                        self.header_date = "invalid"
                        if len(self.other_error) > 0:
                            self.other_error += "|invalid_date"
                        else:
                            self.other_error = "invalid_date"
                elif item.key == ParserKeyWord.MSG_ID:
                    self.message_id = item.value
                elif item.key == ParserKeyWord.USER_AGENT:
                    self.client = item.value
                elif item.key == ParserKeyWord.XMAILER:
                    self.client = item.value
                elif item.key == ParserKeyWord.BODY_STRUCTURE:
                    pass
                elif item.key == ParserKeyWord.DKIM:
                    self.dkim = item.value
                elif item.key == ParsedData.IS_MAILINGLIST:
                    if item.value:
                        self.is_list = 1
                    else:
                        self.is_list = 0
                elif item.key == JsonLevelKeyWord.MAILBOX:
                    self.folder = item.value
                elif item.key == ParserKeyWord.RECEIVED:
                    self.mail_relay_count = item.value
                elif item.key == ParserKeyWord.Virus_SCAN:
                    self.mail_virus_scans = item.value
                elif item.key == ParsedData.CERT_DATA:
                    self.cert_data = item.value
                elif item.key == ParsedData.SMIME_SIGNATURE:
                    self.smime_signature = item.value
                elif item.key == ParsedData.SMIME_ENCRYPTION:
                    self.smime_encryption = item.value
                elif item.key == ParserKeyWord.AUTOCRYPT:
                    self.autocrypt_data = item.value
                elif item.key == ParsedData.PGP_KEY_DATA_ATTACHMENT:
                    self.pgp_key_data_attachment = item.value
                elif item.key == ParsedData.PGP_SIGNATURE_ATTACHMENT:
                    self.pgp_signature_attachment = item.value
                elif item.key == ParsedData.PGP_ENCRYPTION_ATTACHMENT:
                    self.pgp_encryption_attachment = item.value
                elif item.key == ParsedData.PGP_KEY_DATA_INLINE:
                    self.pgp_key_data_inline = item.value
                elif item.key == ParsedData.PGP_SIGNATURE_INLINE:
                    self.pgp_signature_inline = item.value
                elif item.key == ParsedData.PGP_ENCRYPTION_INLINE:
                    self.pgp_encryption_inline = item.value
                elif item.key == ParsedData.PGP_INLINE_FORWARD:
                    self.pgp_inline_forward = item.value
                elif item.key == ParsedData.SMIME_ERROR:
                    self.smime_error = item.value
                elif item.key == ParsedData.PGP_ERROR:
                    self.pgp_error = item.value
                elif item.key == ParsedData.OTHER_ERROR:
                    if len(self.other_error) > 0:
                        self.other_error += "|"+item.value
                    else:
                        self.other_error = item.value
            except Exception as e:
                print("[MailRecord] Exception catched: ", e.__class__.__name__)
                if len(self.other_error) > 0:
                    self.other_error += "|mrwrite"
                else:
                    self.other_error = "mrwrite"
        if self.return_path is not None:
            if self.return_path == self.sender:
                self.return_path_equals_sender = 1
                self.return_path = "eq"
            else:
                self.return_path_equals_sender = 0

        self.cc_list = json.dumps(sorted(self.cc_list))
        self.bcc_list = json.dumps(sorted(self.bcc_list))
        self.receiver_list = json.dumps(sorted(self.receiver_list))

    def get_fields(self):
        for field in headers:
            yield getattr(self, field)

    def get_list(self):
        return self.get_fields()