import enum

@enum.unique
class Group(enum.Enum):
    external = 0
    student = 1
    personal = 2

@enum.unique
class Charset(enum.Enum):
    ISO885915 = "ISO885915"
    USASCII = "USASCII"
    Other = "Other"

@enum.unique
class SMIMESig(enum.Enum):
    un = 0 # un
    sig = 1 # sig
    part = 2 # partially sig
    unknown = 3 #Used when enc mail encountered

@enum.unique
class PGPSig(enum.Enum):
    un = 0 # un
    sig = 1 # sig
    part = 2 # partially sig
    unknown = 3 #Used when enc mail encountered

@enum.unique
class SMIMEEnc(enum.Enum):
    un = 0
    enc = 1

@enum.unique
class PGPEnc(enum.Enum):
    un = 0
    enc = 1

@enum.unique
class Postbox(enum.Enum):
    undefined = -1
    inb = 0
    inb_sub = 1
    sent = 2
    sent_sub = 3
    spam = 4
    trash = 5
    trash_30d = 6
    junk = 7
    Other = 8

@enum.unique
class DKIMState(enum.Enum):
    missing = 0
    valid = 1
    invalid = 2

@enum.unique
class MIME(enum.Enum):
    TextPlain = "TextPlain"
    PKCSSignature = "PKCSSignature"
    SMIMEEnc = "SMIMEEnc"
    Message = "Message"
    Other = "Other"

def find_mime(text):
    #See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Complete_list_of_MIME_types
    text = text.lower()
    if("text/plain" in text):
        return MIME.TextPlain
    elif("application/pkcs7-signature" in text):
        return MIME.PKCSSignature
    elif("filename/smime.p7s" in text):
        return MIME.SMIMEEnc # Should check, if this is correct for signed mails
    elif("message" in text):
        return MIME.Message
    else:
        return MIME.Other

def find_postbox(name):
    name = name.lower()
    if(name == "inbox"):
        return Postbox.inb
    elif(name == "sent"):
        return Postbox.sent
    if(name == "inbox_sub"):
        return Postbox.inb_sub
    elif(name == "sent_sub"):
        return Postbox.sent_sub
    elif(name == "spam"):
        return Postbox.spam
    elif(name == "trash"):
        return Postbox.trash
    elif(name == "30dtrash"):
        return Postbox.trash_30d
    elif(name == "junk"):
        return Postbox.junk
    else:
        return Postbox.Other

def find_charset(text):
    text = text.upper()
    for charset in Charset:
        if text == charset.name:
            return charset
    return Charset.Other
