import pytest
import os
import main

'''
 Test with a larger testset and ensure that no personal data can be found in the results file and that the result file
 contains the expected counts. 
'''


def test_run_larger_and_ensure_no_personal_data():

    blacklisted_words = ["oliver",
                         "wiese",
                         "stransky",
                         "sec.uni-hannover.de",
                         "berlin", #FU Berlin
                         "emailsecurity", #Testmail target
                         "9492113065938320616164791897", #Serial number of Olivers cert
                         "2018-02-20", #Validity of SMIME Cert
                         "2021-02-19", #Validity of SMIME Cert
                         "63AC18A5C2576AB5", #PGP Key Oliver
                         "2EF06E7444F3E96F" #PGP Key Oliver
                         ]
    results_test_file = "results/results_test.csv"
    try:
        #Delete previous test data
        os.remove(results_test_file)
    except FileNotFoundError:
        pass
    main.main(json_input = "data/fetch_json_new.json", csv_output = results_test_file, salt = "123".encode())

    #Ensure we don't find any of the personal data in the results file
    with open(results_test_file, 'r') as testfile:
        data = testfile.read().lower()
        for word in blacklisted_words:
            assert word.lower() not in data, word+" was found in the results file."

        assert(data.count("\n") == 35) # We have 36 Testmails+1 Header row
        assert(data.count("not_valid_b") == 6) # 2 Testmails are signed with an SMIME Cert and are signed by a root and intermediate CA each
        #assert(data.count("c=relaxed/simple") == 2) # 2 DKIM Testmails
        assert(data.count("emailprotection") == 2) # Olivers Cert allows for email Protection as extended_key_usage, should appear 2 times
        assert(data.count("\"subj\":3,\"subju\":2") == 2) # Check that our count for mail addresses in subject alt names, works
        assert(data.count("SMIMESig.sig".lower()) == 2) # 2 Testmails are sig with SMIME
        assert(data.count("SMIMESig.un".lower()) == 32) # 32 Testmails are not sig with SMIME
        assert(data.count("SMIMEEnc.un".lower()) == 34) # 34 Testmails are not enc with SMIME
        assert(data.count("SMIMEEnc.enc".lower()) == 0) # 0 Testmails are enc with SMIME
        assert(data.count(",eq,".lower()) == 20) # All 20 incoming Mails have same return-path