from parser import parser
from parser.KeyWords import ParserKeyWord

session = None

simpleMail = """user: user.mustermann@example.com
    date.received: 2017-09-05 12:07:50
    hdr.from: from@example.com
    hdr.to: example@test.com
    hdr.date: Tue, 05 Sep 2017 12:07:50 +0200
    hdr.message-id: <E1dpAlq-12345p-LB@example.com>
    imap.bodystructure: "text" "plain" ("charset" "us-ascii") NIL NIL "7bit" 188 14 NIL NIL NIL NIL
    hdr.received: server1
    hdr.dkim-signature: """

simpleBrokenMail = """user: user.mustermann@example.com
    date.received: 2017-09-05 12:07:50
    hdr.from: from@example.com
    hdr.to: example@test.com
    hdr.date:
    hdr.message-id: <E1dpAlq-12345p-LB@example.com>
    imap.bodystructure: "text" "plain" ("charset" "us-ascii") NIL NIL "7bit" 188 14 NIL NIL NIL NIL
    hdr.received: server1
    hdr.dkim-signature: """

simpleJsonMail = "[{'user': 'user@example.com', 'mailbox': 'INBOX', 'body': 'This is a plain mail from mac'}]"

'''@pytest.fixture(scope='session', autouse=True)
def engine():
    return create_engine('postgres://postgres:test123@localhost')


@pytest.yield_fixture(scope='session', autouse=True)
def tables(engine):
    BaseModel.Base.metadata.create_all(engine)
    yield
    BaseModel.Base.metadata.drop_all(engine)


@pytest.yield_fixture(scope='session', autouse=True)
def dbsession(engine, tables):
    global session
    """Returns an sqlalchemy session, and after the test tears down everything properly."""
    connection = engine.connect()
    # begin the nested transaction
    transaction = connection.begin()
    # use the connection with the already started transaction
    session = Session(bind=connection)

    yield session

    session.close()
    # roll back the broader transaction
    transaction.rollback()
    # put back the connection to the connection pool
    connection.close()
'''

def test_simple_mail():
    parsed_items = parser.parse_mail(simpleMail, False)
    for item in parsed_items:
        print(item.key)
        if item.key == ParserKeyWord.USER:
            assert (item.value == "user.mustermann@example.com")
        elif item.key == ParserKeyWord.RECEIVED_DATE:
            assert (item.value == "2017-09-05 12:07:50")
        elif item.key == ParserKeyWord.SENDER:
            assert (item.value == "from@example.com")
        elif item.key == ParserKeyWord.TO:
            assert (item.value == "example@test.com")
        elif item.key == ParserKeyWord.HEADER_DATE:
            assert (item.value == "Tue, 05 Sep 2017 12:07:50 +0200")
        elif item.key == ParserKeyWord.MSG_ID:
            assert (item.value == "<E1dpAlq-12345p-LB@example.com>")
        elif item.key == ParserKeyWord.BODY_STRUCTURE:
            assert (item.value == """"text" "plain" ("charset" "us-ascii") NIL NIL "7bit" 188 14 NIL NIL NIL NIL""")
        elif item.key == ParserKeyWord.RECEIVED:
            assert (item.value == "server1")
        elif item.key == parser.KeyWords.ParserKeyWord.DKIM:
            assert (item.value == "")
        else:
            raise ValueError("Too much items! key word: %s value: %s", item.key, item.value)
    assert (len(parsed_items) == 9)

def test_simple_broken_mail():
    parsed_items = parser.parse_mail(simpleBrokenMail, False)
    for item in parsed_items:
        print(item.key)
        if item.key == ParserKeyWord.USER:
            assert (item.value == "user.mustermann@example.com")
        elif item.key == ParserKeyWord.RECEIVED_DATE:
            assert (item.value == "2017-09-05 12:07:50")
        elif item.key == ParserKeyWord.SENDER:
            assert (item.value == "from@example.com")
        elif item.key == ParserKeyWord.TO:
            assert (item.value == "example@test.com")
        elif item.key == ParserKeyWord.HEADER_DATE:
            assert (item.value == "")
        elif item.key == ParserKeyWord.MSG_ID:
            assert (item.value == "<E1dpAlq-12345p-LB@example.com>")
        elif item.key == ParserKeyWord.BODY_STRUCTURE:
            assert (item.value == """"text" "plain" ("charset" "us-ascii") NIL NIL "7bit" 188 14 NIL NIL NIL NIL""")
        elif item.key == ParserKeyWord.RECEIVED:
            assert (item.value == "server1")
        elif item.key == parser.KeyWords.ParserKeyWord.DKIM:
            assert (item.value == "")
        else:
            raise ValueError("Too much items! key word: %s value: %s", item.key, item.value)
    assert (len(parsed_items) == 9)

'''
def test_store_simple_mail():
    parsed_items = parser.parse_mail(simpleMail, False)
    record = MailRecord(incoming_mail=1)
    session.add(record)
    session.commit()

    assert session.query(MailRecord).count() == 1
'''

def test_json_mail():
    mails = parser.parse_json_string(simpleJsonMail)
    assert (len(mails) == 1)
    for mail in mails:
        for item in mail:
            if item.key == parser.KeyWords.JsonLevelKeyWord.USER:
                assert (item.value == "user@example.com")
            elif item.key == parser.KeyWords.JsonLevelKeyWord.MAILBOX:
                assert (item.value == "INBOX")
            elif item.key == parser.KeyWords.JsonLevelKeyWord.BODY:
                assert (item.value == "This is a plain mail from mac")
            else:
                raise ValueError("Too much items! key word: %s value: %s", item.key, item.value)
