from parser.Pseudomizer import hash, extract_address, pseudo_dkim, pseudo_pgp, categorize_date
from parser import PGPHelper, parser, KeyWords

def test_hashing():
    v1 = "Lorem ipsum dolor sit amet, consectetur adipiscing elit"
    v2 = "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua"
    salt = b"E1F53135E559C253"
    h1 = hash(v1, salt)
    h2 = hash(v2, salt)
    print(v1,v2)
    print(h1)
    print(h2)
    assert (h1 != h2)
    h3 = hash(v1,salt)
    print(h3)
    assert (h3 == h1)

def test_email_address_extraction():
    pure_adr = "alice@example.com"
    res = extract_address(pure_adr)
    print(res)
    assert (len(res) == 1)
    assert (res[0] == pure_adr)
    adr1 = "alice <alice@example.com>"
    res = extract_address(adr1)
    print(res)
    assert (len(res) == 1)
    assert (res[0] == pure_adr)
    adr2 = "<alice@example.com>"
    res = extract_address(adr2)
    print(res)
    assert (len(res) == 1)
    assert (res[0] == pure_adr)
    adr3 = "alice@example.com <alice@example.com>"
    res = extract_address(adr3)
    print(res)
    assert (len(res) == 1)
    assert (res[0] == pure_adr)
    adr4 = "alice@example.com, bob@example.com, charlie@example.com"
    res = extract_address(adr4)
    print(res)
    assert (len(res) == 3)
    assert (pure_adr in res and "bob@example.com" in res and "charlie@example.com"in res)

def test_categorize_dkim():
    print("Start testing DKIM")
    salt = b"E1F53135E559C253"
    dkim1 = "v=1; a=rsa-sha258; d=example.net; s=brisbane; c=relaxed/simple; q=dns/txt; t=1117574938; x=1118006938; h=from:to:subject:date:keywords:keywords;bh=MTIzNDU2Nzg5MDEyMzQ1Njc4OTAxMjM0NTY3ODkwMTI=;b=dzdVyOfAKCdLXdJOc9G2q8LoXSlEniSbav+yuU4zGeeruD00lszZVoG4ZHRNiYzR"
    assert(pseudo_dkim(dkim1, salt) == "a=rsa-sha258;c=relaxed/simple;d=14798b0422660b90164595b00b9d23de1572df1f;h=from:to:subject:date:keywords:keywords;q=dns/txt;t=22:2005;v=1;x=22:2005;")
    #TODO Add asserts
    dkim2 = "v=1;  a=rsa-sha256;  c=relaxed/relaxed;  d=example.com;"
    assert(pseudo_dkim(dkim2, salt) == "a=rsa-sha256;c=relaxed/relaxed;d=ed0fb4c58aba14fbe1e78f6fedc1f85c022edf8c;v=1;")

    dkim3 = "v=1; a=rsa-sha1; c=relaxed; d=xyz.com; i=abc@xyz.com;"
    assert(pseudo_dkim(dkim3, salt) == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=1;")

    dkim4 = "v=1; a=rsa-sha1; c=relaxed; d=xyz.com; i=@xyz.com;"
    assert(pseudo_dkim(dkim4, salt) == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=1;")

    dkim5 = "v=1; a=rsa-sha1; c=relaxed; d=xyz.com; i=abc@xyz.com; y=IDontExist"
    assert(pseudo_dkim(dkim5, salt) == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=1;")

    dkim6 = "v=1; a=rsa-sha1; c=relaxed@a; d=xyz.com; i=abc@xyz.com; y=IDontExist"
    assert(pseudo_dkim(dkim6, salt) == "a=rsa-sha1;c=skippedPseudoAt;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=1;")

    dkim7 = "v=1; a=testuser; c=relaxed; d=xyz.com; i=abc@xyz.com; y=IDontExist"
    assert(pseudo_dkim(dkim7, salt, "testuser", "testdomain") == "a=skippedPseudoPriv;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=1;")

    dkim8 = "v=testdomain; a=rsa-sha1; c=relaxed; d=xyz.com; i=abc@xyz.com; y=IDontExist"
    assert(pseudo_dkim(dkim8, salt, "testuser", "testdomain") == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=skippedPseudoPriv;")

    dkim9 = "v=OnlyAPartOftestdomainIsHere; a=rsa-sha1; c=relaxed; d=xyz.com; i=abc@xyz.com; y=IDontExist"
    assert(pseudo_dkim(dkim9, salt, "testuser", "testdomain") == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;i=7c6c76c05436941c281eae11e2d69dcbbd670d2d@eaada6f3ebbb26b66c8490b2cef05728994be9bb;v=skippedPseudoPriv;")

    dkim10 = "v=1; a=rsa-sha1; c=relaxed; d=testdomain.com; i=testuser@testdomain.com; y=IDontExist"
    assert(pseudo_dkim(dkim10, salt, "testuser", "testdomain") == "a=rsa-sha1;c=relaxed;d=5daa68fd1490d9211425eb4b9ac7a211c6d06d36;i=d639ef29033ee9f32272dfc1c8bebc4a67bbc851@5daa68fd1490d9211425eb4b9ac7a211c6d06d36;v=1;")

    dkim11 = "v=1; a=rsa-sha1\n@unneccessaryLineBreak; c=relaxed\r\n\r\ntestuser; d=testdomain.com; i=testuser@testdomain.com; y=IDontExist"
    assert(pseudo_dkim(dkim11, salt, "testuser", "testdomain") == "a=skippedPseudoAt;c=skippedPseudoPriv;d=5daa68fd1490d9211425eb4b9ac7a211c6d06d36;i=d639ef29033ee9f32272dfc1c8bebc4a67bbc851@5daa68fd1490d9211425eb4b9ac7a211c6d06d36;v=1;")

    return

def test_date():
    assert(categorize_date("Mon, 06 Jan 2020 20:00:16 +0100") == "01:2020")
    assert(categorize_date("Tue, 07 Jan 2020 09:00:17 +0100") == "01:2020")
    assert(categorize_date("Tue, 14 Jan 2020 09:00:17 +0100") == "02:2020")
    assert(categorize_date("Tue, 14 Jan 2020 09:00:17 +0100") == "02:2020")
    assert(categorize_date("Thu, 2 May 2019 08:06:17 +0100 (GMT+01:00)") == "17:2019")
    assert(categorize_date("Thu, 6 May 2019 00:00:01 +1200 (GMT+12:00)") == categorize_date("Thu, 6 May 2019 00:00:01 +1200"))

def test_pgp_pseudo():
    keys = PGPHelper.demo() # Just a demo key
    salt = b"E1F53135E559C253"
    for k in keys:
        key = pseudo_pgp(k, salt)
        assert (key['type'] == 'pub')
        assert (key['length'] == '4096')
        assert (key['algo'] == '1')
        assert (key['keyid'] != '20B23B6D439EE43C')
        assert (key['date'] == '27:2019')
        assert (key['cap'] == 'scESC')
        assert (key['uids'] == 3)


test_dkim = '''
Delivered-To: stransky@sec.uni-hannover.de
Received: from mail-proxy2.luis.uni-hannover.de ([127.0.0.1])
	by imap1.p6.rrzn.uni-hannover.de (Dovecot) with LMTP id k3UmFdLgylz4XQAAZsCWPQ
	for <stransky@sec.uni-hannover.de>; Thu, 02 May 2019 14:21:38 +0200
Received: from mail-proxy2.luis.uni-hannover.de ([127.0.0.1])
	by mail-proxy2.luis.uni-hannover.de (Dovecot) with LMTP id bKvSInHgylwneAAAGtkGKA
	; Thu, 02 May 2019 14:21:38 +0200
Authentication-Results: mgw1-erl.srv.dfn.de (amavisd-new);
	dkim=pass (1024-bit key) header.d=xyz.de
Received: from c1121.mx.srv.dfn.de ([127.0.0.1])
	by localhost (mgw1-erl.srv.dfn.de [127.0.0.1]) (amavisd-new, port 20013)
	with ESMTP id HilhJRwXw9_X for <stransky@sec.uni-hannover.de>;
	Thu,  2 May 2019 14:21:37 +0200 (CEST)
DKIM-Signature: v=1; a=rsa-sha256; q=dns/txt; c=relaxed/relaxed; d=xyz.de; s=april2019;
	h=MIME-Version:Content-Type:In-Reply-To:References:Message-ID:Date:Subject:CC:To:From; bh=ubVlQuOgnxy321mT+M9m+SgdmSOJU2bWn5B0kNxgwQw=;
	b=hKpc2biiM0AcacXZy8acf5+4SOZF+3+Ma7HVEuw82/F0lKZQuL+RC0BdaOoBSjfu2aO1+HQpgjKBBofZWImjIp+hV7A14UbwJyUfFw4h+e9FsnsLxvcVw9xfMZ7VFN31rG0x6fm62D2aN94OTsTX//24a73yRvCvkhN+7NhtWww=;
To: 'Christian Stransky' <stransky@sec.uni-hannover.de>'''

test_dkim2 = '''Delivered-To: stransky@sec.uni-hannover.de
Received: from mail-proxy2.luis.uni-hannover.de ([127.0.0.1])
	by mail-proxy2.luis.uni-hannover.de (Dovecot) with LMTP id 7mghOvLQxF+pLwAAGtkGKA
	; Mon, 30 Nov 2020 12:51:01 +0100
Received: from mailin-int-3.uni-hannover.de (mailin-int-3.uni-hannover.de [130.75.2.116])
	by mail-proxy2.luis.uni-hannover.de (Postfix) with ESMTPS id CCAE740250
	for <stransky@sec.uni-hannover.de>; Mon, 30 Nov 2020 12:51:01 +0100 (CET)
Received: from b1121.mx.srv.dfn.de (b1121.mx.srv.dfn.de [194.95.234.47])
	by mailin-int-3.uni-hannover.de (Postfix) with ESMTPS id C34EE60109
	for <stransky@sec.uni-hannover.de>; Mon, 30 Nov 2020 12:51:01 +0100 (CET)
Received: from localhost (localhost [127.0.0.1])
	by b1121.mx.srv.dfn.de (Postfix) with ESMTP id B5580682
	for <stransky@sec.uni-hannover.de>; Mon, 30 Nov 2020 12:51:01 +0100 (CET)
X-DFN-Virus-Scanned: Debian amavisd-new at mgw1-tub.srv.dfn.de
Authentication-Results: mgw1-tub.srv.dfn.de (amavisd-new);
	dkim=pass (2048-bit key) header.d=xyz.com
Received: from b1121.mx.srv.dfn.de ([127.0.0.1])
	by localhost (mgw1-tub.srv.dfn.de [127.0.0.1]) (amavisd-new, port 20013)
	with ESMTP id 6jAqhzaZX1Fm for <stransky@sec.uni-hannover.de>;
	Mon, 30 Nov 2020 12:50:56 +0100 (CET)
DKIM-Signature: v=1; a=rsa-sha1; c=relaxed; d=xyz.com;
	 h=date:to:from:subject:message-id:mime-version:content-type:
	content-transfer-encoding; s=tlddkim; bh=0gXI92OKQW+BC+id90foMkw
	fqJI=; b=VygQ2qbptR/CxNAUFzSIkFoVrBiePenpuscd4VO5xRuAzU/WDRf88Ou
	Ci+EImK7mmbE8mLVNHx5rijILJ58mZBzJZdy79sCDwrq2Bb9tDqHl/M6gQGWJqs9
	794EBiaH6opu2AscW1+dW+sOqzPolzkCWCKK3Dx893LVLPzEkhlSD5twYVDMcWNw
	Deqs9VJELQqsk+ln2tyK2kVhAxqzJ+c4ANfU3JJOiec5lGNWbBs+OZWVtUFNpSXE
	hLNqOEwkMIsZ7ik2u92ot6aSdRdxkGvxWRZD2Nl7qlkwVFYIWYqzjNosMBGiqYzJ
	rJI8tIzwMsRHsRN7MCnyuP0mwOHEfjg==
To: stransky@sec.uni-hannover.de'''

def test_categorize_dkim_2():
    result = parser.parse_mail(test_dkim, True)
    salt = b"E1F53135E559C253"
    for item in result:
        if item.key == KeyWords.ParserKeyWord.DKIM:
            pseudo = pseudo_dkim(item.value, salt)
            assert (pseudo == "a=rsa-sha256;c=relaxed/relaxed;d=c10725b590e466eb4bb382659b58c2a91ed73ff3;h=MIME-Version:Content-Type:In-Reply-To:References:Message-ID:Date:Subject:CC:To:From;q=dns/txt;v=1;")
    return


def test_categorize_dkim_3():
    result = parser.parse_mail(test_dkim2, True)
    salt = b"E1F53135E559C253"
    for item in result:
        if item.key == KeyWords.ParserKeyWord.DKIM:
            pseudo = pseudo_dkim(item.value, salt)
            assert (pseudo == "a=rsa-sha1;c=relaxed;d=eaada6f3ebbb26b66c8490b2cef05728994be9bb;h=date:to:from:subject:message-id:mime-version:content-type:content-transfer-encoding;v=1;")
    return